<?php

namespace Modules\BadrRequest\Entities;

use App\Models\Scopes\ShopIdScope;
use App\Services\Invoice;
use App\Models\User;
use App\Services\Updaters\InvoiceClient;
use App\Services\Updaters\InvoiceSafe;
use App\Traits\RoundingAttributesTrait;
use Illuminate\Database\Eloquent\Model;
use Modules\Invoices\Entities\Client;
use Modules\Invoices\Entities\InvoiceDetails;
use Modules\Invoices\Entities\InvoiceDetailsVats;
use Modules\Invoices\Entities\Einvoices;

class Invoices extends Model
{
    use RoundingAttributesTrait;
    protected $guarded = [];
    protected $table   = 'invoices_';
    protected $appends = ['einvoice_stage'];
    public $timestamps = false;
    // public $incrementing = true;

    protected static $available_types = ['sales', 'purchases', 'returned sales', 'returned purchases', 'sales offers'];
    private static $available_types_no = [
        'purchases' => 0,
        'sales' => 1,
        'returned purchases' => 2,
        'returned sales' => 3,
        'sales offers' => 23,
    ];
    public static function getAvailableTypes()
    {
        return self::$available_types;
    }

    public function referenceBills()
    {
        return $this->hasMany(self::class, 'reference_id', 'id')->where('shop_id', auth()->user()->shop_id)->select('id', 'reference_id')->with('InvoiceDetails:id,invoice_id,item_id,unit_id,quantity,total_price');
    }

    public static function getInvoiceNo($invoice_type)
    {
        return self::$available_types_no[$invoice_type];
    }

    public function client()
    {
        return $this->belongsTo(Client::class, 'client_id');
    }

    public function User()
    {
        return $this->belongsTo(User::class, 'add_user');
    }

    public function issuer()
    {
        return $this->belongsTo(Client::class, 'add_user');
    }

    public function details()
    {
        return $this->hasMany(InvoiceDetails::class, 'invoice_id');
    }

    public function vats()
    {
        return $this->hasMany(InvoiceDetailsVats::class, 'invoice_id');
    }

    public function InvoiceDetails()
    {
        return $this->hasMany(InvoiceDetails::class, 'invoice_id');
    }

    public function einvoice()
    {
        return $this->hasOne(Einvoices::class, 'invoice_id');
    }

    public function einvoice_stages()
    {
        return $this->hasMany(EinvoiceStage::class, 'invoice_id');
    }

    public function getEinvoiceStageAttribute()
    {
        return $this->einvoice_stages()->whereIn('status', [0, 1, 2])->count();
    }

    public function employee()
    {
        return User::find($this->user_id)->user_name ?? '-';
    }

    public static function createNew(Invoice $invoice_obj, InvoiceClient $client, InvoiceSafe $safe)
    {
        $request = $invoice_obj->getRequest();

        $new_invoice = Invoices::create([
            'invoice_date' => $request->invoice_date,
            'transaction_date' => date('Y-m-d h:i:s'),
            'bill_total' => $request->invoice_total,
            'add_user' => auth()->user()->id,
            'client_id' => $request->client_id,
            'net_price' => $request->net,
            'payment' => $request->paid,
            'bill_payment' => $request->paid,
            'rest' => floatVal($request->net) - floatVal($request->paid),
            'store_id' => $request->store_id,
            'bill_no' => $request->bill_no ?? 0,
            'notes' => $request->notes,
            'shop_id' => get_session_val('serial_id', 'Integer'),
            'balance' => ($client->getNewBalance()),
            'currency_id' => $safe->getSafePoint()->currency_id,
            'currency_value' => $safe->getCurrencyValue(),
            'invoice_type' => $invoice_obj->getType(),
            'invoice_total_disc' => 0,
            'invoice_total_disc_type' => 1,
            'invoice_total_disc_status' => get_session_val('company_ratio', 'integer') == 2 ? 0 : 1,
            'insatllments_add'  => $request->installment_add_precent ?? 0,
            'reference_id'  => $request->reference_id ?? 0
        ]);

        return $new_invoice;
    }

    public function Store()
    {
        return $this->belongsTo(Store::class, 'store_id');
    }

    public function childInvoiceDetails()
    {
        return $this->hasManyThrough(InvoiceDetails::class, Invoices::class, 'reference_id', 'invoice_id')
                ->selectRaw('item_id , SUM(quantity) AS total_qty')->groupBy('item_id');
    }

    // one day , not now 😉
    protected static function booted()
    {
        // static::addGlobalScope(new ShopIdScope);
    }

}
