<?php

namespace Modules\Currency\Observers;

use Modules\Currency\Entities\Currency;
use Modules\Currency\Entities\CurrencyLog;
use Modules\Currency\Services\ExchangeRate;

class CurrencyObserver
{
    /**
     * Handle the Currency "creating" event.
     *
     * @param  \Modules\Currency\Entities\Currency  $currency
     * @return void
     */
    public function creating(Currency $currency)
    {
        $currency->shop_id = $currency->shop_id ?? auth()->user()->shop_id;
        $currency->value   = $currency->value ?? ExchangeRate::convert( $currency->code );
    }

    /**
     * Handle the Currency "created" event.
     *
     * @param  \Modules\Currency\Entities\Currency  $currency
     * @return void
     */
    public function created(Currency $currency)
    {
        $main = Currency::main();
        $currency->logs()->create([
            'log'  => "تم <b>إضافة</b> العملة '$currency->name' بسعر صرف $currency->value $main",
            'type' => CurrencyLog::CREATED,
            'details' => ['name' => $currency->name, 'value' => "$currency->value $main"]
        ]);
    }

    /**
     * Handle the Currency "updating" event.
     *
     * @param  \Modules\Currency\Entities\Currency  $currency
     * @return void
     */
    public function updating(Currency $currency)
    {
        // $currency->value = ExchangeRate::convert( $currency->code );
    }

    /**
     * Handle the Currency "updated" event.
     *
     * @param  \Modules\Currency\Entities\Currency  $currency
     * @return void
     */
    public function updated(Currency $currency)
    {
        $changes = $details = [];
        $columns = $currency->getChanges();

        if (isset($columns['name'])) {
            $changes[] = "تم <b>تغيير</b> <u>إسم العملة</u> من  {$currency->getOriginal('name')}  إلي {$columns['name']}";
            $this->setLogDetails($currency, 'name', $details);
        }

        if (isset($columns['code'])) {
            $changes[] = "تم <b>تغيير</b> <u>رمز العملة</u> من  {$currency->getOriginal('code')}  إلي {$columns['code']}";
            $this->setLogDetails($currency, 'code', $details);
        }

        if (isset($columns['value']) && $columns['value'] != $currency->getOriginal('value')) {
            $main_name = Currency::main();
            $changes[] = "تم <b>تغيير</b> <u>سعر الصرف</u> من  {$currency->getOriginal('value')}  $main_name  إلي ".round_money($columns['value'])." $main_name";
            $this->setLogDetails($currency, 'value', $details);
        }

        if (count($changes)) {
            $currency->logs()->create([
                    'log' => implode('<br>', $changes),
                    'type' => CurrencyLog::UPDATED,
                    'details' => $details
                ]);
        }
    }

    protected function setLogDetails(Currency $currency, string|float $column, array &$details): array
    {
        $main_name = $column == 'value' ? " ".Currency::main() : '';
        return $details[$column] = [
            'column' => $column,
            'old'    => $currency->getOriginal($column) . $main_name,
            'new'    => $currency->$column . $main_name
        ];
    }

    /**
     * Handle the Currency "deleted" event.
     *
     * @param  \Modules\Currency\Entities\Currency  $currency
     * @return void
     */
    public function deleted(Currency $currency)
    {
        // $currency->logs()->delete();
        $currency->logs()->create([
                'log' => "تم <b>حذف</b> العملة '$currency->name'",
                'type' => CurrencyLog::DELETED,
                'details' => ['name' => $currency->name, 'value' => "$currency->value ".Currency::main()]
            ]);
    }
}

