<?php

namespace Modules\Dairy\Http\Services;

use Carbon\Carbon;
use Exception;
use Illuminate\Support\Facades\DB;
use Modules\Dairy\Entities\{Badrshop, Client, ClientTransaction, Invoice, InvoiceDetail, Item, ItemTransaction, Log, SalePoint, Store, StoreItem};

class InvoiceService
{
    protected Calculator $calculator;
    protected Badrshop $shop;
    protected SalePoint $salePoint;
    protected Store $store;
    protected Client $client;
    protected Invoice $invoice;
    protected Item $item;
    protected float $new_quantity;
    protected float $old_quantity;
    protected $type;

    public function __construct()
    {
    }

    public function handler(array $invoice = [], int $type = 1)
    {
        $this->type = $type;
        try {
            DB::beginTransaction();

                $this->init($invoice) // To set global properties
                        ->setCalculator($invoice) // To set calculation object
                        ->updateSalePointMoney() // To update sale point money
                        ->updateClientBalance() // To update client balance
                        ->updateItemQuantity() // To update item quantity
                        ->createInvoice($invoice) // To create new invoice
                        ->createInvoiceDetails($invoice['item_id']) // To Create invoice details
                        ->createClientTransaction() // To Create client transaction
                        ->createItemTransaction() // To Create item transaction
                        ->newInvoiceLog(); // To create invoice log

            DB::commit();

            session()->flash('invoice-print', route('invoices.print', $this->invoice->id));
            return ['status' => 200, 'message' => 'تم انشاء الفاتورة بنجاح', 'invoice_id' => $this->invoice->id];
        } catch (Exception $e) {
            return ['status' => 500, 'message' => $e->getMessage()];
        }
    }

    protected function init(array $invoice): InvoiceService
    {
        $this->salePoint = SalePoint::select('id', 'currency_id', 'money_point')->whereId( auth()->user()->sale_point )->first();
        $this->store  = Store::select('id')->whereId( $invoice['store_id'] )->first();
        $this->client = Client::select('id', 'balance', 'price')->find($invoice['client_id']);
        $this->item   = Item::select('id', 'unit_id', 'quantity', 'item_name')->where('id', $invoice['item_id'])->first();

        return $this;
    }

    protected function createInvoice(array $invoice): InvoiceService
    {
        if (! isset($invoice['date'])) { $invoice['date'] = date('Y-m-d'); }

        $this->invoice = Invoice::create([
            'shop_id'      => auth()->user()->shop_id,
            'client_id'    => $this->client->id,
            'store_id'     => $this->store->id,
            'balance'      => $this->client->balance,
            'bill_total'   => $this->calculator->total,
            'net_price'    => $this->calculator->net_price,
            'bill_payment' => $this->calculator->total,
            'invoice_type' => $this->type,
            'invoice_date' => Carbon::parse($invoice['date'])->format('Y-m-d') . ' ' . now()->format('H:i:s'),
            'payment'      => $this->calculator->payment,
            'rest'         => $this->calculator->rest,
            'fee'          => $this->calculator->fee,
            'fee_type'     => 1,
            'order_no'     => $invoice['order_no'] ?? 0,
            'notes'        => $invoice['notes'],
        ]);

        return $this;
    }

    protected function createInvoiceDetails(): InvoiceService
    {
        InvoiceDetail::create([
                    'shop_id'        => auth()->user()->shop_id,
                    'invoice_id'     => $this->invoice->id,
                    'item_id'        => $this->item->id,
                    'quantity'       => $this->item->quantity,
                    'quantity_after' => $this->new_quantity,
                    'unit_id'        => $this->item->unit_id ?? 0,
                    'sale_price'     => $this->calculator->price,
                    'pay_price'      => $this->calculator->price,
                    'total_price'    => $this->calculator->price * $this->calculator->quantity,
                    'quantity'       => $this->calculator->quantity,
                    'price'          => $this->calculator->net_price,
                    'water'          => $this->calculator->water,
                    'protein'        => $this->calculator->protein,
                    'fats'           => $this->calculator->fats,
                    'water_discount' => $this->calculator->water_discount,
                    'protein_bonus'  => $this->calculator->protein_bonus,
                    'fats_bonus'     => $this->calculator->fats_bonus,
                ]);

        return $this;
    }

    protected function newInvoiceLog()
    {
        $message = $this->type ? trans('log.new_sale_invoice') : trans('log.new_purchase_invoice');
        Log::create(['event' => $message . trans('log.invoice_no') . $this->invoice->id, 'type' => $this->type]);
    }

    protected function setCalculator(array $invoice): InvoiceService
    {
        $this->calculator = new Calculator();

        $this->calculator->setQuantity($invoice['quantity'])->setWater($invoice['water'])->setProtein($invoice['protein'])
                        ->setFats($invoice['fats'])->setFee($invoice['fee'])->setPrice($invoice['price'] ?? $this->client->price)
                        ->setPayment($invoice['payment'])->calculate();
        return $this;
    }

    protected function updateSalePointMoney(): InvoiceService
    {
        $payment = $this->calculator->payment;
        if ( in_array($this->type, [0, 2]) ) $payment *= -1;
        $this->salePoint->increment('money_point', $payment);
        return $this;
    }

    protected function updateClientBalance(): InvoiceService
    {
        $rest = $this->calculator->payment;
        if ( in_array($this->type, [0, 2]) ) $rest *= -1;
        $this->client->increment('balance', $rest);
        return $this;
    }

    protected function updateItemQuantity(): InvoiceService
    {
        $row = StoreItem::where(['item_id' => $this->item->id, 'store_id' => $this->store->id])->first();
        $quantity = $this->calculator->quantity;
        if ( in_array($this->type, [1, 3]) ) $quantity *= -1;

        if ($row) {
            $this->old_quantity = $row->store_quant;
            $row->increment('store_quant', $quantity);
            $this->new_quantity = $row->store_quant;
        } else {
            $this->old_quantity = $this->item->quantity;
            $this->item->increment('quantity', $quantity);
            $this->new_quantity = $this->item->quantity;
        }

        return $this;
    }

    protected function createClientTransaction(): InvoiceService
    {
        ClientTransaction::create([
            'shop_id'        => auth()->user()->shop_id,
            'bill_id'        => $this->invoice->id,
            'amount'         => $this->invoice->payment,
            'pay_day'        => $this->invoice->invoice_date,
            'bill_net_total' => $this->invoice->net_price,
            'notes'          => $this->invoice->notes,
            'client_id'      => $this->client->id,
            'balance'        => $this->client->balance,
            'safe_point_id'  => $this->salePoint->id,
            'safe_balance'   => $this->salePoint->money_point,
        ]);

        return $this;
    }

    protected function createItemTransaction(): InvoiceService
    {
        ItemTransaction::create([
            'shop_id'        => auth()->user()->shop_id,
            'quantity'       => $this->calculator->quantity,
            'new_quantity'   => $this->new_quantity,
            'old_quantity'   => $this->old_quantity,
            'item_name'      => $this->item->item_name,
            'item_id'        => $this->item->id,
            'price'          => $this->calculator->price * $this->calculator->quantity,
            'store_id'       => $this->store->id,
            'type'           => $this->type,
            'action_date'    => $this->invoice->invoice_date,
        ]);

        return $this;
    }
}
