<?php

namespace Modules\EInvoice\Http\Controllers;

use App\Services\Einvoice\Einvoice;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Http\Response;
use Illuminate\Routing\Controller;
use Modules\EInvoice\Entities\Invoice;
use Modules\EInvoice\Entities\Einvoices;
use Symfony\Component\HttpFoundation\BinaryFileResponse;

class DocumentController extends Controller
{
    /**
     * Display a listing of the resource.
     * @return Response
     */
    public function index(Request $request)
    {
        $documents = Einvoices::orderBy('dateTimeIssued', 'desc')
                                ->whereHas('invoice', function ($query) {
                                    $query->where('invoices_.shop_id', get_session_val('serial_id', 'integer'));
                                });
        $statuses  = ["Valid" => __('einvoice.Valid'), "Invalid" => __('einvoice.Invalid'), "Rejected" => __('einvoice.Rejected'), "Cancelled" => __('einvoice.Cancelled')];

        // Filter by client name or invoice number or invoice date
        if ($request->has('search') && !in_array($request->search, ['', ' ', null])) {
            $documents = $documents->where(function ($query) use ($request) {
                $query->where('uuid', 'like', '%' . $request->search . '%')
                    ->orWhere('invoice_id', 'like', '%' . $request->search . '%')
                    ->orWhere('receiverName', 'like', '%' . $request->search . '%');
            });
        }

        if ($request->has('status') && !in_array($request->status, ['', ' ', null])) {
            $documents = $documents->where('status', $request->status);
        }

        if ($request->has('date_from') && !in_array($request->date_from, ['', ' ', null])) {
            $documents = $documents->whereDate('dateTimeIssued', '>=', $request->date_from);
        }

        if ($request->has('date_to') && !in_array($request->date_to, ['', ' ', null])) {
            $documents = $documents->whereDate('dateTimeIssued', '<=', $request->date_to);
        }

        $documents = $documents->get();

        $documents->each(function ($document) {
            $document->invoice = Invoice::where('uuid', $document->uuid)->orWhere('submissionId', $document->submissionUUID)->first();

            $type                   = $this->getStatusType($document->status);
            $document->status_color = $type['color'];
            $document->status_text  = $type['text'];
            $document->status_icon  = $type['icon'];
        });

        $invoices = paginate($documents, 20);

        return view('einvoice::documents.index', compact('invoices', 'statuses'));
    }

    /**
     * Get the status type of the document.
     *
     * @param $status
     * @return array
     */
    protected function getStatusType($statusType)
    {
        $data = [
            'Valid' => [
                'color' => 'text-success',
                'text'  => __('einvoice.Valid'),
                'icon'  => 'fas fa-check-circle'
            ],
            'Invalid' => [
                'color' => 'text-danger',
                'text'  => __('einvoice.Invalid'),
                'icon'  => 'fas fa-exclamation-circle'
            ],
            'Rejected' => [
                'color' => 'text-warning',
                'text'  => __('einvoice.Rejected'),
                'icon'  => 'fas fa-ban'
            ],
            'Cancelled' => [
                'color' => 'text-orange',
                'text'  => __('einvoice.Cancelled'),
                'icon'  => 'fas fa-times-circle'
            ],
            'Submitted' => [
                'color' => 'text-purpel',
                'text'  => __('einvoice.Submitted'),
                'icon'  => 'fas fa-sync'
            ],
        ];

        return $data[$statusType] ?? [
            'color' => '',
            'text'  => '',
            'icon'  => ''
        ];
    }

    /**
     * Show the specified resource.
     * @return view
     */
    public function show($documentUUID)
    {
        $document = Einvoices::where('einvoices.uuid', $documentUUID)
            ->with(['invoice', 'details'])
            ->first();

        if (!is_null($document->validationResults)) {
            $document->validationResults = json_decode($document->validationResults, true);
            $document->validationErrors  = !is_null($document->validationResults)
                ? collect($document->validationResults['validationSteps'])->map(function ($validationError) {
                    return !is_null($validationError['error']) ? $validationError['error']['innerError'] : null;
                })->filter()->pluck(app()->getLocale() === 'ar' ? '*.errorAr' : '*.error')->flatten(1)->unique()->toArray()
                : NULL;
        }

        $type                   = $this->getStatusType($document->status);
        $document->status_color = $type['color'];
        $document->status_text  = $type['text'];
        $document->status_icon  = $type['icon'];

        // Add issuer role to the document.
        $document->issuer_role_txt = $document->issuerType == 0 && in_array($document->issuer_role ?? 1, [1, 2])
            ? __('einvoice.supplier')
            : __('einvoice.client');

        // Add receiver to the document.
        $document->receiver_role_txt = $document->receiverType == 0 && in_array($document->receiver_role ?? 1, [1, 2])
            ? __('einvoice.supplier')
            : __('einvoice.client');

        return view('einvoice::documents.show', compact('document'));
    }

    /**
     * Cancel invoice.
     *
     * @param $documentUUID
     * @return JsonResponse
     */
    public function cancel($documentUUID, Request $request)
    {
        $document  = Einvoice::cancelDocument($documentUUID, $request->reason);
        $message   = !is_null($document['message']) ? $document['message'] : $document['details'];

        return response()->json(['type' => 'cancel', 'message' => $message, 'code' => $document['code']]);
    }

    /**
     * Reject Invoice.
     *
     * @param $documentUUID
     * @param Request $request
     * @return JsonResponse
     */
    public function reject($documentUUID, Request $request)
    {
        $document  = Einvoice::rejectDocument($documentUUID, $request->reason);
        $message   = !is_null($document['message']) ? $document['message'] : $document['details'];

        return response()->json(['type' => 'reject', 'message' => $message, 'code' => $document['code']]);
    }

    /**
     * Decline Cancel Invoice.
     *
     * @param $documentUUID
     * @param Request $request
     * @return JsonResponse
     */
    public function declineCancel($documentUUID, Request $request)
    {
        $document  = Einvoice::declineCancelDocument($documentUUID, $request->reason);
        $message   = !is_null($document['message']) ? $document['message'] : $document['details'];

        return response()->json(['type' => 'declineCancel', 'message' => $message, 'code' => $document['code']]);
    }

    /**
     * Decline Reject Invoice.
     *
     * @param $documentUUID
     * @param Request $request
     * @return JsonResponse
     */
    public function declineReject($documentUUID, Request $request)
    {
        $document  = Einvoice::declineRejectDocument($documentUUID, $request->reason);
        $message   = !is_null($document['message']) ? $document['message'] : $document['details'];

        return response()->json(['type' => 'declineReject', 'message' => $message, 'code' => $document['code']]);
    }

    /**
     * Download Invoice As PDF.
     *
     * @param $documentUUID
     * @return BinaryFileResponse $PDF File
     */
    public function download($documentUUID)
    {
        $pdf = storage_path('app/modules/einvoice/documents/' . $documentUUID . '.pdf');

        if (!file_exists($pdf)) {
            $einvoice = new Einvoice();
            Einvoice::getDocumentPDF($documentUUID);
        }

        return response()->file($pdf, [
            'Content-Type'        => 'application/pdf',
            'Content-Disposition' => 'attachment; filename="' . $documentUUID . '.pdf"',
        ]);
    }

    /**
     * Update E-Invoice status.
     *
     * @param Request $request
     * @return JsonResponse
     */
    public function updateStatus(Request $request): JsonResponse
    {
        $einvoice = Einvoices::where('status', 'Submitted')->where('uuid', $request->uuid)->first();

        if ($einvoice) {
            $documentData = Einvoice::getDocumentDetails($request->uuid);

            if (
                $documentData && isset($documentData->original) && is_array($documentData->original) && isset($documentData->original['status'])
                && isset($documentData->original['validationResults']) && !is_null($documentData->original['validationResults'])
            ) {
                $state                       = $documentData->original['status'];
                $einvoice->status            = $state;
                $einvoice->validationResults = json_encode($documentData->original['validationResults']);
                $update_status               = $einvoice->save();

                if ($update_status) {
                    $type         = $this->getStatusType($state);
                    $status_color = $type['color'];
                    $status_text  = $type['text'];
                    $status_icon  = $type['icon'];
                    $txt          = '<td class="' . $status_color . ' ' . $state . '" data-uuid="' . $request->uuid . '"><i class="' . $status_icon . '"></i>' . $status_text . '</td>';

                    return response()->json(['status' => true, 'msg' => $txt]);
                }
                return response()->json(['status' => false, 'msg' => __('einvoice.load_status_failed')]);
            }
            return response()->json(['status' => false, 'msg' => __('extends.load_error')]);
        }

        return response()->json(['status' => false, 'msg' => __('einvoice.load_status_failed')]);
    }
}
