<?php

namespace Modules\EInvoice\Http\Controllers;

use App\Services\Einvoice\Einvoice;
use Illuminate\Http\Request;
use Illuminate\Http\Response;
use Illuminate\Routing\Controller;
use Illuminate\View\View;

class PackageRequestController extends Controller
{
    /**
     * Get the list of document package requests.
     *
     * @param Request $request
     * @return View
     */
    public function index(Request $request): View
    {
        $pageSize = 20;
        $result = Einvoice::getPackageRequests($pageSize, $request->page)->original;
        $packages = isset($result['result']) ? paginate($result['result'], $pageSize, $request->page) : collect();

        return view('einvoice::packageRequests.index', compact('packages'));
    }

    /**
     * Add Document Package Request.
     *
     * @param Request $request
     * @return View
     */
    public function add(Request $request)
    {
        $result = Einvoice::getDocumentTypes();
        $documentTypes = $result->status() == 200 ? $result->original['result'] : collect();

        if ($request->isMethod('post')) {

            $data = [
                "type" => $request->type,
                "format" => $request->input('format'),
                "queryParameters" => [
                    "dateFrom" => date('Y-m-d\TH:i:s\Z', strtotime($request->dateFrom)),
                    "dateTo" => date('Y-m-d\TH:i:s\Z', strtotime($request->dateTo))
                ]
            ];

            /**
             * Add the document type name to the data queryParameters.
             * $request->documentTypeNames must be an array.
             */
            if ($request->has('documentTypeNames')) {
                $documentTypeNames = (gettype($request->documentTypeNames) == "array") ? $request->documentTypeNames : [$request->documentTypeNames];
                $data['queryParameters']['documentTypeNames'] = $documentTypeNames;
            }

            /**
             * Add the statuses to the data queryParameters.
             * $request->statuses must be an array.
             */
            if ($request->has('statuses')) {
                $statuses = (gettype($request->statuses) == "array") ? $request->statuses : [$request->statuses];
                $data['queryParameters']['statuses'] = $statuses;
            }

            /**
             * Add the productsInternalCodes to the data queryParameters.
             * $request->productsInternalCodes must be an array.
             */
            if ($request->has('productsInternalCodes')) {
                $productsInternalCodes = (gettype($request->productsInternalCodes) == "array") ? $request->productsInternalCodes : [$request->productsInternalCodes];
                $data['queryParameters']['productsInternalCodes'] = $productsInternalCodes;
            }

            /**
             * Add receiverSenderType to the data queryParameters.
             */
            if ($request->has('receiverSenderType')) {
                $data['queryParameters']['receiverSenderType'] = $request->receiverSenderType;
            }

            /**
             * Add the receiverSenderId to the data queryParameters.
             */
            if ($request->has('receiverSenderId')) {
                $data['queryParameters']['receiverSenderId'] = $request->receiverSenderId;
            }

            /**
             * Add the branchNumber to the data queryParameters.
             */
            if ($request->has('branchNumber')) {
                $data['queryParameters']['branchNumber'] = $request->branchNumber;
            }

            /**
             * Add the itemCodes to the data queryParameters.
             * $request->itemCodes must be an array.
             */
            if ($request->has('itemCodes')) {
                foreach ($request->itemCodes as $itemCode) {
                    $itemCode = json_decode($itemCode, true);
                    $data['queryParameters']['itemCodes'][] = $itemCode[0];
                }
            }

            $result = Einvoice::requestDocumentPackage($data);

            if (isset($result->original['error'])) {
                $key = 'einvoiceErrors';
                $message = $result->original['error']['details'];
            } else {
                $key = 'error';
                $message = trans('einvoice.send_failed');
            }

            return $result->status() == 200
                ? back()->with('success', trans('einvoice.send_successfully'))
                : back()->with($key, $message);
        }

        return view('einvoice::packageRequests.add', compact('documentTypes'));
    }

    /**
     * Download Document Package Request.
     *
     * @param $id
     * @return Response
     */
    public function download($id): Response
    {
        $package = Einvoice::getDocumentPackage($id);
        if ($package->status() === 200) {
            return response()->make($package->original, 200, ['Content-Type' => 'application/zip']);
        }

        return response(['status' => false], 500);
    }
}
