<?php

namespace Modules\Invoices\Entities;

use App\Scopes\ShopIdScope;
use App\Services\Invoice;
use App\Models\User;
use App\Services\Updaters\InvoiceClient;
use App\Services\Updaters\InvoiceSafe;
use App\Traits\RoundingAttributesTrait;
use Illuminate\Database\Eloquent\Model;
use Modules\Invoices\Entities\Client;
use Modules\Invoices\Entities\InvoiceDetails;
use Modules\Invoices\Entities\InvoiceDetailsVats;
use Modules\Invoices\Entities\Einvoices;
use Carbon\Carbon;
use Modules\ClientInstallments\Entities\Instalment;

class Invoices extends Model
{
    use RoundingAttributesTrait;
    protected $guarded = ['id'];
    protected $table   = 'invoices_';
//    protected $appends = ['einvoice_stage', 'einvoice_count'];
    protected $appends = ['einvoice_stage'];
    public $timestamps = false;
    // public $incrementing = true;

    protected static $available_types = ['sales', 'purchases', 'returned sales', 'returned purchases', 'sales offers'];
    private static $available_types_no = [
        'purchases' => 0,
        'sales' => 1,
        'returned purchases' => 2,
        'returned sales' => 3,
        'sales offers' => 23,
    ];

    public static function getAvailableTypes()
    {
        return self::$available_types;
    }

    public static function getInvoiceNo($invoice_type)
    {
        return self::$available_types_no[$invoice_type];
    }

    public function collector()
    {
        return $this->belongsTo(User::class, 'collector_id', 'id')->select('id', 'name', 'collecting_ratio')->where(['collecting' => true, 'is_employee' => true]);
    }

    public function shop()
    {
        return $this->belongsTo(BadrShop::class, 'shop_id', 'serial_id');
    }

    public function client()
    {
        return $this->belongsTo(Client::class, 'client_id')->withDefault(['client_tax_number' => null, 'client_name' => '---', 'mobile1' => null, 'email' => null, 'address' => null]);
    }

    public function User()
    {
        return $this->belongsTo(User::class, 'add_user');
    }

    public function issuer()
    {
        return $this->belongsTo(Client::class, 'add_user');
    }

    public function vats()
    {
        return $this->hasMany(InvoiceDetailsVats::class, 'invoice_id');
    }

    public function InvoiceDetails()
    {
        return $this->hasMany(InvoiceDetails::class, 'invoice_id')
                    ->select('item_id', 'unit_id', 'invoice_id', 'quantity', 'price_without_vat', 'item_price_with_discount', 'total_price', 'basic_pay_price', 'pay_price', 'sale_price', 'basic_sale_price', 'discount', 'discount_type', 'discount_operation', 'item_about');
    }

    public function einvoice()
    {
        return $this->hasOne(Einvoices::class, 'invoice_id');
    }

    public function einvoice_stages()
    {
        return $this->hasMany(EinvoiceStage::class, 'invoice_id');
    }

    public function getEinvoiceStageAttribute()
    {
        return $this->einvoice_stages()->whereIn('status', [0, 1, 2])->count();
    }

    public function employee()
    {
        return User::find($this->user_id)->user_name ?? '-';
    }

    public function installments()
    {
        return $this->hasMany(Instalment::class, 'bill_id');
    }

    public function scopeOfType($query, $type)
    {
        return $query -> where('invoice_type' , self::$available_types_no[$type]);
    }

    public function setType()
    {
        $this->type = $this->invoice_type == 0
                        ? trans('invoices.type.purchase')
                        : ($this->invoice_type == 1
                            ? trans('invoices.type.sales')
                            : ($this->invoice_type == 2
                                ? trans('invoices.type.purchase_back')
                                : ($this->invoice_type == 3 ? trans('invoices.type.sales_back') : trans('invoices.type.sales_offers'))));
    }

    public function getPaymentType(bool $check = false)
    {
        if (! $check) return false;
        $safe_point_ids = ClientTransaction::select('safe_point_id')->where(['bill_id' => $this->id, 'shop_id' => $this->shop_id])->pluck('safe_point_id')->toArray();
        $sale_point = SalePoint::select('point_type')->where(['shop_id' => $this->shop_id, 'point_type' => 2])->whereIn('id', $safe_point_ids)->first() ?? null;
        return $sale_point ? 'دفع شبكة' : 'دفع نقدي';
    }

    public static function createNew(Invoice $invoice_obj, InvoiceClient $client, InvoiceSafe $safe)
    {
        $request = $invoice_obj->getRequest();

        $new_invoice = Invoices::create([
            'invoice_date' => $request->invoice_date,
            'transaction_date' => date('Y-m-d h:i:s'),
            'bill_total' => $request->invoice_total,
            'add_user' => auth()->user()->id,
            'client_id' => $request->client_id,
            'collector_id' => Client::select('collector_id')->where('id', $request->client_id)->where('client_type', '!=', false)->first()?->collector_id ?? null,
            'net_price' => $request->net,
            'payment' => $request->paid,
            'bill_payment' => $request->paid,
            'rest' => floatVal($request->net) - floatVal($request->paid),
            'store_id' => $request->store_id,
            'bill_no' => $request->bill_no ?? 0,
            'notes' => $request->notes,
            'shop_id' => get_session_val('serial_id', 'Integer'),
            'balance' => ($client->getNewBalance()),
            'currency_id' => $safe->getSafePoint()->currency_id,
            'currency_value' => $safe->getCurrencyValue(),
            'invoice_type' => $invoice_obj->getType(),
            'invoice_total_disc' => 0,
            'invoice_total_disc_type' => 1,
            'invoice_total_disc_status' => get_session_val('company_ratio', 'integer') == 2 ? 0 : 1,
            'insatllments_add'  => $request->installment_add_precent ?? 0,
            'supplier_invoice_no'  => $request->supplier_invoice_no ?? null,
            'reference_id'  => $request->reference_id ?? 0,
            'fee' => $request->delivery_fees,
            'fee_status' => false,
            'bill_service_id' => $invoice_obj->explodeValue($request->service, 0),
            'bill_service_value' => $invoice_obj->explodeValue($request->service, 1),
            'delivery_option_id' => $invoice_obj->explodeValue($request->period, 0),
            'delivery_period_id' => $invoice_obj->explodeValue($request->period, 0),
            'delivery_option_value' => $invoice_obj->explodeValue($request->period, 1),
            'delivery_man_id' => $request->delivery_man ?? 0,
        ]);

        return $new_invoice;
    }



    public function Store()
    {
        return $this->belongsTo(Store::class, 'store_id');
    }

    public function childInvoiceDetails()
    {
        return $this->hasManyThrough(InvoiceDetails::class, Invoices::class, 'reference_id', 'invoice_id')
                ->selectRaw('item_id , SUM(quantity) AS total_qty')->groupBy('item_id');
    }

    public function backInvoice()
    {
        return $this->hasMany(self::class, 'reference_id', 'id');
    }
    
    public function checkAllQuantityIsBack()
    {
        return $this->InvoiceDetails->sum('quantity') <= $this->childInvoiceDetails->sum('total_qty');
    }

    // one day , not now 😉
    // protected static function booted()
    // {
    //     static::addGlobalScope(new ShopIdScope);
    // }

    public function scopeBetweenTwoDates($query, $dates)
    {
        return $query->where(function($query) use($dates) {
                $query->whereDate( 'invoices_.transaction_date', '>=', $dates['start'] )
                        ->whereDate( 'invoices_.transaction_date', '<=', $dates['end'] );
        });
    }

    public function scopeByClient($query, $client_id)
    {
        return  $query->where('client_id',$client_id);
    }

    public function scopeHaveRest($query)
    {
        return $query -> where('rest' , '>' , 0);
    }

    // public function scopeInstalled($query)
    // {
    //     $installments = Instalment::pluck('bill_id');
    //     return $query->whereIn('id',$installments);
    // }
}
