<?php

namespace Modules\Invoices\Http\Controllers;

use App\helpers\ShortUrl;
use App\Http\Controllers\Controller;
use App\Mail\SendInvoiceToClient;
use App\Models\User;
use Exception;
use Illuminate\Support\Facades\Mail;
use Modules\Invoices\Entities\{BadrShop, Client, Invoices, InvoiceSettings, Store};
use Salla\ZATCA\GenerateQrCode;
use Salla\ZATCA\Tags\{InvoiceDate, InvoiceTaxAmount, InvoiceTotalAmount, Seller, TaxNumber};

class ShareInvoiceController extends Controller
{
    public function show($hash_id)
    {
        $data = self::getInvoice($hash_id);

        if (!is_array($data)) abort(404);
        $view = auth()->user()?->invoice_print_template ? 'tax_bill' : 'tax_bill';
        return view("invoices::print.$view", $data);
    }

    public static function mail($hash_id, $transaction = null): bool
    {
        try {
            $data = self::getInvoice($hash_id, transaction:$transaction);

            if (!is_array($data)) return false;
            if(! $data['client']) return  false;
            if ($data['client']->email && $data['client']->send_email) {
                Mail::send(new SendInvoiceToClient($data));
                return true;
            }
        } catch(Exception $e) { return false; }
        return false;
    }

    public static function whatsApp(int $client, string $route, bool $forceSend = true)
    {
        $client = Client::select('mobile1', 'send_whatsapp')->where('shop_id', auth()->user()->shop_id)->when(!$forceSend, fn($query) => $query->where('send_whatsapp', true))->find($client) ?? new Client();

        if (!$client->mobile1) {
            return ['status' => false, 'message' => 'لا يمكن الارسال لعدم توفر رقم واتس اب'];
        }

        if (!$client->send_whatsapp && !$forceSend) {
            return ['status' => false];
        }

        $route = ShortUrl::shorten( $route );
        $prefix = '';
        if (stripos('+2', $client->mobile1) === false) {
            $prefix = '+2';    
        }
        return [
            'status' => true,
            'redirect' => "https://api.whatsapp.com/send?phone={$prefix}{$client->mobile1}&text=$route"
        ];
    }

    protected static function setKeys(?Invoices $invoice, $transaction)
    {
        return [
            'shop_id'       => $invoice->shop_id ?? $transaction->shop_id ?? auth()->user()?->shop_id,
            'client_id'     => $invoice->client_id ?? $transaction->client_id ?? null,
            'collector_id'  => $invoice->collector_id ?? $transaction->collector_id ?? null,
            'store_id'      => $invoice->store_id ?? $transaction->store_id ?? null,
            'user_id'       => $invoice->add_user ?? $transaction->user_id ?? null,
        ];
    }

    public static function getInvoice($hash_id, bool $check_invoice = true, $transaction = null)
    {
        try {
            $decrypt_id = decrypt($hash_id);
            if ($transaction) {
                self::setGuestAuth($transaction->user_id);
            }

            $invoice = Invoices::select('id', 'transaction_date', 'notes', 'balance', 'invoice_total_disc_status', 'invoice_total_disc_status', 'invoice_type', 'rest', 'payment', 'collector_id', 'bill_no', 'invoice_date', 'bill_total', 'net_price', 'payment', 'rest', 'fee', 'store_id', 'shop_id', 'client_id', 'add_user', 'invoice_total_disc_type', 'invoice_total_disc')
                                    ->with('InvoiceDetails.Unit', 'InvoiceDetails.Item', 'vats')
                                    ->withSum('vats', 'vat_weight')->find($decrypt_id) ?? new Invoices();

            self::setGuestAuth($invoice->add_user);

            if (! $invoice && $check_invoice) throw new Exception("Invoice Not Found");
            extract(self::setKeys($invoice, $transaction));

            $client           = self::getClient($client_id, $shop_id);
            $user             = self::getUser($user_id, $shop_id);
            $collector        = self::getCollector($collector_id, $shop_id);
            $store            = self::getStore($store_id, $shop_id);
            $shop             = self::getShop($shop_id);
            $invoice_settings = self::getInvoiceSetting($shop_id);
            $settings = $shop->printing_shop_info_src ? $store : $invoice_settings;
            $invoice->setType();
            $invoice->vat_number = $invoice_settings->vat_number;

            try {
                $invoice->qr = self::getQrCode($invoice, $shop->shop_name);
            } catch (\Exception $e){
                $invoice->qr = '';
            }

            return [
                'invoice'           => $invoice,
                'client'            => $client,
                'collector'         => $collector,
                'user'              => $user,
                'shop'              => $shop,
                'settings'          => $settings,
                'invoice_settings'  => $invoice_settings,
                'store'             => $store
            ];
        } catch (Exception $e) {
            dd($e);
            return false;
            return $e->getMessage();
        }
    }

    protected static function setGuestAuth(?int $id): void
    {
        if (! auth()->check()) {
            $auth = \App\Models\User::select('id', 'shop_id', 'invoice_print_template')->find($id);
            if ($auth) auth()->setUser($auth);
        }

        if (! auth()->check()) {
            throw new Exception("لا يمكن عرض هذه البيانات");
        }
    }

    protected static function getClient(?int $id, ?int $shop_id): Client
    {
        return Client::select('id', 'client_name', 'code', 'balance', 'mobile1', 'tele', 'email', 'client_tax_number', 'address', 'send_email', 'client_type')->with('type')->where('id', $id)->where('shop_id', $shop_id)->first() ?? new Client();
    }

    protected static function getCollector(?int $id, ?int $shop_id): User
    {
        return User::select('id', 'name', 'collecting_ratio')->where(['collecting' => true, 'is_employee' => true])->where('id', $id)->where('shop_id', $shop_id)->first() ?? new User();
    }

    protected static function getUser(?int $id, ?int $shop_id): User
    {
        return User::select('id', 'user_name', 'invoice_print_template')->where(['id' => $id, 'shop_id' => $shop_id])->first() ?? new User();
    }

    protected static function getStore(?int $id, ?int $shop_id): Store
    {
        return Store::select('message_bill', 'bill_header', 'tax_number', 'address', 'store_name')->where(['id' => $id, 'shop_id' => $shop_id])->first() ?? new Store();
    }

    protected static function getShop(?int $shop_id): BadrShop
    {
        return BadrShop::select('serial_id','telephone', 'tel_bill','use_parts_cash_proforma_invoice','shop_name','company_ratio','shop_name_en','printing_shop_info_src','address','address_en','logo_path','apply_vats','show_delivery_charges_in_the_invoice','currency','decimal_num_price','country')
                            ->where(['serial_id' => $shop_id])->with('getCountry')->first() ?? new BadrShop();
    }

    protected static function getInvoiceSetting(?int $shop_id): InvoiceSettings
    {
        return InvoiceSettings::where(['shop_id' => $shop_id])->first() ?? new InvoiceSettings();
    }

    public static function getQrCode(Invoices $invoice, string $shop_name)
    {
        try {
            $qrcode = GenerateQrCode::fromArray([
                new Seller($shop_name), // seller name
                new TaxNumber($invoice->vat_number), // seller tax number
                new InvoiceDate($invoice->transaction_date), // invoice date as Zulu ISO8601 @see https://en.wikipedia.org/wiki/ISO_8601
                new InvoiceTotalAmount( round_money($invoice->net_price) ), // invoice total amount
                new InvoiceTaxAmount( round_money($invoice->vats_sum_vat_weight) ) // invoice tax amount
            ])->render();

            return view('QR-code', compact('qrcode'))->render();
        } catch (Exception $e) {
            return '';
        }
    }
}
