<?php

namespace Modules\Item\Http\Controllers;

use Illuminate\Contracts\Support\Renderable;
use Illuminate\Http\Request;
use Illuminate\Routing\Controller;
use Illuminate\Support\Carbon;
use Illuminate\Support\Facades\Response;
use Illuminate\Support\Facades\Validator;
use Illuminate\Validation\Rule;
use Modules\Item\Entities\PriceList;

class MultiPriceController extends Controller
{
    /**
     * Display a listing of the resource.
     * @return Renderable
     */
    public function index()
    {
        $prices_list = PriceList::where('shop_id', get_session_val('serial_id', 'integer'))->get();
        return view('item::multiPrices.index', compact('prices_list'));
    }

    /**
     * Store a newly created resource in storage.
     * @param Request $request
     * @return Renderable
     */
    public function store(Request $request)
    {

        $validator = Validator::make($request->all(), [
            'name'  => ['required', 'min:3', Rule::unique('prices_list', 'list_name')->where('shop_id', get_session_val('serial_id', 'integer'))]
        ]);

        if ($validator->fails()) {
            return Response()->json([
                'status'    => false,
                'errors'    => $this->getValidationFailsErrors($validator),
                'target'    => 'store'
            ]);
        }

        $new_list  = PriceList::create([
            'list_name' => $request->name,
            'add_user'  => auth()->id(),
            'add_date'  => date('Y-m-d H:i:s'),
            'shop_id'   => get_session_val('serial_id', 'integer')
        ]);

        return Response()->json([
            'status'    => true,
            'new_list'  => $new_list
        ]);
    }

    /**
     * Update the specified resource in storage.
     * @param Request $request
     * @param int $id
     * @return Renderable
     */
    public function update(Request $request)
    {

        $validator = Validator::make($request->all(), [
            'id'    => ['required', Rule::exists('prices_list', 'id')->where('shop_id', get_session_val('serial_id', 'integer'))],
            'name'  => ['required', 'min:3', Rule::unique('prices_list', 'list_name')->where('shop_id', get_session_val('serial_id', 'integer'))]
        ]);

        if ($validator->fails()) {
            return Response()->json([
                'status'    => false,
                'errors'    => $this->getValidationFailsErrors($validator),
                'target'    => 'update'
            ]);
        }

        $updated_list  = PriceList::find($request->id);
        $updated_list->list_name = $request->name;
        $updated_list->edit_user = auth()->id();
        $updated_list->edit_date = Carbon::now();
        $updated_list->save();

        return Response()->json([
            'status'    => true,
            'updated_list'  => $updated_list
        ]);
    }

    /**
     * Remove the specified resource from storage.
     * @param int $id
     * @return Renderable
     */
    public function destroy(Request $request)
    {

        $validator = Validator::make($request->all(), [
            'id'    => ['required', Rule::exists('prices_list', 'id')->where('shop_id', get_session_val('serial_id', 'integer'))],
        ]);

        if ($validator->fails()) {
            return Response()->json([
                'status'    => false,
                'errors'    => $this->getValidationFailsErrors($validator)
            ]);
        }
        $clients_list=PriceList::where('prices_list.shop_id', get_session_val('serial_id', 'integer'))
        ->join('clients', function ($q) use ($request) {
            $q->where('clients.client_type', $request->id);
        })->count();
        $item_list=PriceList::where('prices_list.shop_id', get_session_val('serial_id', 'integer'))
        ->join('items_prices', function ($q) use ($request) {
            $q->where('items_prices.list_id', $request->id);
        })->count();

        if ($clients_list > 0 || $item_list >0) {
            return Response()->json([
                'status'    => false,

            ]);
        } else {
            $new_list  = PriceList::find($request->id);
            $new_list->delete();
            return Response()->json([
                'status'    => true,
                'id'  => $request->id
            ]);
        }
    }

    private function getValidationFailsErrors($validation)
    {
        $errors = $validation->errors();
        $error_data = [];
        foreach ($errors->all() as $error) {
            array_push($error_data, $error);
        }
        return $error_data;
    }
}
