<?php

namespace App\Services\Einvoice;

use Illuminate\Support\Facades\Log;
use Ixudra\Curl\Facades\Curl;
use Modules\EInvoice\Entities\EinvoiceStage;
use Modules\EInvoice\Entities\Settings;

abstract class Basic
{
    public static string $idSrvBaseUrl = 'https://id.preprod.eta.gov.eg';
    public static string $apiBaseUrl   = 'https://api.preprod.invoicing.eta.gov.eg';
    public string $token;
    public int $statusCode;
    public string $clientId;
    public string $clientSecret;

    public function __construct()
    {
        $shop               = Settings::shop()->first();
        abort_if(! $shop, 404, 'Shop not found');
        $this->clientId     = $shop->client_id;
        $this->clientSecret = $shop->client_secret;
    }

    /**
     * This Method allows taxpayer system to get access token to be used to access other protected APIs of the solution.
     *
     * @return string token
     */
    abstract public function loginTaxpayer(): string;

    /**
     * This Method allows to send any request to the taxpayer system.
     *
     * @param array $data
     * @return object
     */
    protected function sendCurl(array $data): object
    {
        $response = Curl::to($data['url'])->withContentType($data['contentType'] ?? 'application/json')->returnResponseObject();

        if (isset($data['body'])) {
            $response->withData($data['body']);
        }

        if (isset($data['headers'])) {
            (count($data['headers']) > 1)
                ? $response->withHeaders($data['headers'])
                : $response->withHeader(array_key_first($data['headers']) , $data['headers'][array_key_first($data['headers'])]);
        }

        if (isset($data['asJson'])) {
            $response->asJson(true);
        }

        if (isset($data['token'])) {
            $response->withBearer($data['token']);
        }

        if (isset($data['connectTimeout'])){
            $response->withConnectTimeout($data['connectTimeout']);
        }

        if (isset($data['timeout'])){
            $response->withTimeout($data['timeout']);
        }

        return isset($data['download'])
            ? $response->download($data['download'])
            : $response->{$data['method'] ?? 'post'}();
    }

    /**
     * This Method is used to log the stages of the einvoice in einvoice stage table.
     * @param string $type
     * @param $data
     * @param int|null $stage
     * @param int|null $status
     * @param $request
     * @param $response
     * @param string $channel
     * @return void
     *
     * type:
     * 1 => invoice
     * 2 => item code
     *
     * status:
     * 0 => initializing
     * 1 => pending
     * 2 => error
     * 3 => done
     *
     * stages:
     * 0  => submit einvoice
     * 1  => update invoice table (invoices_)
     * 2  => get einvoice
     * 3  => get einvoice details
     * 4  => save einvoice in db
     * 5  => save einvoice details in db
     * 6  => save einvoice detail in vats in db
     * 7  => save item code in einvoice solution
     * 8  => update items table with item code
     * 9  => save item code in db
     */
    public function logEinvoiceStage(string $type, $data, int $stage = null, int $status = null, $request = null, $response = null, string $channel = 'invoice'): void
    {
        try {
            $column             = $type === 'invoice' ? 'invoice_id' : 'item_code';
            $query              = EinvoiceStage::where($column, $data);
            $einvoiceStageCount = $query->count();
            $einvoiceStage      = $query->where('stage', $stage);
            $dataText           = is_array($data) ? implode(', ', $data) : $data;
            $logText            = "Intialize submit stages in einvoice stage table (Intialize stages for $type: $dataText)";

            if (is_null($stage)) {
                if ($type === 'invoice' && !is_null($data)) {
                    $stages = [
                        ['invoice_id' => $data, 'stage' => 0],
                        ['invoice_id' => $data, 'stage' => 1],
                        ['invoice_id' => $data, 'stage' => 2],
                        ['invoice_id' => $data, 'stage' => 3],
                        ['invoice_id' => $data, 'stage' => 4],
                        ['invoice_id' => $data, 'stage' => 5],
                        ['invoice_id' => $data, 'stage' => 6],
                    ];
                } else if ($type === 'item_code' && !is_null($data)) {
                    $stages = [
                        ['item_code' => $data, 'stage' => 7],
                        ['item_code' => $data, 'stage' => 8],
                        ['item_code' => $data, 'stage' => 9],
                    ];
                }
                Log::channel($channel)->info($logText);
                if ($einvoiceStageCount === 0) {
                    EinvoiceStage::insert($stages);
                }
            } else {
                $data = ['status' => $status, 'updated_at' => now(), 'shop_id' => auth()->user()->shop_id];

                if (!empty($request)) {
                    $data['request'] = $request;
                    $data['request_at'] = now();
                }

                if (!empty($response)) {
                    $data['response'] = $response;
                    $data['response_at'] = now();
                }

                // Check if the stage is not done
                $einvoiceStage = ($einvoiceStage->first()->status !== 3) ? $einvoiceStage->update($data) : true;

                $einvoiceStage
                    ? Log::channel($channel)->info("Einvoice Stage Updated Successfully , Stage : $stage , Status : $status")
                    : Log::channel($channel)->error("Error in updating einvoice stage , Stage : $stage , Status : $status");
            }
        } catch (\Exception $e) {
            Log::channel($channel)->error('Error in logEinvoiceStage method: ' . $e->getMessage() . ' , Line Number: ' . $e->getLine());
        }
    }
}
