<?php

namespace App\Services\Einvoice\Interfaces;

use Illuminate\Http\Response;

interface Document
{
    /**
     * Send E-Invoice to the Taxpayer
     *
     * @param $body
     * @return array
     */
    public function sendEinvoice($body): array;

    /**
     * This Method allows taxpayers to cancel previously issued document.
     *
     * @param string $documentUUID
     * @param string $reason
     * @return Response $data
     *
     * This Method Can Cancel (Invoices, Credit Notes, Debit Notes).
     */
    public function cancelDocument(string $documentUUID, string $reason): Response;

    /**
     * This Method allows taxpayer that received an invalid invoice to reject it.
     *
     * @param string $documentUUID
     * @param string $reason
     * @return Response $data
     *
     * This Method Can Reject (Invoices, Credit Notes, Debit Notes).
     */
    public function rejectDocument(string $documentUUID, string $reason): Response;

    /**
     * This Method allows taxpayers "recipient" to decline document that was previously cancelled by other taxpayer "issuer".
     *
     * @param string $documentUUID
     * @return Response $data
     */
    public function declineCancelDocument(string $documentUUID): Response;

    /**
     * This Method allows taxpayers "Issuer" to decline document that was previously rejected by other taxpayer "Receiver".
     *
     * @param string $documentUUID
     * @return Response $data
     */
    public function declineRejectDocument(string $documentUUID): Response;

    /**
     * This Method allows taxpayer systems to retrieve list of document types published by the eInvoicing solution.
     *
     * @return Response $data
     */
    public function getDocumentTypes(): Response;

    /**
     * This Method allows taxpayer systems to query latest documents sent or received that are available on the eInvoicing solution.
     * @param int $pageSize
     * @param int $pageNumber
     * @return Response $data
     */
    public function getRecentDocuments(int $pageSize = 20, int $pageNumber = 1): Response;

    /**
     * This Method returns information on documents submitted during a single submission by taxpayer.
     *
     * @param string $submitionUUID
     * @param int $pageSize
     * @param int $pageNumber
     * @return Response $data
     */
    public function getSubmittedDocument(string $submitionUUID, int $pageSize = 10, int $pageNumber = 1): Response;

    /**
     * This Method allows taxpayers to retrieve document.
     *
     * @param string $documentUUID
     * @return document
     */
    public function getDocument(string $documentUUID);

    /**
     * This Method enables ERP system to download the PDF version of the document having predefined layout.
     *
     * @param string $documentUUID
     * @return document as PDF
     */
    public function getDocumentPDF(string $documentUUID);

    /**
     * This Method allows taxpayers to retrieve a single document full details including validation results.
     *
     * @param string $documentUUID
     * @return Response $data
     */
    public function getDocumentDetails(string $documentUUID): Response;
}
