<?php

namespace App\Services\Einvoice\Invoice;

use App\Rules\NotNull;
use App\Services\Einvoice\Basic;
use App\Services\Einvoice\Interfaces\InitiateDocument;
use Illuminate\Support\Facades\{Log, Validator};
use Illuminate\Support\Str;
use Modules\EInvoice\Entities\{Clients, Invoice, Item, ItemUnit, Settings, Unit, Vat};
use Illuminate\Support\Collection;

class Initiate extends Basic implements InitiateDocument
{
    public Collection $documents;
    public Collection $failedBills;
    public Collection $notFoundBills;

    public function __construct()
    {
        parent::__construct();
        $this->documents = collect();
        $this->failedBills = collect();
        $this->notFoundBills = collect();
    }

    /**
     * This Method allows taxpayer system to get access token to be used to access other protected APIs of the solution.
     *
     * @return string token
     */
    public function loginTaxpayer(): string
    {
        try {
            $curlData = [
                'url'         => self::$idSrvBaseUrl . '/connect/token',
                'contentType' => 'application/x-www-form-urlencoded',
                'body'        => [
                    'grant_type'    => 'client_credentials',
                    'client_id'     => $this->clientId,
                    'client_secret' => $this->clientSecret,
                    'scope'         => 'InvoicingAPI',
                ],
            ];

            // Send Invoice to e-invoice.
            $response = $this->sendCurl($curlData);

            // If the response is successful data is returned as associative array
            $data = json_decode($response->content, true);
            $this->statusCode = $response->status;

            return ($this->statusCode >= 200 && $this->statusCode < 300)
                        ? $data['access_token']
                        : $data['error'];

        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in loginTaxpayer Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method to set Document to be submitted.
     *
     * @param mixed $bill
     * @param string $cert_token
     * @return Initiate
     */
    public function setDocument(mixed $bill, string $cert_token): Initiate
    {
        // All the data used to create the document
        $invoice = Invoice::with('reciever', 'vats')->where('id', $bill)->first();

        // if the invoice is not found
        if (!$invoice) {
            $this->notFoundBills->push($bill);
            return $this;
        }

        $settings = Settings::shop()->first();
        $buyer    = $invoice->reciever;
//        $vats     = $invoice->vats;

        $vats           = collect($invoice->vats->toArray())->map(function ($vat) {
            $vat['add'] = Vat::where('id', $vat['vat_id'])->first()->toArray();
            return $vat;
        });


        // Validate the bill
        $this->validateInvoice($invoice, $settings, $buyer, $vats);

        // Issuer Data.
        $issuer = $this->setIssuer($settings);

        // Buyer Data
        $receiver = $this->setReceiver($buyer, $invoice);

        // Invoice Calculation.
        $invoiceCalculation = $this->invoiceCalculation($invoice, $settings);

        // Invoice Data
        $document = [
            "issuer"                   => $issuer,
            "receiver"                 => $receiver,
            "documentType"             => $invoiceCalculation['documentType'],
            "documentTypeVersion"      => "0.9",
            "dateTimeIssued"           => now()->tz('UTC')->format('Y-m-d\TH:i:s\Z'),
            "taxpayerActivityCode"     => $settings->activity_id,
            "internalID"               => Str::random(12),
            "invoiceLines"             => $invoiceCalculation['invoiceLines'],
            "totalDiscountAmount"      => einvoice_decimal($invoiceCalculation['details']['totalDiscount']),
            "totalSalesAmount"         => einvoice_decimal($invoiceCalculation['details']['totalSalesAmount']),
            "netAmount"                => einvoice_decimal($invoiceCalculation['netAmount']),
            "totalAmount"              => einvoice_decimal($invoiceCalculation['totalAmount']),
            "extraDiscountAmount"      => einvoice_decimal($invoiceCalculation['extraDiscountAmount']),
            "totalItemsDiscountAmount" => einvoice_decimal($invoiceCalculation['details']['totalItemsDiscountAmount']),
            "signatures"               => [
                [
                    "signatureType" => "I",
                    "value"         => $cert_token,
                ],
            ],
        ];

        // If the item has a vat.
        if ($invoiceCalculation['totalTaxes'] > 0) {
            $document['taxTotals'] = $invoiceCalculation['totalTaxes']['taxTotals'];
        }

        // Add the data to the document
        if (!$this->failedBills->has($bill)) {
            $this->documents->put($bill, $document);
        }

        return $this;
    }

    /**
     * This Method to set Issuer to be submitted.
     *
     * @param $settings
     * @return array
     */
    public function setIssuer($settings) : array
    {
        $issuer = [
            'address' => [
                "branchID"              => $settings->branch_id ?? '',
                "country"               => $settings->iso_alpha2 ?? '',
                "governate"             => $settings->governate ?? '',
                "regionCity"            => $settings->regionCity ?? '',
                "street"                => $settings->street ?? '',
                "buildingNumber"        => (string) $settings->buildingNumber ?? '',
                "postalCode"            => (string) $settings->postalCode ?? '',
                "floor"                 => (string) $settings->floor ?? '',
                "room"                  => (string) $settings->room ?? '',
                "landmark"              => $settings->landmark ?? '',
            ],
            'type' => "B",
            'id'   => $settings->registration_id ?? '',
            'name' => $settings->company_name ?? '',
        ];

        // If address has additional data
        if (!is_null($settings->additional_information)) {
            $issuer['address']['additionalInformation'] = $settings->additional_information;
        }

        return $issuer;
    }

    /**
     * This Method to set Receiver to be submitted.
     *
     * @param $buyer
     * @param $invoice
     * @return array
     */
    public function setReceiver($buyer, $invoice): array
    {
        $receiver = [
            'type' => in_array($buyer->client_role, ['1', '2']) ? 'B' : 'P',
            'name' => $buyer->client_name
        ];

        // Push the data to receiver array if person buyer and amount less than threshold limit defined (50000).
        if (!in_array($receiver['type'], ['P', 'F']) || $invoice['net_price'] >= 50000) {

            $receiver['id'] = in_array($receiver['type'], ['P', 'F']) ? $buyer->national_id : $buyer->client_tax_number;

            $receiver['address'] = [
                "country"               => $buyer->country->iso_alpha2 ?? '',
                "governate"             => $buyer->governate ?? '',
                "regionCity"            => $buyer->regionCity ?? '',
                "street"                => $buyer->street ?? '',
                "buildingNumber"        => (string) $buyer->buildingNumber ?? '',
                "postalCode"            => (string) $buyer->postalCode ?? '',
                "floor"                 => (string) $buyer->floor ?? '',
                "room"                  => (string) $buyer->room ?? '',
                "landmark"              => $buyer->landmark ?? '',
            ];

            if (!is_null($buyer->additional_information)) {
                $receiver['address']['additionalInformation'] = $buyer->additional_information;
            }
        }

        return $receiver;
    }

    /**
     * Invoice Items Calculation.
     *
     * @param $invoice
     * @param $settings
     * @return array
     */
    public function invoiceCalculation($invoice, $settings): array
    {
        // Invoice Items.
        $invoiceLines             = collect();
        $taxes                    = collect();
        $totalDiscount            = 0;
        $totalSalesAmount         = 0;
        $extraDiscountAmount      = 0;
        $totalItemsDiscountAmount = 0;
        $documentType             = 'I';

        // Get the items from the invoice.
        $details = $this->invoiceDetails($invoice->details, $invoice, $totalSalesAmount, $totalItemsDiscountAmount, $settings, $totalDiscount, $invoiceLines, $taxes);

        // Total Tax Amount.
        $totalTaxes = $taxes->count() > 0 ? $this->invoiceTaxTotal($taxes) : 0;

        // Sum of all tax amounts.
        $totalTaxAmount = $totalTaxes['totalTaxAmount'] ?? 0  + $details['totalTaxableFees'];

        // Total amount of total sales after applying discount.
        $netAmount = $details['totalSalesAmount'] - $details['totalDiscount'];

        // Total amount of the invoice calculated as NetAmount + Totals of tax amounts.
        $totalAmount = $netAmount + $totalTaxAmount;

        if ($invoice->invoice_type === 3) {
            $documentType = "C";
        }

        return [
            'details'                 => $details,
            'documentType'            => $documentType,
            'totalDiscount'           => $totalDiscount,
            'totalSalesAmount'        => $totalSalesAmount,
            'netAmount'               => $netAmount,
            'totalAmount'             => $totalAmount,
            'extraDiscountAmount'     => $extraDiscountAmount,
            'totalItemsDiscountAmount'=> $totalItemsDiscountAmount,
            'totalTaxes'              => $totalTaxes,
            'totalTaxAmount'          => $totalTaxAmount,
            'invoiceLines'            => $invoiceLines,
            'taxes'                   => $taxes,
        ];
    }

    /**
     * This Method is used to validate the parameters of the invoice.
     *
     * @param $invoice
     * @param $merchant
     * @param $buyer
     * @param $vats
     * @return void
     */
    public function validateInvoice($invoice, $settings, $buyer, $vats): void
    {
        $request  = [
            "merchant" => optional($settings)->toArray() ?? [],
            "buyer"    => optional($buyer)->toArray() ?? [],
            "settings" => optional($settings)->toArray() ?? [],
            "invoice"  => optional($invoice)->toArray() ?? [],
            "vats"     => optional($vats)->toArray() ?? [],
        ];

        $rules = [
            "merchant"                        => ['required', 'array'],
            "merchant.governate"              => ['required', 'string', new NotNull],
            "merchant.regionCity"             => ['required', 'string', new NotNull],
            "merchant.street"                 => ['required', 'string', new NotNull],
            "merchant.buildingNumber"         => ['required', 'numeric', new NotNull],
            "merchant.postalCode"             => ['required', 'numeric', new NotNull],
            "merchant.floor"                  => ['required', 'numeric', new NotNull],
            "merchant.room"                   => ['required', 'numeric', new NotNull],
            "merchant.landmark"               => ['required', 'string', new NotNull],
            "merchant.additional_information" => ['nullable', 'string', new NotNull],
            "buyer"                           => ['required', 'array'],
            "buyer.governate"                 => ['required', 'string', new NotNull],
            "buyer.regionCity"                => ['required', 'string', new NotNull],
            "buyer.street"                    => ['required', 'string', new NotNull],
            "buyer.buildingNumber"            => ['required', 'numeric', new NotNull],
            "buyer.postalCode"                => ['required', 'numeric', new NotNull],
            "buyer.floor"                     => ['required', 'numeric', new NotNull],
            "buyer.room"                      => ['required', 'numeric', new NotNull],
            "buyer.landmark"                  => ['required', 'string', new NotNull],
            "buyer.additional_information"    => ['nullable', 'string', new NotNull],
            "settings"                        => ['required', 'array'],
            "settings.company_name"           => ['required', 'string', new NotNull],
            "settings.activity_id"            => ['required', 'string', new NotNull],
            "settings.branch_id"              => ['required', 'string', new NotNull],
            "settings.country_id"             => ['required', 'string', new NotNull],
            "settings.currency_id"            => ['required', 'string', new NotNull],
            "settings.registration_id"        => ['required', 'string', new NotNull],
            "settings.client_id"              => ['required', 'string', new NotNull],
            "settings.client_secret"          => ['required', 'string', new NotNull]
        ];

        $messages = [
            "merchant.governate.required"           => __('messages.merchant.governate_required'),
            "merchant.regionCity.required"          => __('messages.merchant.regionCity_required'),
            "merchant.street.required"              => __('messages.merchant.street_required'),
            "merchant.buildingNumber.required"      => __('messages.merchant.buildingNumber_required'),
            "merchant.postalCode.required"          => __('messages.merchant.postalCode_required'),
            "merchant.floor.required"               => __('messages.merchant.floor_required'),
            "merchant.room.required"                => __('messages.merchant.room_required'),
            "merchant.landmark.required"            => __('messages.merchant.landmark_required'),
            "buyer.governate.required"              => __('messages.buyer.governate_required'),
            "buyer.regionCity.required"             => __('messages.buyer.regionCity_required'),
            "buyer.street.required"                 => __('messages.buyer.street_required'),
            "buyer.buildingNumber.required"         => __('messages.buyer.buildingNumber_required'),
            "buyer.postalCode.required"             => __('messages.buyer.postalCode_required'),
            "buyer.floor.required"                  => __('messages.buyer.floor_required'),
            "buyer.room.required"                   => __('messages.buyer.room_required'),
            "buyer.landmark.required"               => __('messages.buyer.landmark_required'),
            "settings.company_name.required"        => __('messages.settings.company_name_required'),
            "settings.activity_id.required"         => __('messages.settings.activity_id_required'),
            "settings.branch_id.required"           => __('messages.settings.branch_id_required'),
            "settings.country_id.required"          => __('messages.settings.country_id_required'),
            "settings.currency_id.required"         => __('messages.settings.currency_id_required'),
            "settings.registration_id.required"     => __('messages.settings.registration_id_required'),
            "settings.client_id.required"           => __('messages.settings.client_id_required'),
            "settings.client_secret.required"       => __('messages.settings.client_secret_required'),
        ];

        // Check if vats count is greater than 0 then add the rules.
        if ($vats->count() > 0) {
            $rules["vats"]                         = ['required', 'array'];
            $rules["vats.*.add"]                   = ['required', 'array'];
            $rules["vats.*.vat_id"]                = ['required', 'numeric', new NotNull];
            $rules["vats.*.vat_type"]              = ['required', 'numeric', new NotNull];
            $rules["vats.*.vat_value"]             = ['required', 'numeric', new NotNull];
            $rules["vats.*.add.einvoice_code"]     = ['required', 'string', new NotNull];
            $rules["vats.*.add.sub_einvoice_code"] = ['required', 'string', new NotNull];

            $messages["vats.required"]                         = __('messages.vats.required');
            $messages["vats.*.add.required"]                   = __('messages.vats_add_required');
            $messages["vats.*.vat_id.required"]                = __('messages.vats_vat_id_required');
            $messages["vats.*.vat_type.required"]              = __('messages.vats_vat_type_required');
            $messages["vats.*.vat_value.required"]             = __('messages.vats_vat_value_required');
            $messages["vats.*.add.einvoice_code.required"]     = __('messages.vats_add_einvoice_code_required');
            $messages["vats.*.add.sub_einvoice_code.required"] = __('messages.vats_add_sub_einvoice_code_required');
        }

        $validator = Validator::make($request, $rules, $messages);

        if ($validator->fails()){
            $this->failedBills[$invoice->id] = $validator->errors()->all();
        }
    }

    /**
     * This Method to get the invoice details.
     *
     * @param $invoiceDetails
     * @param $invoice
     * @param $totalSalesAmount
     * @param $totalItemsDiscountAmount
     * @param $settings
     * @param $totalDiscount
     * @param $invoiceLines
     * @param $taxes
     * @return array
     */
    public function invoiceDetails($invoiceDetails, $invoice, $totalSalesAmount, $totalItemsDiscountAmount, $settings, $totalDiscount, $invoiceLines, $taxes): array
    {
        foreach ($invoiceDetails as $item_) {

            // Get the item
            $item = Item::where('id', $item_->item_id)->first();

            /**
             * Get the document type
             * (invoice_type = 0) => فاتورة شراء , (invoice_type = 1) => فاتورة بيع
             * (invoice_type = 2) => فاتورة مرتجع شراء , (invoice_type = 3) => فاتورة مرتجع بيع
             */
            $item_price = in_array($invoice->invoice_type, [0, 2]) ? $item_->basic_pay_price : $item_->basic_sale_price;

            /**
             * Item Discount
             * (discount_type = 0) => نسبة , (discount_type = 1) => مبلغ
             * (discount_operation = 0) => تخفيض , (discount_operation = 1) => زيادة
             */
            $discount         = 0;
            $discount_percent = 0;

            if ($item_->discount > 0) {
                $discount         = (($item_->discount_type === 1) ? $item_->discount : ($item_->discount * $item_price) / 100) * $item_->quantity;
                $discount_percent = ($item_->discount_type === 0) ? $item_->discount : ($item_->discount / $item_price) * 100;
            }

            // Item Unit Price
            if ($item_->unit_id !== $item->unit_id) {
                $unit       = ItemUnit::where('unit_id', $item_->unit_id)->where('item_id', $item_->item_id)->first();
                $item_price = $unit->unit_value * $item_price;
            } else {
                $unit = Unit::where('id', $item_->unit_id)->first();
            }

            // net price is the unit price - discount
            $net_price  = ($item_price * $item_->quantity) - $discount;

            $taxableItems     = collect();
            $totalTaxableFees = 0;
            $invoiceVats      = $invoice->vats->where('item_id', $item_->item_id);

            // Get item vats.
            $totalItemTax = count($invoiceVats) > 0
                ? $this->invoiceVats($item_, $invoiceVats, $taxableItems, $totalTaxableFees, $taxes, $net_price, $item_->quantity)
                : 0;

            // Calculate the total amount of the item.
            $itemTotal = $this->invoiceTotal($item_price, $item_->quantity, $discount, $totalItemTax);

            // Total amount of all the items sales.
            $totalSalesAmount += $itemTotal['salesTotal'];

            // Non-taxable items discount.
            $itemsDiscount = $item_->items_discount ?? 0;

            // Total amount of all the items discount.
            $totalItemsDiscountAmount += $itemsDiscount;

            // Add the item to the invoice items.
            $invoiceItems = [
                "description"      => $item->about ?? '',
                "itemType"         => "EGS",
                "itemCode"         => $item->item_einvoice_code,
                "unitType"         => $unit->unit_einvoice_code,
                "quantity"         => (float) $item_->quantity,
                "salesTotal"       => einvoice_decimal($itemTotal['salesTotal']),
                "total"            => einvoice_decimal($itemTotal['total']),
                "valueDifference"  => einvoice_decimal($itemTotal['valueDifference']),
                "totalTaxableFees" => einvoice_decimal($totalTaxableFees),
                "netTotal"         => einvoice_decimal($itemTotal['netTotal']),
                "itemsDiscount"    => einvoice_decimal($itemsDiscount),
                "unitValue" => [
                    "currencySold" => $settings->currency_id,
                    "amountEGP"    => einvoice_decimal($item_price),
                ],
            ];

            // If the item has discount.
            if (isset($discount)) {
                $invoiceItems['discount'] = [
                    "rate"   => einvoice_decimal($discount_percent),
                    "amount" => einvoice_decimal($discount),
                ];
            }

            // if currency is not equal to EGP.
            if ($invoiceItems['unitValue']['currencySold'] !== 'EGP') {
                $invoiceItems['unitValue']['amountSold'] = (float) $item_price;
            }

            // Calculate the total discount amount.
            $totalDiscount += $invoiceItems['discount']['amount'];

            // Add the taxable items to the invoice items if item has vats.
            if ($taxableItems->count() > 0) {
                $invoiceItems['taxableItems'] = $taxableItems;
            }

            $invoiceLines->push($invoiceItems);
        }

        return [
            'totalTaxableFees'         => einvoice_decimal($totalTaxableFees),
            'totalSalesAmount'         => einvoice_decimal($totalSalesAmount),
            'totalDiscount'            => einvoice_decimal($totalDiscount),
            'totalItemsDiscountAmount' => einvoice_decimal($totalItemsDiscountAmount),
        ];
    }

    /**
     * This Method to get the item vats.
     *
     * @param $item_
     * @param $vats
     * @param $taxableItems
     * @param $totalTaxableFees
     * @param $taxes
     * @param $net_price
     * @param $quantity
     * @return numeric-string $totalItemTax
     */
    public function invoiceVats($item_, $vats, $taxableItems, $totalTaxableFees, $taxes, $net_price, $quantity): string
    {
        foreach ($vats as $vat) {

            // Get the vat from the bills_add
            $vat_ = Vat::where('id', $vat->vat_id)->first();

            /**
             * Add the vat to the taxable items
             * (vat_type = 0) => نسبة , (vat_type = 1) => مبلغ
             * (vat_operation = 0) => إضافة , (vat_operation = 1) => خصم
             */
            $rate   = $vat->vat_type == 0 ? $vat->vat_value : ($vat->vat_value / $net_price) * 100;
            $amount = ($vat->vat_type == 1 ? $vat->vat_value : ($vat->vat_value * $net_price) / 100);

            // Check if the vat is already added to the taxable items.
            $allTaxableItem = $taxableItems->pluck('taxType')->toArray();
            $allTaxes       = $taxes->pluck('taxType')->toArray();

            if (!in_array($vat_->einvoice_code, $allTaxableItem)) {
                $taxableItems->push([
                    "taxType" => $vat_->einvoice_code,
                    "rate"    => einvoice_decimal($rate),
                    "amount"  => einvoice_decimal($amount),
                    "subType" => $vat_->sub_einvoice_code
                ]);
            }

            if (!in_array($vat_->einvoice_code, $allTaxes)) {
                $taxes->push([
                    "taxType" => $vat_->einvoice_code,
                    "amount"  => einvoice_decimal($amount),
                ]);
            } else {
                $tax            = $taxes->where('taxType', $vat_->einvoice_code)->first();
                $key            = $taxes->search($tax);

                $taxes->forget($key);
                $taxes->push([
                    "taxType" => $vat_->einvoice_code,
                    "amount"  => einvoice_decimal($tax['amount'] + $amount),
                ]);
            }

            // Total amount of additional taxable fees to be used in final tax calculation.
            $totalTaxableFees += $item_->additional_taxable_fees ?? 0;
        }

        // Total item tax.
        $totalItemTax = $taxableItems->sum('amount') + $totalTaxableFees;

        return einvoice_decimal($totalItemTax);
    }

    /**
     * This Method is used to calculate the tax totals.
     *
     * @param $taxes
     * @return array $taxTotals
     */
    public function invoiceTaxTotal($taxes): array
    {
        $taxTotals = collect();

        // Calculate the tax totals amount.
        $taxes->groupBy('taxType')->each(function ($taxes, $key) use ($taxTotals) {
            $taxTotals->push([
                "taxType" => $key,
                "amount"  => einvoice_decimal($taxes->sum('amount')),
            ]);
        });

        // Calculate the sum of all the tax totals.
        $totalTaxAmount = $taxTotals->sum('amount');

        return [
            "taxTotals"      => $taxTotals,
            "totalTaxAmount" => einvoice_decimal($totalTaxAmount),
        ];
    }

    /**
     * This Method is used to calculate totals.
     *
     * @param $item_price
     * @param $quantity
     * @param $discount
     * @param $totalItemTax
     * @return array
     */
    public function invoiceTotal($item_price, $quantity, $discount, $totalItemTax): array
    {
        // Total amount considering quantity and unit price.
        $salesTotal = $item_price * $quantity;

        // Total amount after adding all pricing items, taxes, removing discounts.
        $total = $salesTotal + $totalItemTax - $discount;

        // Value difference when selling goods already taxed.
        $valueDifference = $total - $salesTotal;

        // Total amount for the invoice line after applying discount.
        $netTotal = $total - $totalItemTax;

        return [
            "salesTotal"      => einvoice_decimal($salesTotal),
            "total"           => einvoice_decimal($total),
            "valueDifference" => einvoice_decimal(0 ?? $valueDifference),
            "netTotal"        => einvoice_decimal($netTotal),
        ];
    }
}
