<?php

namespace App\Services\Einvoice\Invoice;

use App\Services\Einvoice\Interfaces\UpdateDocument;
use Exception;
use Illuminate\Http\Response;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Log;
use Modules\EInvoice\Entities\EinvoiceDetail;
use Modules\EInvoice\Entities\EinvoiceDetailVat;
use Modules\EInvoice\Entities\Einvoices;
use Modules\EInvoice\Entities\EinvoiceStage;

class Invoice extends InvoiceDocument implements UpdateDocument
{
    public array $invoiceResult;
    public Collection $einvoiceLogs;
    public Collection $submitLog;
    public Collection $operationLogs;
    public array $submitLogCount;
    public array $operationLogsCount;

    public function __construct()
    {
        parent::__construct();
        $this->invoiceResult      = [];
        $this->einvoiceLogs       = collect();
        $this->submitLog          = collect();
        $this->operationLogs      = collect();
        $this->submitLogCount     = [];
        $this->operationLogsCount = [];
    }

    /**
     * Taxpayer can submit one or more signed documents to eInvoicing solution using this Method.
     *
     * @param mixed $bills
     * @param string cert_token
     * @return Response
     *
     * This Method Can Submit (Invoices, Credit Notes, Debit Notes).
     */
    public function submitDocuments($bills, $cert_token): Response
    {
        try {
            if (blank($cert_token)) throw new Exception('Please use token');

            // Check If $bills is array or not.
            $bills          = is_array($bills) ? $bills : collect([$bills]);
            $data           = [];
            $response       = [];

            foreach ($bills as $bill){
                /**
                 * Intialize submit stages in einvoice stage table (Intialize stages for each document).
                 * einvoiceLogs   => All logs with status 0 , 1 , 2.
                 * submitLogCount => First log (submit einvoice stage) with status 3 (done).
                 */
                $this->initEinvoiceStage($bill);

                /**
                 * Get stages from einvoice stage table. (submitLog is the first stage and operationLogs are the rest of stages).
                 * submitLog          => Submit einvoice stage that has stage == 0.
                 * operationLogs      => All stages from $einvoiceLogs with stage != 0 (all logs except submit einvoice stage).
                 * operationLogsCount => Count of $operationLogs.
                 */
                $this->getEinvoiceStage($bill);

                /**
                 * Get the submitted document data from einvoice stage table.
                 * If the document is already submitted and done, then get the submitted document data from einvoice stage table.
                 */
                if (empty($this->submitLog[$bill]) && $this->submitLogCount[$bill] > 0) {
                    $sumittedDocument = $this->getEinvoiceStageDocument($bill);
                    $data             = $sumittedDocument['data'];
                }

                /**
                 * Submit the documents to eInvoicing solution or retry to submit if failed.
                 * If the document is not submitted or failed, then retry to submit the document to eInvoicing solution.
                 */
                if (!empty($this->submitLog[$bill]) || $this->submitLogCount[$bill] === 0 ||
                    (!empty($data) && !empty($data['rejectedDocuments']) && empty($data['acceptedDocuments']))) {

                    // Set Bills In Document Template For E-Invoice.
                    $this->setDocument($bill, $cert_token);
                }

                // Log the submitted documents in einvoice stage table (Request).
                $this->logEinvoiceStage('invoice', $bill, 0, 1, $response);
            }

            // Check If $bills are already submmited or not.
            $submitStatus = EinvoiceStage::whereIn('invoice_id',$bills)->where(['stage' => 0, 'status' => 3])->first();
            $otherStages  = EinvoiceStage::whereIn('invoice_id',$bills)
                ->where('stage','>',0)->whereBetween('status', [0, 1 ,2])
                ->count();

            if (!$submitStatus && $otherStages > 0) {
                // Submit the documents to eInvoicing solution or retry to submit if failed (If there are valid documents) (stage 0).
                $submittedEinvoice = $this->submitEinvoice($bills);

                // Abort if $submittedEinvoice is instance of Response.
                $responseError = $this->getErrorResponse($submittedEinvoice);

                if (!is_null($responseError)) {
                    return $responseError;
                }
            }else {
                $submittedEinvoice                  = json_decode($submitStatus->request, true);
                $submittedEinvoice['status']        = 200;
                $submittedEinvoice['data']          = json_decode($submitStatus->response, true);
                $submittedEinvoice['rejectedBills'] = [
                    'failedBills'   => $submittedEinvoice["failedBills"],
                    "notFoundBills" => $submittedEinvoice["notFoundBills"]
                ];
            }

            $invoices              = $submittedEinvoice['invoices'];
            $status                = $submittedEinvoice['status'];
            $data                  = $submittedEinvoice['data'];
            $data['rejectedBills'] = $submittedEinvoice['rejectedBills'];

            if ($this->failedBills->count()) {
                Log::channel('invoice')->error('There are some bills that are not valid to be submitted to e-invoice service.',
                    ['bills' => $this->failedBills]);
            }

            if ($this->notFoundBills->count()) {
                Log::channel('invoice')->error('There are some bills that are not exists in the system.', ['bills' => $this->notFoundBills]);
            }

            // Update the database with the submitted document. (If the document is submitted successfully).
            foreach ($bills as $bill) {
                if (($status >= 200 && $status < 300) || ($this->operationLogsCount[$bill] > 0 && empty($data['rejectedDocuments'])
                        && !empty($data['acceptedDocuments']))) {
                    sleep(2);

                    $response = $this->updateEinvoice($data, $invoices, $this->operationLogs[$bill]);
                }
            }

            // Data from the response
            return response($data, $status);
        } catch (\Exception $e) {
            // Log the error in einvoice stage table (Response).
            $this->logEinvoiceStage('invoice', $bills, 0, 2, $bills, $e->getMessage());
            Log::channel('invoice')->error('Error in submitDocuments method: ' . $e->getMessage() . ' , Line: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * Initialize e-invoice Stage to submit to e-invoice service.
     *
     * @param $bill
     * @return Invoice
     */
    public function initEinvoiceStage($bill): Invoice
    {
        $this->einvoiceLogs[$bill] = EinvoiceStage::where('invoice_id', $bill)->whereIn('status', [0, 1, 2])->orderBy('stage', 'asc')->get();
        $submitCheck               =  EinvoiceStage::where('invoice_id', $bill)->where(['status' => 3, 'stage' => 0])->first();

        // Intialize submit stages in einvoice stage table (Intialize stages for document).
        ($this->einvoiceLogs[$bill]->count() === 0 && !$submitCheck)
            ? $this->logEinvoiceStage('invoice', $bill)
            : Log::channel('invoice')
                ->info("Re-submit Invoice , Invoice Id : $bill (Retrying from stage : {$this->einvoiceLogs[$bill]->first()->stage})");

        // Update value of einvoice stages.
        $this->einvoiceLogs[$bill]   = EinvoiceStage::where('invoice_id', $bill)->whereIn('status', [0, 1, 2])->orderBy('stage', 'asc')->get();
        $this->submitLogCount[$bill] = EinvoiceStage::where('invoice_id', $bill)->where('stage', 0)->whereIn('status', [3])->count();

        return $this;
    }

    /**
     * Get Stages From E-invoice Stage Table.
     *
     * @param $bill
     * @return Invoice
     */
    public function getEinvoiceStage($bill): Invoice
    {
        $submitStage                     = $this->einvoiceLogs[$bill]->where('stage', 0);
        $this->submitLog[$bill]          = ($submitStage->count() === 1) ? $submitStage->first()->toArray() : [];
        $this->operationLogs[$bill]      = $this->einvoiceLogs[$bill]->whereBetween('stage', [1, 6])->values();
        $this->operationLogsCount[$bill] = count($this->operationLogs);

        return $this;
    }

    /**
     * Get the submitted document data from e-invoice stage table.
     *
     * @param $bill
     * @return array
     */
    public function getEinvoiceStageDocument($bill): array
    {
        $query    = EinvoiceStage::where('invoice_id', $bill)->where('stage', 0)->selectRaw('request, response')->first();
        $request  = json_decode($query->request, true);
        $data     = json_decode($query->response, true);
        $status   = !is_null($data['rejectedDocuments']) && !empty($data['acceptedDocuments']) ? 200 : 400;

        return ['body' => $request['body'], 'invoices' => $request['invoices'], 'status' => $status, 'data' => $data];
    }

    /**
     * Submit the documents to eInvoicing solution or retry to submit if failed.
     *
     * @param $bills
     * @return array|Response
     */
    public function submitEinvoice($bills): Response|array
    {
        // Intialize submit stages in einvoice stage table (Intialize stages for each document).
        $body          = ["documents" => $this->documents->values()->toArray()];
        $invoices      = $this->documents->keys()->toArray();
        $submitRequest = [
            'body'          => $body,
            'invoices'      => $invoices,
            'failedBills'   => $this->failedBills->toArray(),
            'notFoundBills' => $this->notFoundBills->toArray()
        ];

        if (!empty($body['documents']) && !empty($invoices)) {

            // Submit the documents to eInvoicing solution.
            $submitedData = $this->sendEinvoice($body);
            $data         = $submitedData['data'];
            $status       = $submitedData['status'];
            $stageStatus  = ($status >= 200 && $status < 300 && empty($data['rejectedDocuments'])) ? 3 : 2;

            // Update einvoice stage table with submitted document data (Response).
            foreach ($bills as $bill) {
                $this->logEinvoiceStage('invoice', $bill, 0, $stageStatus, $submitRequest, $data);
            }

            // Log Error if failed to submit the documents to eInvoicing solution.
            if ($status >= 200 && $status < 300) {
                $message = is_array($bills) ? 'All E-Invoice Submitted Successfully' : 'E-Invoice Submitted Successfully';
                Log::channel('invoice')->info($message);
            } else {
                Log::channel('invoice')->error('Failed to submit the einvoice to E-Invoicing solution with error: ' . $data['error']);
                return response(['payload_error' => $data['error']], 400);
            }

            return [
                'status'        => $status,
                'data'          => $data,
                'invoices'      => $invoices,
                'body'          => $body,
                'rejectedBills' => ['failedBills' => $this->failedBills->toArray(), 'notFoundBills' => $this->notFoundBills->toArray()]
            ];
        }

        Log::channel('invoice')->error('There are no valid documents to submit to e-invoice service.', ['bills' => $bills]);
        return response(['rejectedBills' => ['failedBills' => $this->failedBills->toArray(), 'notFoundBills' => $this->notFoundBills->toArray()]], 422);
    }

    /**
     * This Method Returns Error Message.
     * @param $submittedEinvoice
     * @return Response|null
     */
    public function getErrorResponse($submittedEinvoice): Response|null
    {
        if ($submittedEinvoice instanceof Response) {
            if (($submittedEinvoice->getStatusCode() === 400) && isset($submittedEinvoice->original['payload_error'])) {
                $payloadTime = explode(' ', $submittedEinvoice->original['payload_error']);
                $payloadTime = $payloadTime[19] / 60;
                $payloadTime = ceil($payloadTime);
                return response(['status' => 'payload_error', 'payload_msg' => __('einvoice.payload_error', ['time' => $payloadTime])], 400);
            }

            if (($submittedEinvoice->getStatusCode() === 422) && isset($submittedEinvoice->original['rejectedBills'])) {
                return response(['status' => 'rejectedBills', 'rejectedBills' => $submittedEinvoice->original['rejectedBills']], 422);
            }
        }

        return null;
    }

    /**
     * Update the database with the submitted document.
     *
     * @param $data
     * @param $invoices
     * @param $operationLogs
     * @return response
     */
    public function updateEinvoice($data, $invoices, $operationLogs): Response
    {
        $einvoice = $this->updateEinvoiceTables($data, $invoices, $operationLogs);

        if (isset($einvoice) && !empty($einvoice) && in_array($einvoice->status(), [422, 500])) {
            Log::channel('invoice')->error('Error in updating invoice, Error: ' . $einvoice->original . ' , Line: ' . __LINE__ . ' , Status: ' . $einvoice->status());
            return response(['error' => $einvoice], $einvoice->status());
        }

        if (isset($data['rejectedDocuments']) && !empty($data['rejectedDocuments'] && empty($data['acceptedDocuments']))) {
            $errors = collect($data['rejectedDocuments'][0]['error']['details'])->map(function ($item) {
                return $item['message'];
            })->implode(', ');

            Log::channel('invoice')->error('Invoice Rejected , Errors : ' . json_encode($errors));

            return response(['error' => explode(',', $errors)], 422);
        }

        Log::channel('invoice')->info('Invoice Submitted Successfully , Invoice UUID : ' . $data['acceptedDocuments'][0]['uuid']);
        return response(['success' => 'Invoice Submitted Successfully'], 200);
    }

    /**
     * This Method Update the Invoice in database.
     *
     * @param $data
     * @param $invoices
     * @param $operationLogs
     * @return Response
     */
    public function updateEinvoiceTables($data, $invoices, $operationLogs): Response
    {
        // Send Notification to Taxpayer and update the invoice in database.
        if (isset($data['acceptedDocuments'])) {
            foreach ($data['acceptedDocuments'] as $index => $document) {

                // Get the einvoice status from log einvoice stage table (stage 1).
                $this->getEinvoiceStatus($operationLogs, $invoices, $index, $data, $document);

                // Update the invoice detail in database.
                try {

                    // Get the invoice document details and retry if failed.
                    DB::transaction(function () use ($index, $document, $invoices, $operationLogs) {

                        // Get the invoice document (stage 2).
                        $einvoiceDocument = $this->getEinvoiceDocument($operationLogs, $invoices, $index, $document);

                        $einvoice         = $einvoiceDocument['einvoice'];

                        // Get the invoice details (stage 3).
                        $einvoiceDetails = $this->getEinvoiceDetails($operationLogs, $invoices, $index, $document);
                        $details         = $einvoiceDetails['details'];

                        // Update the invoice in database.
                        if (isset($einvoice['uuid']) && !empty($einvoice) && isset($details['uuid']) && !empty($details)) {

                            // Check log einvoice stage table for invoice (stage 4).
                            $e_invoice = $this->saveEinvoice($operationLogs, $invoices, $index, $details, $einvoice);

                            // Check log einvoice stage table for invoice details (stage 5).
                            $invoice_detail = $this->saveInvoiceDetail($operationLogs, $invoices, $index, $details, $e_invoice);

                            // Check log einvoice stage table for vats (stage 6).
                            $this->saveInvoiceVat($operationLogs, $invoices, $index, $details, $e_invoice, $invoice_detail);

                            Log::channel('invoice')->info('Invoice Added Successfully To Database , Invoice UUID : ' . $einvoice['uuid']);
                            return response(__('einvoice.einvoice_added_successfully'), 201);
                        }

                        if (is_string($einvoice)) {
                            Log::channel('invoice')->error('Invoice Details is empty , Invoice Details : ' . $einvoice);
                        }
                        return response(__('einvoice.invoice_details_is_empty'), 422);
                    });
                } catch (\Exception $e) {
                    Log::channel('invoice')->error('Error in updateEinvoiceTables method: ' . $e->getMessage() . ' , Line Number: ' . $e->getLine() . ' , File: ' . $e->getFile());
                    return response(__('einvoice.error_in_update_invoice'), 500);
                }
            }
        }

        return response(__('einvoice.no_accepted_documents'), 422);
    }

    /**
     * Get the e-invoice status from log e-invoice stage table (stage 1).
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $data
     * @param $document
     * @return void
     */
    public function getEinvoiceStatus($operationLogs, $invoices, $index, $data, $document): void
    {
        $updateEinvoiceLog = $operationLogs->where('stage', 1);
        $updateEinvoiceLog = $updateEinvoiceLog->count() === 1 ? $updateEinvoiceLog->first()->toArray() : [];

        if (!empty($updateEinvoiceLog) && in_array($updateEinvoiceLog['status'], [0, 1, 2], true)) {
            // Update the invoice in database.
            $invoice  = \Modules\EInvoice\Entities\Invoice::where('id', $invoices[$index])->first();

            // Log the submitted document in einvoice stage table (Request)
            $this->logEinvoiceStage('invoice', $invoices[$index], 1, 1, [$data, $document]);

            $update   = $invoice->update([
                'submissionId' => $data['submissionId'],
                'uuid'         => $document['uuid'],
                'longId'       => $document['longId'],
                'internalId'   => $document['internalId'],
                'hashKey'      => $document['hashKey'],
            ]);

            $stageStatus = $update ? 3 : 2;
            $response    = $update ? __('extends.update_sucessfully') : __('ex tends.update_failed');

            // Log the invoice submission in einvoice stage table (Response)
            $this->logEinvoiceStage('invoice', $invoices[$index], 1, $stageStatus, [$data, $document], $response);
        }
    }

    /**
     * Get the invoice document from einvoice stage table (stage 2).
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $document
     * @return array
     */
    public function getEinvoiceDocument($operationLogs, $invoices, $index, $document): array
    {
        $getEinvoiceLog = $operationLogs->where('stage', 2);
        $getEinvoiceLog = $getEinvoiceLog->count() === 1 ? $getEinvoiceLog->first()->toArray() : [];

        if (!empty($getEinvoiceLog) && in_array($getEinvoiceLog['status'], [0, 1, 2], true)) {
            $result = $this->getDocument($document['uuid']);

            // Log the einvoice in einvoice stage table (Request).
            $this->logEinvoiceStage('invoice', $invoices[$index], 2, 1, $document['uuid']);

            // Log the einvoice details in einvoice stage table (Request).
            $einvoiceStageStatus = $result->status() === 200 ? 3 : 2;
            $einvoice            = $result->original;
            $this->logEinvoiceStage('invoice', $invoices[$index], 2, $einvoiceStageStatus, $document['uuid'], $einvoice);
        } else {
            $result   = EinvoiceStage::where('invoice_id', $invoices[$index])->where('stage', 2)->where('status', 3)->select('response')->first();
            $einvoice = json_decode($result->response, true);
        }

        return ['einvoice' => $einvoice];
    }

    /**
     * Get the invoice details from einvoice stage table (stage 3).
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $document
     * @return array
     */
    public function getEinvoiceDetails($operationLogs, $invoices, $index, $document): array
    {
        $getEinvoiceDetailsLog = $operationLogs->where('stage', 3);
        $getEinvoiceDetailsLog = $getEinvoiceDetailsLog->count() === 1 ? $getEinvoiceDetailsLog->first()->toArray() : [];

        if (!empty($getEinvoiceDetailsLog) && in_array($getEinvoiceDetailsLog['status'], [0, 1, 2], true)) {
            $details_result      = $this->getDocumentDetails($document['uuid']);

            // Log the einvoice details in einvoice stage table (Request)
            $this->logEinvoiceStage('invoice', $invoices[$index], 3, 1, $document['uuid']);

            // Log the einvoice details in einvoice stage table (Response)
            $detailsStageStatus  = $details_result->status() === 200 ? 3 : 2;
            $details             = $details_result->original;
            $this->logEinvoiceStage('invoice', $invoices[$index], 3, $detailsStageStatus, $document['uuid'], $details);
        } else {
            $result  = EinvoiceStage::where('invoice_id', $invoices[$index])->where('stage', 3)->where('status', 3)->select('response')->first();
            $details = json_decode($result->response, true);
        }

        return ['details' => $details];
    }

    /**
     * Check log einvoice stage table for invoice.
     * If the invoice is already exist in database then update the invoice.
     * If the invoice is not exist in database then create the invoice.
     * (stage 4).
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $details
     * @param $einvoice
     * @return Collection|Einvoices
     */
    public function saveEinvoice($operationLogs, $invoices, $index, $details, $einvoice): Einvoices|Collection
    {
        $updateEinvoiceLog = $operationLogs->where('stage', 4);
        $updateEinvoiceLog = $updateEinvoiceLog->count() === 1 ? $updateEinvoiceLog->first()->toArray() : [];

        if (!empty($updateEinvoiceLog) && in_array($updateEinvoiceLog['status'], [0, 1, 2])) {

            // Check the invoice is already exist in database.
            $einvoice_ = Einvoices::where('invoice_id', $invoices[$index]);
            $e_invoice = $einvoice_->exists() ? $einvoice_->first() : null;

            if (is_null($e_invoice)) {
                // Log the submitted document in einvoice stage table (Request)
                $this->logEinvoiceStage('invoice', $invoices[$index], 4, 1, [$einvoice, $details]);
                $einvoices_ = Einvoices::where('invoice_id', $invoices[$index]);
                $e_invoice_ = $einvoices_->exists() ? $einvoices_->first() : null;

                if (is_null($e_invoice_)) {
                    $e_invoice = Einvoices::create([
                        "uuid"                          => $einvoice['uuid'],
                        "submissionUUID"                => $einvoice['submissionUUID'],
                        "longId"                        => $einvoice['longId'],
                        "internalId"                    => $einvoice['internalId'],
                        "invoice_id"                    => $invoices[$index],
                        "typeName"                      => $einvoice['typeName'],
                        "documentTypeNamePrimaryLang"   => $details['documentTypeNamePrimaryLang'],
                        "documentTypeNameSecondaryLang" => $details['documentTypeNameSecondaryLang'],
                        "typeVersionName"               => $einvoice['typeVersionName'],
                        "issuerId"                      => $einvoice['issuerId'],
                        "issuerName"                    => $einvoice['issuerName'],
                        "issuerType"                    => $details['issuer']['type'],
                        "receiverId"                    => $einvoice['receiverId'],
                        "receiverName"                  => $einvoice['receiverName'],
                        "receiverType"                  => $details['receiver']['type'],
                        "taxpayerActivityCode"          => $details['taxpayerActivityCode'],
                        "dateTimeIssued"                => $einvoice['dateTimeIssued'],
                        "dateTimeReceived"              => $einvoice['dateTimeReceived'],
                        "totalSales"                    => $einvoice['totalSales'],
                        "totalDiscount"                 => $einvoice['totalDiscount'],
                        "totalItemsDiscountAmount"      => $details['totalItemsDiscountAmount'],
                        "extraDiscountAmount"           => $details['extraDiscountAmount'],
                        "taxTotals"                     => collect($details['taxTotals'])->sum('amount'),
                        "netAmount"                     => $einvoice['netAmount'],
                        "total"                         => $einvoice['total'],
                        "maxPercision"                  => $einvoice['maxPercision'],
                        "cancelRequestDate"             => $details['cancelRequestDate'],
                        "rejectRequestDate"             => $details['rejectRequestDate'],
                        "cancelRequestDelayedDate"      => $details['cancelRequestDelayedDate'],
                        "rejectRequestDelayedDate"      => $details['rejectRequestDelayedDate'],
                        "declineCancelRequestDate"      => $details['declineCancelRequestDate'],
                        "declineRejectRequestDate"      => $details['declineRejectRequestDate'],
                        "canbeCancelledUntil"           => $details['canbeCancelledUntil'],
                        "canbeRejectedUntil"            => $details['canbeRejectedUntil'],
                        "documentStatusReason"          => $details['documentStatusReason'],
                        "currenciesSold"                => $details['currenciesSold'],
                        "statusId"                      => $details['statusId'],
                        "status"                        => $details['status'],
                        "publicUrl"                     => $details['publicUrl'],
                        "validationResults"             => json_encode($details['validationResults']),
                    ]);
                }

                // Log the invoice submission in einvoice stage table (Response)
                $e_invoiceStatusStage = !empty($e_invoice) ? 3 : 2;
                $e_invoiceResponse    = $e_invoice ? __('extends.save_sucessfully') : __('extends.save_failed');
                $this->logEinvoiceStage('invoice', $invoices[$index], 4, $e_invoiceStatusStage, [$einvoice, $details], $e_invoiceResponse);
            }
        }

        return $e_invoice ?? collect();
    }

    /**
     * Check log einvoice stage table for invoice.
     * If the invoice detials is already exist in database then update the invoice.
     * If the invoice details is not exist in database then create the invoice.
     * (stage 5).
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $details
     * @param $e_invoice
     * @return Collection|EinvoiceDetail
     */
    public function saveInvoiceDetail($operationLogs, $invoices, $index, $details, $e_invoice): EinvoiceDetail|Collection
    {
        $updateEinvoiceDetailLog = $operationLogs->where('stage', 5);
        $updateEinvoiceDetailLog = $updateEinvoiceDetailLog->count() === 1 ? $updateEinvoiceDetailLog->first()->toArray() : [];

        if (!empty($updateEinvoiceDetailLog) && in_array($updateEinvoiceDetailLog['status'], [0, 1, 2])) {

            // Log the submitted document in einvoice stage table (Request)
            $this->logEinvoiceStage('invoice', $invoices[$index], 5, 1, $details['invoiceLines']);

            // Add Einvoice details to the einvoice details table.
            foreach ($details['invoiceLines'] as $item) {
                $invoiceDetail  = EinvoiceDetail::where('einvoice_id', $e_invoice->id)->where('itemCode', $item['itemCode']);
                $invoice_detail = $invoiceDetail->exists() ? $invoiceDetail->first() : null;

                if (is_null($invoice_detail)) {
                    $invoice_detail = EinvoiceDetail::create([
                        "einvoice_id"                    => $e_invoice->id,
                        "itemPrimaryName"                => $item['itemPrimaryName'],
                        "itemPrimaryDescription"         => $item['itemPrimaryDescription'],
                        "itemSecondaryName"              => $item['itemSecondaryName'],
                        "itemSecondaryDescription"       => $item['itemSecondaryDescription'],
                        "unitTypePrimaryName"            => $item['unitTypePrimaryName'],
                        "unitTypePrimaryDescription"     => $item['unitTypePrimaryDescription'],
                        "unitTypeSecondaryName"          => $item['unitTypeSecondaryName'],
                        "unitTypeSecondaryDescription"   => $item['unitTypeSecondaryDescription'],
                        "unitValue_amountSold"           => $item['unitValue']['amountSold'],
                        "unitValue_amountEGP"            => $item['unitValue']['amountEGP'],
                        "unitValue_currencyExchangeRate" => $item['unitValue']['currencyExchangeRate'],
                        "description"                    => $item['description'],
                        "itemType"                       => $item['itemType'],
                        "itemCode"                       => $item['itemCode'],
                        "unitType"                       => $item['unitType'],
                        "quantity"                       => $item['quantity'],
                        "salesTotal"                     => $item['salesTotal'],
                        "discount_rate"                  => $item['discount']['rate'],
                        "discount_amount"                => $item['discount']['amount'],
                        "taxableItems"                   => $item['taxableItems'],
                        "internalCode"                   => $item['internalCode'],
                        "itemsDiscount"                  => $item['itemsDiscount'],
                        "netTotal"                       => $item['netTotal'],
                        "totalTaxableFees"               => $item['totalTaxableFees'],
                        "valueDifference"                => $item['valueDifference'],
                        "total"                          => $item['total'],
                    ]);
                }
            }

            // Log the invoice submission in einvoice details stage table.
            $detailStatusStage = !empty($invoice_detail) ? 3 : 2;
            $detailResponse    = $invoice_detail ? __('extends.save_sucessfully') : __('extends.save_failed');
            $this->logEinvoiceStage('invoice', $invoices[$index], 5, $detailStatusStage, $details['invoiceLines'], $detailResponse);
        }

        return $invoice_detail ?? collect();
    }

    /**
     * Check log e-invoice stage table for invoice.
     * If the invoice details vats is already exist in database then update the invoice.
     * If the invoice details vats is not exist in database then create the invoice.
     * (stage 6).
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $details
     * @param $e_invoice
     * @param $invoice_detail
     * @return void
     */
    public function saveInvoiceVat($operationLogs, $invoices, $index, $details, $e_invoice, $invoice_detail): void
    {
        $updateEinvoiceVatLog = $operationLogs->where('stage', 6);
        $updateEinvoiceVatLog = $updateEinvoiceVatLog->count() === 1 ? $updateEinvoiceVatLog->first()->toArray() : [];

        if (!empty($updateEinvoiceVatLog) && in_array($updateEinvoiceVatLog['status'], [0, 1, 2])) {

            // Add E-invoice vats to the e-invoice vats table.
            $vats = Collect($details['invoiceLines'])->pluck('lineTaxableItems');

            // Log the submitted document in einvoice stage table (Request)
            $this->logEinvoiceStage('invoice', $invoices[$index], 6, 1, $details['invoiceLines']);
            foreach ($vats as $vat) {
                $invoice_vat = null;
                if (!empty($vat)) {
                    $invoiceVat  = EinvoiceDetailVat::where('einvoice_id', $e_invoice->id)->where('item_id', $invoice_detail->id);
                    $invoice_vat = $invoiceVat->exists() ? $invoiceVat->first() : null;

                    if (is_null($invoice_vat)) {
                        $invoice_vat = EinvoiceDetailVat::create([
                            "einvoice_id"   => $e_invoice->id,
                            "item_id"       => $invoice_detail->id,
                            "amountForeign" => $vat[0]['amountForeign'],
                            "taxType"       => $vat[0]['taxType'],
                            "amount"        => $vat[0]['amount'],
                            "subType"       => $vat[0]['subType'],
                            "rate"          => $vat[0]['rate'],
                        ]);
                    }
                }
            }

            // Check if items vats are added successfully or if there is no vats.
            $this->checkInvoiceVat($invoice_vat, $e_invoice, $invoices, $details, $index, $vats);
        }
    }

    /**
     * Check if items vats are added successfully or if there is no vats.
     *
     * @param $invoice_vat
     * @param $e_invoice
     * @param $invoices
     * @param $details
     * @param $index
     * @param $vats
     * @return void
     */
    public function checkInvoiceVat($invoice_vat, $e_invoice, $invoices, $details, $index, $vats): void
    {
        $einvoiceID = $e_invoice->id ?? Einvoices::where('invoice_id', $invoices[$index])->select('id')->first()?->id;
        $vatsCount  = EinvoiceDetailVat::where('einvoice_id', $einvoiceID)->count();

        if ($vatsCount > 0 || empty($vats)) {
            // Log the invoice submission in einvoice vats stage table (Response)
            $vatStatusStage = !empty($invoice_vat) ? 3 : 2;
            $vatResponse    = $invoice_vat ? __('extends.save_sucessfully') : __('extends.save_failed');
            $this->logEinvoiceStage('invoice', $invoices[$index], 6, $vatStatusStage, $details['invoiceLines'], $vatResponse);
        } else {
            // Log the invoice vats submission in einvoice stages table (No Vats)
            $this->logEinvoiceStage('invoice', $invoices[$index], 6, 2, $details['invoiceLines'], __('einvoice.no_vats'));
        }
    }
}
