<?php

namespace App\Services\Einvoice\Receipt;

use Exception;

class ReceiptDocument extends Initiate
{
    public function __construct()
    {
        parent::__construct();
        $this->token = $this->loginTaxpayer();
       // abort_if($this->statusCode != 200, $this->statusCode, $this->token);
    }

    /**
     * Send E-Invoice to the Taxpayer
     *
     * @param $body
     * @return array
     */
    public function sendEinvoice($body): array
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/v1/receiptsubmissions',
                'token' => $this->token,
                'body' => $body,
                'asJson' => true,
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);
            dd($response, $this->token, $body);

            // If the response is successful data is returned as associative array.
            return ['data' => $response->content, 'status' => $response->status];
        } catch (Exception $e) {
            return response($e->getMessage(), 500);
            return ['data' => json_encode(['message' => $e->getMessage()]), 'status' => 500];
        }
    }

    /**
     * This Method allows taxpayers to cancel previously issued document.
     *
     * @param string $documentUUID
     * @param string $reason
     * @return Response $data
     *
     * This Method Can Cancel (Invoices, Credit Notes, Debit Notes).
     */
    public function cancelDocument(string $documentUUID, string $reason): Response
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documents/state/' . $documentUUID . '/state',
                'token' => $this->token,
                'body' => [
                    "status" => "cancelled",
                    "reason" => $reason
                ],
                'asJson' => true,
                'method' => 'put',
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);
            $data = $response->content;

            // Data from the response
            return ($response->status >= 200 && $response->status < 300) ? response($data, $response->status) : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in cancelDocument method: ' . $e->getMessage() . ' Line Number: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayer that received an invalid invoice to reject it.
     *
     * @param string $documentUUID
     * @param string $reason
     * @return Response $data
     *
     * This Method Can Reject (Invoices, Credit Notes, Debit Notes).
     */
    public function rejectDocument(string $documentUUID, string $reason): Response
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documents/state/' . $documentUUID . '/state',
                'token' => $this->token,
                'body' => [
                    "status" => "rejected",
                    "reason" => $reason
                ],
                'asJson' => true,
                'method' => 'put',
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);
            $data = $response->content;

            // Data from the response
            return ($response->status >= 200 && $response->status < 300) ? response($data, $response->status) : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in rejectDocument method: ' . $e->getMessage() . ' Line Number: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers "recipient" to decline document that was previously cancelled by other taxpayer "issuer".
     *
     * @param string $documentUUID
     * @return Response $data
     */
    public function declineCancelDocument(string $documentUUID): Response
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documents/state/' . $documentUUID . '/decline/cancelation',
                'token' => $this->token,
                'asJson' => true,
                'method' => 'put',
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);
            $data = $response->content;

            // Data from the response
            return ($response->status >= 200 && $response->status < 300) ? response($data, $response->status) : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in declineCancelDocument method: ' . $e->getMessage() . ' Line Number: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers "Issuer" to decline document that was previously rejected by other taxpayer "Receiver".
     *
     * @param string $documentUUID
     * @return Response $data
     */
    public function declineRejectDocument(string $documentUUID): Response
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documents/state/' . $documentUUID . '/decline/rejection',
                'token' => $this->token,
                'asJson' => true,
                'method' => 'put',
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);
            $data = $response->content;

            // Data from the response
            return ($response->status >= 200 && $response->status < 300)
                ? response($data, $response->status)
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in declineRejectDocument method: ' . $e->getMessage() . ' Line Number: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayer systems to retrieve list of document types published by the eInvoicing solution.
     *
     * @return Response $data
     */
    public function getDocumentTypes(): Response
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documenttypes',
                'token' => $this->token,
                'headers' => ['Accept-Language' => app()->getLocale()],
                'asJson' => true,
                'method' => 'get',
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);
            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getDocumentTypes: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayer systems to query latest documents sent or received that are available on the eInvoicing solution.
     *
     * @param int $pageSize
     * @param int $pageNumber
     * @return Response $data
     */
    public function getRecentDocuments(int $pageSize = 20, int $pageNumber = 1): Response
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documents/recent?pageNo=' . $pageNumber . '&pageSize=' . $pageSize,
                'token' => $this->token,
                'asJson' => true,
                'method' => 'get',
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);
            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getRecentDocuments: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method returns information on documents submitted during a single submission by taxpayer.
     *
     * @param string $submitionUUID
     * @param int $pageSize
     * @param int $pageNumber
     * @return Response $data
     */
    public function getSubmittedDocument(string $submitionUUID, int $pageSize = 10, int $pageNumber = 1): Response
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documentSubmissions/' . $submitionUUID . '?PageSize=' . $pageSize . '&PageNumber=' . $pageNumber,
                'token' => $this->token,
                'asJson' => true,
                'method' => 'get',
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);

            // Retry if the response is not successful.
            while (!($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
            }

            $data = $response->content;

            // Data from the response
            return ($response->status >= 200 && $response->status < 300) ? response($data, $response->status) : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getSubmittedDocument: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers to retrieve document.
     *
     * @param string $documentUUID
     * @return Response $data
     */
    public function getDocument(string $documentUUID): Response
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documents/' . $documentUUID . '/raw',
                'token' => $this->token,
                'asJson' => true,
                'method' => 'get',
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);

            // Retry if the request fails.
            $count = 1;
            while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
                $count++;
            }

            $data = $response->content;

            // Data from the response
            return ($response->status >= 200 && $response->status < 300) ? response($data, $response->status) : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getDocument: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method enables ERP system to download the PDF version of the document having predefined layout.
     *
     * @param string $documentUUID
     * @return Response $data
     */
    public function getDocumentPDF(string $documentUUID): Response
    {
        try {
            // Dowanload Path
            if (!file_exists(storage_path('app/modules/einvoice/documents')) && !mkdir($concurrentDirectory = storage_path('app/modules/einvoice/documents'), 0777, true) && !is_dir($concurrentDirectory)) {
                throw new \RuntimeException(sprintf('Directory "%s" was not created', $concurrentDirectory));
            }

            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documents/' . $documentUUID . '/pdf',
                'token' => $this->token,
                'contentType' => 'application/pdf',
                'download' => storage_path('app/modules/einvoice/documents/' . $documentUUID . '.pdf'),
            ];

            // Send a request to the Taxpayer API
            $response = $this->sendCurl($curlData);
            $data = json_decode($response->content);

            // Data from the response
            return ($response->status >= 200 && $response->status < 300)
                ? response($data, $response->status)->header('Content-Type', 'application/pdf')
                : response($data->error->details[0]->message, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getDocumentPDF: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers to retrieve a single document full details including validation results.
     *
     * @param string $documentUUID
     * @return Response $data
     */
    public function getDocumentDetails(string $documentUUID): Response
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documents/' . $documentUUID . '/details',
                'token' => $this->token,
                'asJson' => true,
                'connectTimeout' => 60,
                'timeout' => 10,
                'method' => 'get',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->sendCurl($curlData);
            $count = 0;
            while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
                $count++;
            }

            Log::channel('invoice')->error('success in getDocumentDetails: ' . json_encode($response));
            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getDocumentDetails: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers to retrieve information on previous document package requests, their status.
     *
     * @param number $pageNo
     * @param number $pageSize
     * @return $data
     */
    public function getPackageRequests($pageSize = 20, $pageNo = 1)
    {
        try {
            $curlData = [
                'url' => self::$apiBaseUrl . '/api/1/documentPackages/requests?pageSize=' . $pageSize . '&pageNo=' . $pageNo,
                'token' => $this->token,
                'asJson' => true,
                'method' => 'get',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->sendCurl($curlData);
            $count    = 0;
            while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
                $count++;
            }

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getPackageRequests: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayer systems to download previously requested document packages as XML or JSON files (compressed).
     *
     * @param string $rid (Unique ID of the package)
     * @return $data
     */
    public function getDocumentPackage($rid)
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/api/1/documentPackages/' . $rid,
                'token'  => $this->token,
                'asJson' => true,
                'method' => 'get',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->sendCurl($curlData);
            $count    = 0;
            while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
                $count++;
            }

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getPackageRequests: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayer system to request larger packages of previously sent or received documents to be prepared for download.
     *
     * @param array $data
     * @return $data
     */
    public function requestDocumentPackage($data)
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/api/1/documentPackages/requests',
                'token'  => $this->token,
                'method' => 'post',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->sendCurl($curlData);
            $count    = 0;
            while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
                $count++;
            }

            $new_data = $response->content;

            // Data from the response
            return response($new_data, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getPackageRequests: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow eInvoicing to verify during registration of the system notifications that API is accessible from the eInvoicing solution.
     *
     * @param  string $rin
     * @return string $response
     */
    public function ping($rin)
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/ping',
                'token'  => $this->token,
                'body'   => ['rin' => $rin],
                'headers' => [
                    'Accept-Language' =>  app()->getLocale(),
                ],
                'asJson' => true,
                'method' => 'put',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->sendCurl($curlData);
            $count    = 0;
            while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
                $count++;
            }

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in ping: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows ERP system to query for previously received notifications.
     *
     * @param $query
     * @return mixed
     */
    public function getNotification($query)
    {
        try {
            $url = in_array($query, [' ', '', null], true)
                ? self::$apiBaseUrl . '/api/1/notifications/taxpayer'
                : self::$apiBaseUrl . '/api/1/notifications/taxpayer?' . $query;

            $curlData = [
                'url'    => $url,
                'token'  => $this->token,
                'asJson' => true,
                'method' => 'get',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->sendCurl($curlData);
            $count    = 0;
            while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
                $count++;
            }

            // Data from the response
            return $response->content;
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in getNotification: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method enables ERP system to receive notifications on document related events - validation, issuance, rejection and cancellation.
     *
     * @param $document
     * @param $count
     * @return mixed
     */
    public function receiveDocumentNotifications($document, $count = 1)
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/notifications/documents',
                'token'  => $this->token,
                'body'   => [
                    "deliveryId" => Str::random(10),
                    "type"       => "document",
                    "count"      => $count,
                    "message"    => [
                        "type"           => $document['type'],
                        "uuid"           => $document['uuid'],
                        "longId"         => $document['longId'],
                        "status"         => $document['status'],
                        "internalId"     => $document['internalId'],
                        "submissionUUID" => $document['submissionUUID'],
                    ]
                ],
                'headers' => [
                    'Accept-Language' => app()->getLocale(),
                ],
                'asJson' => true,
                'method' => 'put',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->sendCurl($curlData);
            $count    = 0;
            while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
                $count++;
            }

            // Data from the response
            return $response->content;
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in receiveDocumentNotifications: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows ERP system to receive notification when document package is ready for download.
     *
     * @param $count
     * @return mixed
     */
    public function receiveDownloadReadyNotification($count = 1)
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/notifications/documentpackages',
                'token'  => $this->token,
                'body'   => [
                    "deliveryId" => Str::random(10),
                    "type"       => "document-package-ready",
                    "count"      => $count,
                    "message"    => [
                        "type"      => "document-package-ready",
                        "packageId" => Str::random(10),
                    ]
                ],
                'headers' => [
                    'Accept-Language' => app()->getLocale(),
                ],
                'asJson' => true,
                'method' => 'put',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->sendCurl($curlData);
            $count    = 0;
            while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
                sleep(1);
                $response = $this->sendCurl($curlData);
                $count++;
            }

            // Data from the response
            return $response->content;
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in receiveDownloadReadyNotification: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to create their own codes in the system (EGS Codes).
     *
     * @param array $data
     */
    public function createEgsCode($data)
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/api/1/codetypes/requests/codes',
                'token'  => $this->token,
                'body'   => ["items" => $data],
                'asJson' => true,
                'method' => 'post',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->loopRequest($curlData, method: 'createEgsCode');

            // Data from the response
            return response($response->content, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in createEgsCode: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to update their own codes in the system (EGS Codes).
     *
     * @param string $itemCode
     * @param array $body
     * @param string $codeType
     * @return response
     */
    public function updateCode($itemCode, $body, $codeType = 'EGS')
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/api/1/codetypes/' . $codeType . '/codes/' . $itemCode,
                'token'  => $this->token,
                'body'   => $body,
                'asJson' => true,
                'method' => 'put',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->loopRequest($curlData, method: 'updateCode');

            // Data from the response
            return response($response->content, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in updateCode: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayer to Update his previous code usage requests as long they are under ‘Submitted’ state.
     *
     * @param string $codeUsageRequestID
     * @param array $body
     * @param string $codeType
     * @return response
     */
    public function updateEgsCode($codeUsageRequestID, $body)
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/api/1/codetypes/requests/codes/' . $codeUsageRequestID,
                'token'  => $this->token,
                'body'   => $body,
                'asJson' => true,
                'method' => 'put',
            ];

            // Send a request to the Taxpayer API/
            $response = $this->loopRequest($curlData, method: 'updateEgsCode');

            // Data from the response
            return response($response->content, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in updateEgsCode: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to get code details from the system (EGS Codes).
     *
     * @param string $codeType
     * @param string $itemCode
     * @return response
     */
    public function getEgsCodeDetails($itemCode, $codeType = 'EGS')
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/api/1/codetypes/' . $codeType . '/codes/' . $itemCode,
                'token'  => $this->token,
                'asJson' => true,
                'method' => 'put',
            ];

            $response = $this->loopRequest($curlData, method: 'getEgsCodeDetails');

            // Data from the response
            return response($response->content, $response->status);
        } catch (\Exception $e) {
            Log::channel('invoice')->error('Error in updateEgsCode: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to search for codes in the system (EGS Codes).
     *
     * @param string $query
     * @return response
     */
    public function searchEgsCode($query)
    {
        try {
            // Check if there is a query or not.
            $url = in_array($query, [' ', '', null])
                ? self::$apiBaseUrl . '/api/1/codetypes/requests/my'
                : self::$apiBaseUrl . '/api/1/codetypes/requests/my?' . $query;

            $curlData = [
                'url'    => $url,
                'token'  => $this->token,
                'asJson' => true,
                'method' => 'get',
            ];

            $response = $this->loopRequest($curlData, method: 'searchEgsCode');
            // Data from the response
            return response($response->content, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in searchEgsCode Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to request code reusue from current existing code.
     *
     * @param array $data
     * @return response
     */
    public function requestCodeReuse($data)
    {
        try {
            $curlData = [
                'url'    => self::$apiBaseUrl . '/api/1/codetypes/requests/codeusages',
                'token'  => $this->token,
                'body'   => ["items" => $data],
                'asJson' => true,
                'method' => 'put',
            ];

            $response = $this->loopRequest($curlData, method: 'requestCodeReuse');
            // Data from the response
            return response($response->content, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in requestCodeReuse Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to search EGS and GS1 published codes in the solution.
     *
     * @param string $query
     * @param string $codeType
     * @return response
     */
    public function searchPublishedCodes($query, $codeType)
    {
        try {
            // Check if there is a query or not.
            $url = in_array($query, [' ', '', null])
                ? self::$apiBaseUrl . '/api/1/codetypes/' . $codeType . '/codes'
                : self::$apiBaseUrl . '/api/1/codetypes/' . $codeType . '/codes?' . $query;

            $curlData = [
                'url'    => $url,
                'token'  => $this->token,
                'asJson' => true,
                'method' => 'get',
            ];

            $response = $this->loopRequest($curlData, method: 'searchPublishedCodes');
            // Data from the response
            return response($response->content, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in searchPublishedCodes Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    protected function loopRequest($curlData, $count = 0, $method = '')
    {
        $response = $this->sendCurl($curlData);
        while ($count <= 5 && !($response->status >= 200 && $response->status < 300)) {
            Log::channel('einvoice')->error("Retrying to get $method , Error: " . $response->content . ' Status: ' . $response->status);
            sleep(1);
            $response = $this->sendCurl($curlData);
            $count++;
        }
        return $response;
    }
}
