<?php

namespace App\Traits\Einvoice;

use Ixudra\Curl\Facades\Curl;
use Illuminate\Support\Facades\Log;
use Modules\EInvoice\Entities\Settings;
use Modules\EInvoice\Entities\EinvoiceStage;

trait AuthTrait
{
    public static $idSrvBaseUrl = 'https://id.preprod.eta.gov.eg';
    public static $apiBaseUrl   = 'https://api.preprod.invoicing.eta.gov.eg';
    public $token;
    public $posToken;
    public $clientId;
    public $clientSecret;

    function __construct($token = '', $posToken = '')
    {
        $this->token        = $token;
        $this->posToken     = $posToken;
        $this->clientId     = Settings::shop()->first()->client_id;
        $this->clientSecret = Settings::shop()->first()->client_secret;
    }

    /**
     * This Method allows taxpayer system to get access token to be used to access other protected APIs of the solution.
     *
     * @return token
     */
    public function loginTaxpayer()
    {
        try {
            $body = [
                'grant_type'    => 'client_credentials',
                'client_id'     => $this->clientId,
                'client_secret' => $this->clientSecret,
                'scope'         => 'InvoicingAPI',
            ];

            $response = Curl::to(self::$idSrvBaseUrl . '/connect/token')
                ->withContentType('application/x-www-form-urlencoded')
                ->withData($body)
                ->returnResponseObject()
                ->post();

            // Retry if token is not valid
            while ($response->status != 200) {
                sleep(1);
                Log::channel('einvoice')->error('Retrying to get access token , Error: ' . $response->content . ' Status: ' . $response->status);
                $response = Curl::to(self::$idSrvBaseUrl . '/connect/token')
                    ->withContentType('application/x-www-form-urlencoded')
                    ->withData($body)
                    ->returnResponseObject()
                    ->post();
            }

            // If the response is successful data is returned as associative array
            $data = json_decode($response->content, true);

            // Data from the response
            return $response->status == 200
                ? $data['access_token']
                : $data;
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in loginTaxpayer Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayer system to get access token to be used to access other protected APIs of the solution.
     *
     * @return posToken
     */
    public function authenticatePOS()
    {
        try {
            $body = [
                'grant_type'    => 'client_credentials',
                'client_id'     => $this->clientId,
                'client_secret' => $this->clientSecret,
            ];

            $response = Curl::to(self::$idSrvBaseUrl . '/connect/token')
                ->withContentType('application/x-www-form-urlencoded')
                ->withHeaders([
                    'posserial: YV200T573',
                    'pososversion: V/OS (linux-based)',
                    'presharedkey: C68FEC13-FE34-4839-89F9-ADB0D678CFE0'
                ])
                ->withData($body)
                ->returnResponseObject()
                ->post();

            // Retry if token is not valid
            while ($response->status != 200) {
                sleep(1);
                Log::channel('einvoice')->error('Retrying to get pos access token , Error: ' . $response->content . ' Status: ' . $response->status);
                $response = Curl::to(self::$idSrvBaseUrl . '/connect/token')
                    ->withContentType('application/x-www-form-urlencoded')
                    ->withHeaders([
                        'posserial: YV200T573',
                        'pososversion: V/OS (linux-based)',
                        'presharedkey: C68FEC13-FE34-4839-89F9-ADB0D678CFE0'
                    ])
                    ->withData($body)
                    ->returnResponseObject()
                    ->post();
            }

            // If the response is successful data is returned as associative array
            $data = json_decode($response->content, true);

            // Data from the response
            return $response->status == 200
                ? $data['access_token']
                : $data['error'];
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in authenticatePOS Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method is used to log the stages of the einvoice in einvoice stage table.
     * @param $data
     * @param $stage
     * @param $status
     * @param $request
     * @param $response
     *
     * type:
     * 1 => invoice
     * 2 => item code
     *
     * status:
     * 0 => initializing
     * 1 => pending
     * 2 => error
     * 3 => done
     *
     * stages:
     * 0  => submit einvoice
     * 1  => update invoice table
     * 2  => get einvoice
     * 3  => get einvoice details
     * 4  => save einvoice in db
     * 5  => save einvoice details in db
     * 6  => save einvoice detail in vats in db
     * 7  => save item code in einvoice solution
     * 8  => update items table with item code
     * 9  => save item code in db
     */
    public function logEinvoiceStage($type, $data, $stage = null, $status = null, $request = null, $response = null)
    {
        try {
            // Check type of einvoice and set stages and count
            if ($type == 'invoice' && !is_null($data)) {
                $einvoiceStageCount = EinvoiceStage::where('invoice_id', $data)->count();
                $einvoiceStage      = EinvoiceStage::where('invoice_id', $data)->where('stage', $stage);
                $logText            = 'Intialize submit stages in einvoice stage table (Intialize stages for invoice id: ' . $data . ')';
                $stages             = [
                    ['invoice_id' => $data, 'stage' => 0],
                    ['invoice_id' => $data, 'stage' => 1],
                    ['invoice_id' => $data, 'stage' => 2],
                    ['invoice_id' => $data, 'stage' => 3],
                    ['invoice_id' => $data, 'stage' => 4],
                    ['invoice_id' => $data, 'stage' => 5],
                    ['invoice_id' => $data, 'stage' => 6],
                ];
            } else if ($type == 'item_code' && !is_null($data)) {
                $einvoiceStageCount = EinvoiceStage::where('item_code', $data)->count();
                $einvoiceStage      = EinvoiceStage::where('item_code', $data)->where('stage', $stage);
                $logText            = 'Intialize submit stages in einvoice stage table (Intialize stages for item code: ' . $data . ')';
                $stages             = [
                    ['item_code' => $data, 'stage' => 7],
                    ['item_code' => $data, 'stage' => 8],
                    ['item_code' => $data, 'stage' => 9],
                ];
            }

            if (is_null($stage)) {
                Log::channel('einvoice')->info($logText);
                if ($einvoiceStageCount == 0) {
                    foreach ($stages as $stage_) {
                        EinvoiceStage::create($stage_);
                    }
                }
            } else {
                $data = ['status' => $status, 'updated_at' => now()];

                if ($request != null && !empty($request)) {
                    $data['request']    = $request;
                    $data['request_at'] = now();
                }

                if ($response != null && !empty($response)) {
                    $data['response']    = $response;
                    $data['response_at'] = now();
                }

                $einvoiceStage = $einvoiceStage->update($data);
                $einvoiceStage
                    ? Log::channel('einvoice')->info('Einvoice Stage Updated Successfully , Stage : ' . $stage . ' , Status : ' . $status)
                    : Log::channel('einvoice')->error('Error in updating einvoice stage , Stage : ' . $stage . ' , Status : ' . $status);
            }
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in logEinvoiceStage method: ' . $e->getMessage() . ' , Line Number: ' . $e->getLine());
        }
    }
}
