<?php

namespace App\Traits\Einvoice\Invoice;

use Illuminate\Http\Response;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Log;
use Modules\EInvoice\Entities\Invoice;
use Modules\EInvoice\Entities\Einvoices;
use Modules\EInvoice\Entities\EinvoiceStage;
use Modules\EInvoice\Entities\EinvoiceDetail;
use Modules\EInvoice\Entities\EinvoiceDetailVat;

trait InitiateDocumentTrait
{
    use InvoiceDocumentTrait;

    /**
     * Intialize einvoice to submit to einvoice service.
     *
     * @param $bills
     * @return array
     */
    public function initEinvoice($bills)
    {
        // Login to the Taxpayer API.
         $this->token = in_array($this->token, [null, ''], true) ? $this->loginTaxpayer() : $this->token;

        // Set Document to be submitted.
        $documents = $this->setDocument($bills);

        // Validate Document.
        if (isset($documents['status']) && $documents['status'] === 'validation') {
            return response(['validation_errors' => $documents['errors']], 422);
        }

        $body = ["documents" => $documents->values()->toArray()];

        // Log the submitted documents in einvoice stage table (Request).
        $this->logEinvoiceStage('invoice', $bills, 0, 1, ['body' => $body, 'invoices' => $documents->keys()->toArray()]);

        return ['body' => $body, 'invoices' => $documents->keys()];
    }

    /**
     * Intialize einvoice Stage to submit to einvoice service.
     *
     * @param $bills
     * @return array
     */
    public function initEinvoiceStage($bills)
    {
        $einvoiceLogs   = EinvoiceStage::where('invoice_id', $bills)->whereIn('status', [0, 1, 2])->orderBy('stage', 'asc')->get();
        $submitLogCount = EinvoiceStage::where('invoice_id', $bills)->where('stage', 0)->whereIn('status', [3])->count();

        // Intialize submit stages in einvoice stage table (Intialize stages for document).
        ($einvoiceLogs->count() == 0)
            ? $this->logEinvoiceStage('invoice', $bills)
            : Log::channel('einvoice')->info('Re-submit Invoice , Invoice Id : ' . $bills . ' (Retrying from stage : ' . $einvoiceLogs->first()->stage . ')');

        // Update value of einvoice stages.
        $einvoiceLogs   = EinvoiceStage::where('invoice_id', $bills)->whereIn('status', [0, 1, 2])->orderBy('stage', 'asc')->get();
        $submitLogCount = EinvoiceStage::where('invoice_id', $bills)->where('stage', 0)->whereIn('status', [3])->count();

        return ['einvoiceLogs' => $einvoiceLogs, 'submitLogCount' => $submitLogCount];
    }

    /**
     * Get Stages From Einvoice Stage Table.
     *
     * @param $einvoiceLogs
     * @return array
     */
    public function getEinvoiceStage($einvoiceLogs)
    {
        $submitLog          = $einvoiceLogs->where('stage', 0);
        $submitLog          = ($submitLog->count() == 1) ? $submitLog->first()->toArray() : [];
        $operationLogs      = $einvoiceLogs->whereNotIn('stage', [0])->values();
        $operationLogsCount = count($operationLogs);

        return ['submitLog' => $submitLog, 'operationLogs' => $operationLogs, 'operationLogsCount' => $operationLogsCount];
    }

    /**
     * Get the submitted document data from einvoice stage table.
     *
     * @param $bills
     * @return array
     */
    public function getEinvoiceStageDocument($bills): array
    {
        $einvoiceRequest  = EinvoiceStage::where('invoice_id', $bills)->where('stage', 0)->select('request')->first();
        $einvoiceResponse = EinvoiceStage::where('invoice_id', $bills)->where('stage', 0)->select('response')->first();

        $einvoiceRequest  = json_decode($einvoiceRequest->request, true);
        $body             = $einvoiceRequest['body'];
        $data             = json_decode($einvoiceResponse->response, true);
        $invoices         = $einvoiceRequest['invoices'];
        $status           = !is_null($data['rejectedDocuments']) && !empty($data['acceptedDocuments']) ? 200 : 400;

        return ['body' => $body, 'invoices' => $invoices, 'status' => $status, 'data' => $data];
    }

    /**
     * Submit the documents to eInvoicing solution or retry to submit if failed.
     *
     * @param $bills
     * @return array | \Illuminate\Http\Response
     */
    public function submitEinvoice($bills)
    {
        // Intialize submit stages in einvoice stage table (Intialize stages for each document).
        $intialize = $this->initEinvoice($bills);

        // Check if the document is not valid.
        if ($intialize instanceof \Illuminate\Http\Response && $intialize->getStatusCode() == 422 && $intialize->original['validation_errors']) {
            return $intialize;
        }

        $body      = $intialize['body'];
        $invoices  = $intialize['invoices'];

        // Submit the documents to eInvoicing solution.
        $submitedData = $this->sendEinvoice($body);
        $data         = $submitedData['data'];
        $status       = $submitedData['status'];
        $stageStatus  = in_array($status, [200, 201]) ? 3 : 2;

        // Update einvoice stage table with submitted document data (Response).
        $this->logEinvoiceStage('invoice', $bills, 0, $stageStatus, ['body' => $data, 'invoices' => $invoices], $data);

        // Log Error if failed to submit the documents to eInvoicing solution.
        if (in_array($status, [200, 201])) {
            Log::channel('einvoice')->info('Einvoice submitted successfully');
        } else {
            Log::channel('einvoice')->error('Failed to submit the einvoice to E-Invoicing solution with error: ' . $data['error']);
            return response(['payload_error' => $data['error']], 400);
        }

        return ['status' => $status, 'data' => $data, 'invoices' => $invoices, 'body' => $body];
    }

    /**
     * Update the database with the submitted document.
     *
     * @param $data
     * @param $invoices
     * @param $operationLogs
     * @return response
     */
    public function updateEinvoice($data, $invoices, $operationLogs)
    {
        sleep(1);
        $einvoice = $this->updateInvoice($data, $invoices, $operationLogs);

        if (isset($einvoice) && !empty($einvoice) && in_array($einvoice->status(), [422, 500])) {
            Log::channel('einvoice')->error('Error in updating invoice');
            return response(['error' => $einvoice], $einvoice->status());
        }

        if (isset($data['acceptedDocuments']) && !empty($data['acceptedDocuments'] && empty($data['rejectedDocuments']))) {
            Log::channel('einvoice')->info('Invoice Submitted Successfully , Invoice UUID : ' . $data['acceptedDocuments'][0]['uuid']);
        } else {
            $errors = collect($data['rejectedDocuments'][0]['error']['details'])->map(function ($item) {
                return $item['message'];
            })->implode(', ');

            Log::channel('einvoice')->error('Invoice Rejected , Errors : ' . $errors);

            return response(['error' => explode(',', $errors)], 422);
        }
    }

    /**
     * Taxpayer can submit one or more signed documents to eInvoicing solution using this Method.
     *
     * @param $bills
     * @return $data of submitted document
     *
     * This Method Can Submit (Invoices, Credit Notes, Debit Notes).
     */
    public function submitDocuments($bills)
    {
        try {
            // Intialize submit stages in einvoice stage table (Intialize stages for each document).
            $initStage      = $this->initEinvoiceStage($bills);
            $einvoiceLogs   = $initStage['einvoiceLogs'];
            $submitLogCount = $initStage['submitLogCount'];

            // Get stages from einvoice stage table. (submitLog is the first stage and operationLogs are the rest of stages).
            $stages             = $this->getEinvoiceStage($einvoiceLogs);
            $submitLog          = $stages['submitLog'];
            $operationLogs      = $stages['operationLogs'];
            $operationLogsCount = $stages['operationLogsCount'];

            // Get the submitted document data from einvoice stage table. (If the document is already submitted).
            if (empty($submitLog) && $submitLogCount > 0) {
                $sumittedDocument = $this->getEinvoiceStageDocument($bills);
                $body             = $sumittedDocument['body'];
                $data             = $sumittedDocument['data'];
                $invoices         = $sumittedDocument['invoices'];
                $status           = $sumittedDocument['status'];
            }

            // Data is the response of the submitted document.
            $data = isset($data) ? $data : [];

            // Submit the documents to eInvoicing solution or retry to submit if failed. (If the document is not submitted).
            if (!empty($submitLog) || $submitLogCount == 0 || (!empty($data['rejectedDocuments']) && empty($data['acceptedDocuments']))) {
                $submittedEinvoice = $this->submitEinvoice($bills);

                if ($submittedEinvoice instanceof \Illuminate\Http\Response && $submittedEinvoice->getStatusCode() == 400) {
                    $payloadTime = explode(' ', $submittedEinvoice->original['payload_error']);
                    $payloadTime = $payloadTime[19] / 60;
                    $payloadTime = ceil($payloadTime);
                    return response(['status' => 'payload_error', 'payload_msg' => __('einvoice.payload_error', ['time' => $payloadTime])], 400);
                }

                $body     = $submittedEinvoice['body'];
                $invoices = $submittedEinvoice['invoices'];
                $status   = $submittedEinvoice['status'];
            }

            // Update the database with the submitted document. (If the document is submitted successfully).
            if (in_array($status, [200, 201]) || ($operationLogsCount > 0 && empty($data['rejectedDocuments']) && !empty($data['acceptedDocuments']))) {
                $this->updateEinvoice($data, $invoices, $operationLogs);
            }

            // Data from the response
            return response($data, $status);
        } catch (\Exception $e) {
            // Log the error in einvoice stage table (Response).
            $intialize = $this->initEinvoice($bills);
            $body      = $intialize['body'];
            $this->logEinvoiceStage('invoice', $bills, 0, 2, $body, $e->getMessage());
            Log::channel('einvoice')->error('Error in submitDocuments method: ' . $e->getMessage() . ' , Line: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * Get the einvoice status from log einvoice stage table.
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $data
     * @param $document
     * @return void
     */
    public function getEinvoiceStatus($operationLogs, $invoices, $index, $data, $document): void
    {
        $updateEinvoiceLog = $operationLogs->where('stage', 1);
        $updateEinvoiceLog = $updateEinvoiceLog->count() == 1 ? $updateEinvoiceLog->first()->toArray() : [];

        if (!empty($updateEinvoiceLog) && in_array($updateEinvoiceLog['status'], [0, 1, 2])) {
            // Update the invoice in database.
            $invoice  = Invoice::where('id', $invoices[$index])->first();

            // Log the submitted document in einvoice stage table (Request)
            $this->logEinvoiceStage('invoice', $invoices[$index], 1, 1, [$data, $document]);

            $update   = $invoice->update([
                'submissionId'  => $data['submissionId'],
                'uuid'          => $document['uuid'],
                'longId'        => $document['longId'],
                'internalId'    => $document['internalId'],
                'hashKey'       => $document['hashKey'],
            ]);

            $stageStatus = $update == true ? 3 : 2;
            $response    = $update == true ? __('extends.update_sucessfully') : __('extends.update_failed');

            // Log the invoice submission in einvoice stage table (Response)
            $this->logEinvoiceStage('invoice', $invoices[$index], 1, $stageStatus, [$data, $document], $response);
        }
    }

    /**
     * Get the invoice document from einvoice stage table.
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $document
     * @return array
     */
    public function getEinvoiceDocument($operationLogs, $invoices, $index, $document)
    {
        $getEinvoiceLog = $operationLogs->where('stage', 2);
        $getEinvoiceLog = $getEinvoiceLog->count() == 1 ? $getEinvoiceLog->first()->toArray() : [];

        if (!empty($getEinvoiceLog) && in_array($getEinvoiceLog['status'], [0, 1, 2])) {
            $result              = $this->getDocument($document['uuid']);

            // Log the einvoice in einvoice stage table (Request)
            $this->logEinvoiceStage('invoice', $invoices[$index], 2, 1, $document['uuid']);

            // Log the einvoice details in einvoice stage table (Request)
            $einvoiceStageStatus = $result->status() == 200 ? 3 : 2;
            $einvoice            = $result->original;
            $this->logEinvoiceStage('invoice', $invoices[$index], 2, $einvoiceStageStatus, $document['uuid'], $einvoice);
        } else {
            $result   = EinvoiceStage::where('invoice_id', $invoices[$index])->where('stage', 2)->where('status', 3)->select('response')->first();
            $einvoice = json_decode($result->response, true);
        }

        return ['einvoice_result' => $result, 'einvoice' => $einvoice];
    }

    /**
     * Get the invoice details from einvoice stage table.
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $document
     * @return array
     */
    public function getEinvoiceDetails($operationLogs, $invoices, $index, $document)
    {
        $getEinvoiceDetailsLog = $operationLogs->where('stage', 3);
        $getEinvoiceDetailsLog = $getEinvoiceDetailsLog->count() == 1 ? $getEinvoiceDetailsLog->first()->toArray() : [];

        if (!empty($getEinvoiceDetailsLog) && in_array($getEinvoiceDetailsLog['status'], [0, 1, 2])) {
            $details_result      = $this->getDocumentDetails($document['uuid']);

            // Log the einvoice details in einvoice stage table (Request)
            $this->logEinvoiceStage('invoice', $invoices[$index], 3, 1, $document['uuid']);

            // Log the einvoice details in einvoice stage table (Response)
            $detailsStageStatus  = $details_result->status() == 200 ? 3 : 2;
            $details             = $details_result->original;
            $this->logEinvoiceStage('invoice', $invoices[$index], 3, $detailsStageStatus, $document['uuid'], $details);
        } else {
            $result  = EinvoiceStage::where('invoice_id', $invoices[$index])->where('stage', 3)->where('status', 3)->select('response')->first();
            $details = json_decode($result->response, true);
        }

        return ['details_result' => $result, 'details' => $details];
    }

    /**
     * Check log einvoice stage table for invoice.
     * If the invoice is already exist in database then update the invoice.
     * If the invoice is not exist in database then create the invoice.
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $details
     * @param $einvoice
     * @return $e_invoice
     */
    public function saveEinvoice($operationLogs, $invoices, $index, $details, $einvoice)
    {
        $updateEinvoiceLog = $operationLogs->where('stage', 4);
        $updateEinvoiceLog = $updateEinvoiceLog->count() == 1 ? $updateEinvoiceLog->first()->toArray() : [];

        if (!empty($updateEinvoiceLog) && in_array($updateEinvoiceLog['status'], [0, 1, 2])) {

            // Check the invoice is already exist in database.
            $einvoice_ = Einvoices::where('invoice_id', $invoices[$index]);
            $e_invoice = $einvoice_->exists() ? $einvoice_->first() : null;

            if ($e_invoice == null) {
                // Log the submitted document in einvoice stage table (Request)
                $this->logEinvoiceStage('invoice', $invoices[$index], 4, 1, [$einvoice, $details]);
                $einvoices_ = Einvoices::where('invoice_id', $invoices[$index]);
                $e_invoice_ = $einvoices_->exists() ? $einvoices_->first() : null;

                if ($e_invoice_ == null) {
                    $e_invoice = Einvoices::create([
                        "uuid"                          => $einvoice['uuid'],
                        "submissionUUID"                => $einvoice['submissionUUID'],
                        "longId"                        => $einvoice['longId'],
                        "internalId"                    => $einvoice['internalId'],
                        "invoice_id"                    => $invoices[$index],
                        "typeName"                      => $einvoice['typeName'],
                        "documentTypeNamePrimaryLang"   => $details['documentTypeNamePrimaryLang'],
                        "documentTypeNameSecondaryLang" => $details['documentTypeNameSecondaryLang'],
                        "typeVersionName"               => $einvoice['typeVersionName'],
                        "issuerId"                      => $einvoice['issuerId'],
                        "issuerName"                    => $einvoice['issuerName'],
                        "issuerType"                    => $details['issuer']['type'],
                        "receiverId"                    => $einvoice['receiverId'],
                        "receiverName"                  => $einvoice['receiverName'],
                        "receiverType"                  => $details['receiver']['type'],
                        "taxpayerActivityCode"          => $details['taxpayerActivityCode'],
                        "dateTimeIssued"                => $einvoice['dateTimeIssued'],
                        "dateTimeReceived"              => $einvoice['dateTimeReceived'],
                        "totalSales"                    => $einvoice['totalSales'],
                        "totalDiscount"                 => $einvoice['totalDiscount'],
                        "totalItemsDiscountAmount"      => $details['totalItemsDiscountAmount'],
                        "extraDiscountAmount"           => $details['extraDiscountAmount'],
                        "taxTotals"                     => collect($details['taxTotals'])->sum('amount'),
                        "netAmount"                     => $einvoice['netAmount'],
                        "total"                         => $einvoice['total'],
                        "maxPercision"                  => $einvoice['maxPercision'],
                        "cancelRequestDate"             => $details['cancelRequestDate'],
                        "rejectRequestDate"             => $details['rejectRequestDate'],
                        "cancelRequestDelayedDate"      => $details['cancelRequestDelayedDate'],
                        "rejectRequestDelayedDate"      => $details['rejectRequestDelayedDate'],
                        "declineCancelRequestDate"      => $details['declineCancelRequestDate'],
                        "declineRejectRequestDate"      => $details['declineRejectRequestDate'],
                        "canbeCancelledUntil"           => $details['canbeCancelledUntil'],
                        "canbeRejectedUntil"            => $details['canbeRejectedUntil'],
                        "documentStatusReason"          => $details['documentStatusReason'],
                        "currenciesSold"                => $details['currenciesSold'],
                        "statusId"                      => $details['statusId'],
                        "status"                        => $details['status'],
                        "publicUrl"                     => $details['publicUrl'],
                        "validationResults"             => json_encode($details['validationResults']),
                    ]);
                }

                // Log the invoice submission in einvoice stage table (Response)
                $e_invoiceStatusStage = !empty($e_invoice) ? 3 : 2;
                $e_invoiceResponse    = $e_invoice ? __('extends.save_sucessfully') : __('extends.save_failed');
                $this->logEinvoiceStage('invoice', $invoices[$index], 4, $e_invoiceStatusStage, [$einvoice, $details], $e_invoiceResponse);
            }
        }

        return $e_invoice ?? [];
    }

    /**
     * Check log einvoice stage table for invoice.
     * If the invoice detials is already exist in database then update the invoice.
     * If the invoice details is not exist in database then create the invoice.
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $details
     * @param $e_invoice
     * @return $invoice_detail
     */
    public function saveInvoiceDetail($operationLogs, $invoices, $index, $details, $e_invoice)
    {
        $updateEinvoiceDetailLog = $operationLogs->where('stage', 5);
        $updateEinvoiceDetailLog = $updateEinvoiceDetailLog->count() == 1 ? $updateEinvoiceDetailLog->first()->toArray() : [];

        if (!empty($updateEinvoiceDetailLog) && in_array($updateEinvoiceDetailLog['status'], [0, 1, 2])) {

            // Log the submitted document in einvoice stage table (Request)
            $this->logEinvoiceStage('invoice', $invoices[$index], 5, 1, $details['invoiceLines']);

            // Add Einvoice details to the einvoice details table.
            foreach ($details['invoiceLines'] as $item) {
                $invoiceDetail  = EinvoiceDetail::where('einvoice_id', $e_invoice->id)->where('itemCode', $item['itemCode']);
                $invoice_detail = $invoiceDetail->exists() ? $invoiceDetail->first() : null;

                if ($invoice_detail == null) {
                    $invoice_detail = EinvoiceDetail::create([
                        "einvoice_id"                    => $e_invoice->id,
                        "itemPrimaryName"                => $item['itemPrimaryName'],
                        "itemPrimaryDescription"         => $item['itemPrimaryDescription'],
                        "itemSecondaryName"              => $item['itemSecondaryName'],
                        "itemSecondaryDescription"       => $item['itemSecondaryDescription'],
                        "unitTypePrimaryName"            => $item['unitTypePrimaryName'],
                        "unitTypePrimaryDescription"     => $item['unitTypePrimaryDescription'],
                        "unitTypeSecondaryName"          => $item['unitTypeSecondaryName'],
                        "unitTypeSecondaryDescription"   => $item['unitTypeSecondaryDescription'],
                        "unitValue_amountSold"           => $item['unitValue']['amountSold'],
                        "unitValue_amountEGP"            => $item['unitValue']['amountEGP'],
                        "unitValue_currencyExchangeRate" => $item['unitValue']['currencyExchangeRate'],
                        "description"                    => $item['description'],
                        "itemType"                       => $item['itemType'],
                        "itemCode"                       => $item['itemCode'],
                        "unitType"                       => $item['unitType'],
                        "quantity"                       => $item['quantity'],
                        "salesTotal"                     => $item['salesTotal'],
                        "discount_rate"                  => $item['discount']['rate'],
                        "discount_amount"                => $item['discount']['amount'],
                        "taxableItems"                   => $item['taxableItems'],
                        "internalCode"                   => $item['internalCode'],
                        "itemsDiscount"                  => $item['itemsDiscount'],
                        "netTotal"                       => $item['netTotal'],
                        "totalTaxableFees"               => $item['totalTaxableFees'],
                        "valueDifference"                => $item['valueDifference'],
                        "total"                          => $item['total'],
                    ]);
                }
            }

            // Log the invoice submission in einvoice details stage table.
            $detailStatusStage = !empty($invoice_detail) ? 3 : 2;
            $detailResponse    = $invoice_detail ? __('extends.save_sucessfully') : __('extends.save_failed');
            $this->logEinvoiceStage('invoice', $invoices[$index], 5, $detailStatusStage, $details['invoiceLines'], $detailResponse);
        }

        return $invoice_detail ?? collect([]);
    }

    /**
     * Check log e-invoice stage table for invoice.
     * If the invoice details vats is already exist in database then update the invoice.
     * If the invoice details vats is not exist in database then create the invoice.
     *
     * @param $operationLogs
     * @param $invoices
     * @param $index
     * @param $details
     * @param $e_invoice
     * @return $invoice_detail
     */
    public function saveInvoiceVat($operationLogs, $invoices, $index, $details, $e_invoice, $invoice_detail)
    {
        $updateEinvoiceVatLog    = $operationLogs->where('stage', 6);
        $updateEinvoiceVatLog    = $updateEinvoiceVatLog->count() == 1 ? $updateEinvoiceVatLog->first()->toArray() : [];

        if (!empty($updateEinvoiceVatLog) && in_array($updateEinvoiceVatLog['status'], [0, 1, 2])) {

            // Add E-invoice vats to the e-invoice vats table.
            $vats = Collect($details['invoiceLines'])->pluck('lineTaxableItems');

            // Log the submitted document in einvoice stage table (Request)
            $this->logEinvoiceStage('invoice', $invoices[$index], 6, 1, $details['invoiceLines']);

            foreach ($vats as $vat) {
                if (!empty($vat)) {
                    $invoiceVat  = EinvoiceDetailVat::where('einvoice_id', $e_invoice->id)->where('item_id', $invoice_detail->id);
                    $invoice_vat = $invoiceVat->exists() ? $invoiceVat->first() : null;

                    if ($invoice_vat == null) {
                        $invoice_vat = EinvoiceDetailVat::create([
                            "einvoice_id"   => $e_invoice->id,
                            "item_id"       => $invoice_detail->id,
                            "amountForeign" => $vat[0]['amountForeign'],
                            "taxType"       => $vat[0]['taxType'],
                            "amount"        => $vat[0]['amount'],
                            "subType"       => $vat[0]['subType'],
                            "rate"          => $vat[0]['rate'],
                        ]);
                    }
                }
            }

            // Check if items vats are added successfully or if there is no vats.
            $this->checkInvoiceVat($invoice_vat, $e_invoice, $invoices, $details, $index);
        }
    }

    /**
     * Check if items vats are added successfully or if there is no vats.
     *
     * @param $invoice_vat
     * @param $e_invoice
     * @param $invoices
     * @param $details
     * @param $index
     * @return void
     */
    public function checkInvoiceVat($invoice_vat, $e_invoice, $invoices, $details, $index): void
    {
        $einvoiceID = isset($e_invoice->id) ? $e_invoice->id : Einvoices::where('invoice_id', $invoices[$index])->first(['id'])->id;
        $vatsCount  = EinvoiceDetailVat::where('einvoice_id', $einvoiceID)->count();

        if ($vatsCount > 0) {
            // Log the invoice submission in einvoice vats stage table (Response)
            $vatStatusStage = !empty($invoice_vat) ? 3 : 2;
            $vatResponse    = $invoice_vat ? __('extends.save_sucessfully') : __('extends.save_failed');
            $this->logEinvoiceStage('invoice', $invoices[$index], 6, $vatStatusStage, $details['invoiceLines'], $vatResponse);
        } else {
            // Log the invoice vats submission in einvoice stages table (No Vats)
            $this->logEinvoiceStage('invoice', $invoices[$index], 6, 2, $details['invoiceLines'], __('einvoice.no_vats'));
        }
    }

    /**
     * This Method Update the Invoice in database.
     *
     * @param $data
     * @param $invoices
     * @return $data of submitted document
     */
    public function updateInvoice($data, $invoices, $operationLogs)
    {
        // Send Notification to Taxpayer and update the invoice in database.
        if (isset($data['acceptedDocuments'])) {
            foreach ($data['acceptedDocuments'] as $index => $document) {

                // Get the einvoice status from log einvoice stage table.
                $this->getEinvoiceStatus($operationLogs, $invoices, $index, $data, $document);

                // Update the invoice detail in database.
                try {

                    // Get the invoice document details and retry if failed.
                    DB::transaction(function () use ($index, $document, $invoices, $operationLogs) {

                        // Get the invoice document.
                        $einvoiceDocument = $this->getEinvoiceDocument($operationLogs, $invoices, $index, $document);
                        $einvoice         = $einvoiceDocument['einvoice'];
                        $einvoice_result  = $einvoiceDocument['einvoice_result'];

                        // Get the invoice details.
                        $einvoiceDetails = $this->getEinvoiceDetails($operationLogs, $invoices, $index, $document);
                        $details         = $einvoiceDetails['details'];
                        $detail_result   = $einvoiceDetails['detail_result'];

                        // Update the invoice in database.
                        if (isset($einvoice['uuid']) && !empty($einvoice) && isset($details['uuid']) && !empty($details)) {

                            // Check log einvoice stage table for invoice.
                            $e_invoice = $this->saveEinvoice($operationLogs, $invoices, $index, $details, $einvoice);

                            // Check log einvoice stage table for invoice details.
                            $invoice_detail = $this->saveEinvoiceDetail($operationLogs, $invoices, $index, $details, $einvoice);

                            // Check log einvoice stage table for vats.
                            $this->saveInvoiceVat($operationLogs, $invoices, $index, $details, $e_invoice, $invoice_detail);

                            Log::channel('einvoice')->info('Invoice Added Successfully To Database , Invoice UUID : ' . $einvoice['uuid']);
                            return response(__('einvoice.einvoice_added_successfully'), 201);
                        }

                        Log::channel('einvoice')->error('Invoice Details is empty , Invoice Details : ' . $einvoice . ' , Status : ' . $einvoice_result->status() ?? 200);
                        return response(__('einvoice.invoice_details_is_empty'), 422);
                    });
                } catch (\Exception $e) {
                    Log::channel('einvoice')->error('Error in updateInvoice method: ' . $e->getMessage() . ' , Line Number: ' . $e->getLine());
                    return response(__('einvoice.error_in_update_invoice'), 500);
                }
            }
        }

        return response(__('einvoice.no_accepted_documents'), 422);
    }
}
