<?php

namespace App\Traits\Einvoice\Invoice;

use Ixudra\Curl\Facades\Curl;
use App\Traits\Einvoice\AuthTrait;
use Illuminate\Support\Facades\Log;

trait InvoiceDocumentTrait
{
    use AuthTrait;

    /**
     * This Method allows taxpayer systems to retrieve list of document types published by the eInvoicing solution.
     *
     * @return data for document types
     */
    public function getDocumentTypes()
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1/documenttypes')
                ->withBearer($this->token)
                ->withHeader('Accept-Language', app()->getLocale())
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getDocumentTypes: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayer systems to query latest documents sent or received that are available on the eInvoicing solution.
     *
     * @return documents
     */
    public function getRecentDocuments($pageSize = 20, $pageNumber = 1)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documents/recent?pageNo=' . $pageNumber . '&pageSize=' . $pageSize)
                ->withBearer($this->token)
                ->withContentType('application/json')
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getRecentDocuments: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method returns information on documents submitted during a single submission by taxpayer.
     *
     * @param submitionUUID
     * @return document
     */
    public function getSubmittedDocument($submitionUUID, $pageSize = 10, $pageNumber = 1)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documentSubmissions/' . $submitionUUID . '?PageSize=' . $pageSize . '&PageNumber=' . $pageNumber)
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            // Retry if the response is not successful.
            while (!in_array($response->status, [200, 201])) {
                sleep(1);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documentSubmissions/' . $submitionUUID . '?PageSize=' . $pageSize . '&PageNumber=' . $pageNumber)
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->get();
            }

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getSubmittedDocument: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers to retrieve document.
     *
     * @param documentUUID
     * @return document
     */
    public function getDocument($documentUUID)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/raw')
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            // Retry if the request fails.
            $count = 0;
            while ($count <= 5 && !in_array($response->status, [200, 201])) {
                sleep(1);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/raw')
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->get();
                $count++;
            }

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getDocument: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method enables ERP system to download the PDF version of the document having predefined layout.
     *
     * @param documentUUID
     * @return document as PDF
     */
    public function getDocumentPDF($documentUUID)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Dowanload Path
            if (!file_exists(storage_path('app/modules/einvoice/documents'))) {
                mkdir(storage_path('app/modules/einvoice/documents'), 0777, true);
            }

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/pdf')
                ->withBearer($this->token)
                ->withContentType('application/pdf')
                ->returnResponseObject()
                ->download(storage_path('app/modules/einvoice/documents/' . $documentUUID . '.pdf'));

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)->header('Content-Type', 'application/pdf')
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getDocumentPDF: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers to retrieve a single document full details including validation results.
     *
     * @param documentUUID
     * @return document
     */
    public function getDocumentDetails($documentUUID)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/details')
                ->withBearer($this->token)
                ->withTimeout(10)
                ->withConnectTimeout(60)
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            $count = 0;
            while ($count <= 5 && !in_array($response->status, [200, 201])) {
                sleep(1);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/details')
                    ->withBearer($this->token)
                    ->withTimeout(10)
                    ->withConnectTimeout(60)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->get();
                $count++;
            }

            $data = $response->content;

            // Data from the response
            return  response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getDocumentDetails: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }
}
