<?php

namespace App\Traits\Einvoice\Invoice;

use Ixudra\Curl\Facades\Curl;
use Illuminate\Support\Facades\Log;
use App\Traits\Einvoice\NotificationTrait;
use App\Traits\Einvoice\Invoice\SetDocumentTrait;
use App\Traits\Einvoice\Invoice\InitiateDocumentTrait;

trait InvoiceTrait
{
    use NotificationTrait, InitiateDocumentTrait, SetDocumentTrait;

    /**
     * Send E-Invoice to the Taxpayer
     *
     * @param $body
     * @return $data of submitted document
     */
    public function sendEinvoice($body)
    {
        // Send a request to the Taxpayer API
        $response = Curl::to(self::$apiBaseUrl . '/api/v1/documentsubmissions')
            ->withData($body)
            ->withBearer($this->token)
            ->withContentType('application/json')
            ->returnResponseObject()
            ->asJson(true)
            ->post();

        // Retry if the response is not successful
        $counter = 1;
        while ($counter <= 5 && !in_array($response->status, [200, 201])) {
            sleep(1);
            $response = Curl::to(self::$apiBaseUrl . '/api/v1/documentsubmissions')
                ->withData($body)
                ->withBearer($this->token)
                ->withContentType('application/json')
                ->returnResponseObject()
                ->asJson(true)
                ->post();
            $counter++;
        }

        // If the response is successful data is returned as associative array.
        return [
            'data'   => $response->content,
            'status' => $response->status,
        ];
    }

    /**
     * This Method allows taxpayers to cancel previously issued document.
     *
     * @param $documentUUID
     * @param $reason
     * @return $data
     *
     * This Method Can Cancel (Invoices, Credit Notes, Debit Notes).
     */
    public function cancelDocument($documentUUID, $reason)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            $body = [
                "status" => "cancelled",
                "reason" => $reason
            ];

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documents/state/' . $documentUUID . '/state')
                ->withData($body)
                ->withBearer($this->token)
                ->withContentType('application/json')
                ->returnResponseObject()
                ->asJson(true)
                ->put();

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in cancelDocument method: ' . $e->getMessage() . ' Line Number: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayer that received an invalid invoice to reject it.
     *
     * @param $documentUUID
     * @param $reason
     * @return $data
     *
     * This Method Can Reject (Invoices, Credit Notes, Debit Notes).
     */
    public function rejectDocument($documentUUID, $reason)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            $body = [
                "status" => "rejected",
                "reason" => $reason
            ];

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documents/state/' . $documentUUID . '/state')
                ->withData($body)
                ->withBearer($this->token)
                ->withContentType('application/json')
                ->returnResponseObject()
                ->asJson(true)
                ->put();

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in rejectDocument method: ' . $e->getMessage() . ' Line Number: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers "recipient" to decline document that was previously cancelled by other taxpayer "issuer".
     *
     * @param $documentUUID
     * @return $data
     */
    public function declineCancelDocument($documentUUID)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documents/state/' . $documentUUID . '/decline/cancelation')
                ->withBearer($this->token)
                ->withContentType('application/json')
                ->returnResponseObject()
                ->asJson(true)
                ->put();

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in declineCancelDocument method: ' . $e->getMessage() . ' Line Number: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers "Issuer" to decline document that was previously rejected by other taxpayer "Receiver".
     *
     * @param $documentUUID
     * @return $data
     */
    public function declineRejectDocument($documentUUID)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documents/state/' . $documentUUID . '/decline/rejection')
                ->withBearer($this->token)
                ->withContentType('application/json')
                ->returnResponseObject()
                ->asJson(true)
                ->put();

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in declineRejectDocument method: ' . $e->getMessage() . ' Line Number: ' . $e->getLine());
            return response($e->getMessage(), 500);
        }
    }
}
