<?php

namespace App\Traits\Einvoice\Invoice;

use App\Rules\NotNull;
use Illuminate\Support\Str;
use Modules\EInvoice\Entities\Vat;
use Modules\EInvoice\Entities\Item;
use Modules\EInvoice\Entities\Unit;
use Modules\EInvoice\Entities\Clients;
use Modules\EInvoice\Entities\Invoice;
use Modules\EInvoice\Entities\ItemUnit;
use Modules\EInvoice\Entities\Settings;
use Illuminate\Support\Facades\Validator;

trait SetDocumentTrait
{

    use AuthTrait;

    /**
     * This Method allows taxpayer systems to retrieve list of document types published by the eInvoicing solution.
     *
     * @return data for document types
     */
    public function getDocumentTypes()
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1/documenttypes')
                ->withBearer($this->token)
                ->withHeader('Accept-Language', app()->getLocale())
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getDocumentTypes: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayer systems to query latest documents sent or received that are available on the eInvoicing solution.
     *
     * @return documents
     */
    public function getRecentDocuments($pageSize = 20, $pageNumber = 1)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documents/recent?pageNo=' . $pageNumber . '&pageSize=' . $pageSize)
                ->withBearer($this->token)
                ->withContentType('application/json')
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getRecentDocuments: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method returns information on documents submitted during a single submission by taxpayer.
     *
     * @param submitionUUID
     * @return document
     */
    public function getSubmittedDocument($submitionUUID, $pageSize = 10, $pageNumber = 1)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documentSubmissions/' . $submitionUUID . '?PageSize=' . $pageSize . '&PageNumber=' . $pageNumber)
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            // Retry if the response is not successful.
            while (!in_array($response->status, [200, 201])) {
                sleep(1);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/documentSubmissions/' . $submitionUUID . '?PageSize=' . $pageSize . '&PageNumber=' . $pageNumber)
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->get();
            }

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getSubmittedDocument: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers to retrieve document.
     *
     * @param documentUUID
     * @return document
     */
    public function getDocument($documentUUID)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/raw')
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            // Retry if the request fails.
            $count = 0;
            while ($count <= 5 && !in_array($response->status, [200, 201])) {
                sleep(1);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/raw')
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->get();
                $count++;
            }

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getDocument: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method enables ERP system to download the PDF version of the document having predefined layout.
     *
     * @param documentUUID
     * @return document as PDF
     */
    public function getDocumentPDF($documentUUID)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Dowanload Path
            if (!file_exists(storage_path('app/modules/einvoice/documents'))) {
                mkdir(storage_path('app/modules/einvoice/documents'), 0777, true);
            }

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/pdf')
                ->withBearer($this->token)
                ->withContentType('application/pdf')
                ->returnResponseObject()
                ->download(storage_path('app/modules/einvoice/documents/' . $documentUUID . '.pdf'));

            $data = $response->content;

            // Data from the response
            return $response->status == 200
                ? response($data, $response->status)->header('Content-Type', 'application/pdf')
                : response($data['error'], $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getDocumentPDF: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allows taxpayers to retrieve a single document full details including validation results.
     *
     * @param documentUUID
     * @return document
     */
    public function getDocumentDetails($documentUUID)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/details')
                ->withBearer($this->token)
                ->withTimeout(10)
                ->withConnectTimeout(60)
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            $count = 0;
            while ($count <= 5 && !in_array($response->status, [200, 201])) {
                sleep(1);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1/documents/' . $documentUUID . '/details')
                    ->withBearer($this->token)
                    ->withTimeout(10)
                    ->withConnectTimeout(60)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->get();
                $count++;
            }

            $data = $response->content;

            // Data from the response
            return  response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getDocumentDetails: ' . $e->getMessage());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method to set Document to be submitted.
     *
     * @param bills
     * @return array
     */
    public function setDocument($bills)
    {
        // All the bills to be submitted.
        $documents = collect();

        // Check if the bills are array or not.
        $bills = is_array($bills) ? $bills : collect($bills);

        foreach ($bills as $bill) {

            // All the data used to create the document
            $invoice = Invoice::where('id', $bill)->first();
            $settings          = Settings::shop()->first();
            $merchant          = Clients::where('id', $invoice->add_user)->first();
            $buyer             = Clients::where('id', $invoice->client_id)->first();
            $vats              = collect($invoice->vats->toArray())->map(function ($vat) {
                $vat['add'] = Vat::where('id', $vat['vat_id'])->first()->toArray();
                return $vat;
            });

            // Validate the bill
            $validation = $this->validateInvoice($invoice, $settings, $merchant, $buyer, $vats);

            if (is_array($validation) && count($validation)) {
                return ['status' => 'validation', 'errors' => $validation];
            }

            // Issuer Data.
            $issuer = [
                'address' => [
                    "branchID"              => $settings->branch_id,
                    "country"               => $merchant->country->iso_alpha2,
                    "governate"             => $merchant->governate,
                    "regionCity"            => $merchant->regionCity,
                    "street"                => $merchant->street,
                    "buildingNumber"        => (string) $merchant->buildingNumber,
                    "postalCode"            => (string) $merchant->postalCode,
                    "floor"                 => (string) $merchant->floor,
                    "room"                  => (string) $merchant->room,
                    "landmark"              => $merchant->landmark,
                ],
                'type' => "B",
                'id'   => $settings->registration_id,
                'name' => $settings->company_name,
            ];

            // If address has additional data
            if (!is_null($merchant->additional_information)) {
                $issuer['address']['additionalInformation'] = $merchant->additional_information;
            }

            // Buyer Data
            $receiver = [
                'type' => in_array($buyer->client_role, ['1', '2']) ? 'B' : 'P',
                'name' => $buyer->client_name
            ];

            // Push the data to receiver array if person buyer and amount less than threshold limit defined (50000).
            if (!in_array($receiver['type'], ['P', 'F']) || $invoice['net_price'] >= 50000) {

                $receiver['id'] = in_array($receiver['type'], ['P', 'F']) ? $buyer->national_id : $buyer->client_tax_number;

                $receiver['address'] = [
                    "country"               => $buyer->country->iso_alpha2,
                    "governate"             => $buyer->governate,
                    "regionCity"            => $buyer->regionCity,
                    "street"                => $buyer->street,
                    "buildingNumber"        => (string) $buyer->buildingNumber,
                    "postalCode"            => (string) $buyer->postalCode,
                    "floor"                 => (string) $buyer->floor,
                    "room"                  => (string) $buyer->room,
                    "landmark"              => $buyer->landmark,
                ];

                if (!is_null($buyer->additional_information)) {
                    $receiver['address']['additionalInformation'] = $buyer->additional_information;
                }
            }

            // Invoice Items
            $invoiceLines             = collect();
            $taxes                    = collect();
            $totalDiscount            = 0;
            $totalSalesAmount         = 0;
            $netAmount                = 0;
            $totalAmount              = 0;
            $extraDiscountAmount      = 0;
            $totalItemsDiscountAmount = 0;

            // Get the items from the invoice.
            $details = $this->invoiceDetails($invoice->details, $invoice, $totalSalesAmount, $totalItemsDiscountAmount, $settings, $totalDiscount, $invoiceLines, $taxes);

            // Total Tax Amount.
            $totalTaxes = $taxes->count() > 0 ? $this->invoiceTaxTotal($taxes) : 0;

            // Sum of all tax amounts.
            $totalTaxAmount = $totalTaxes['totalTaxAmount'] ?? 0  + $details['totalTaxableFees'];

            // Total amount of total sales after applying discount.
            $netAmount = $details['totalSalesAmount'] - $details['totalDiscount'];

            // Total amount of the invoice calculated as NetAmount + Totals of tax amounts.
            $totalAmount = $netAmount + $totalTaxAmount;

            if ($invoice->invoice_type == 1) {
                $documentType = "I";
            } elseif ($invoice->invoice_type == 3) {
                $documentType = "C";
            }

            // Invoice Data
            $document = [
                "issuer"                   => $issuer,
                "receiver"                 => $receiver,
                "documentType"             => $documentType,
                "documentTypeVersion"      => "0.9",
                "dateTimeIssued"           => now()->format('Y-m-d\TH:i:s\Z'),
                "taxpayerActivityCode"     => $settings->activity_id,
                "internalID"               => Str::random(12),
                "invoiceLines"             => $invoiceLines,
                "totalDiscountAmount"      => einvoice_decimal($details['totalDiscount']),
                "totalSalesAmount"         => einvoice_decimal($details['totalSalesAmount']),
                "netAmount"                => einvoice_decimal($netAmount),
                "totalAmount"              => einvoice_decimal($totalAmount),
                "extraDiscountAmount"      => einvoice_decimal($extraDiscountAmount),
                "totalItemsDiscountAmount" => einvoice_decimal($details['totalItemsDiscountAmount']),
            ];

            // If the item has a vat.
            if ($totalTaxes > 0) {
                $document['taxTotals'] = $totalTaxes['taxTotals'];
            }

            // Add the data to the document
            $documents->put($bill, $document);

        }

        return $documents;
    }

    /**
     * This Method is used to validate the parameters of the invoice.
     *
     * @param $merchant
     * @param $buyer
     * @param $invoice
     * @param $settings
     * @return response
     */
    public function validateInvoice($invoice, $settings, $merchant, $buyer, $vats)
    {
        $request  = [
            "merchant" => optional($merchant)->toArray() ?? [],
            "buyer"    => optional($buyer)->toArray() ?? [],
            "settings" => optional($settings)->toArray() ?? [],
            "invoice"  => optional($invoice)->toArray() ?? [],
            "vats"     => optional($vats)->toArray() ?? [],
        ];

        $rules = [
            "merchant"                        => ['required', 'array'],
            "merchant.governate"              => ['required', 'string', new NotNull],
            "merchant.regionCity"             => ['required', 'string', new NotNull],
            "merchant.street"                 => ['required', 'string', new NotNull],
            "merchant.buildingNumber"         => ['required', 'numeric', new NotNull],
            "merchant.postalCode"             => ['required', 'numeric', new NotNull],
            "merchant.floor"                  => ['required', 'numeric', new NotNull],
            "merchant.room"                   => ['required', 'numeric', new NotNull],
            "merchant.landmark"               => ['required', 'string', new NotNull],
            "merchant.additional_information" => ['nullable', 'string', new NotNull],
            "buyer"                           => ['required', 'array'],
            "buyer.governate"                 => ['required', 'string', new NotNull],
            "buyer.regionCity"                => ['required', 'string', new NotNull],
            "buyer.street"                    => ['required', 'string', new NotNull],
            "buyer.buildingNumber"            => ['required', 'numeric', new NotNull],
            "buyer.postalCode"                => ['required', 'numeric', new NotNull],
            "buyer.floor"                     => ['required', 'numeric', new NotNull],
            "buyer.room"                      => ['required', 'numeric', new NotNull],
            "buyer.landmark"                  => ['required', 'string', new NotNull],
            "buyer.additional_information"    => ['nullable', 'string', new NotNull],
            "settings"                        => ['required', 'array'],
            "settings.company_name"           => ['required', 'string', new NotNull],
            "settings.activity_id"            => ['required', 'string', new NotNull],
            "settings.branch_id"              => ['required', 'string', new NotNull],
            "settings.country_id"             => ['required', 'string', new NotNull],
            "settings.currency_id"            => ['required', 'string', new NotNull],
            "settings.registration_id"        => ['required', 'string', new NotNull],
            "settings.client_id"              => ['required', 'string', new NotNull],
            "settings.client_secret"          => ['required', 'string', new NotNull]
        ];

        $messages = [
            "merchant.governate.required"           => __('messages.merchant.governate_required'),
            "merchant.regionCity.required"          => __('messages.merchant.regionCity_required'),
            "merchant.street.required"              => __('messages.merchant.street_required'),
            "merchant.buildingNumber.required"      => __('messages.merchant.buildingNumber_required'),
            "merchant.postalCode.required"          => __('messages.merchant.postalCode_required'),
            "merchant.floor.required"               => __('messages.merchant.floor_required'),
            "merchant.room.required"                => __('messages.merchant.room_required'),
            "merchant.landmark.required"            => __('messages.merchant.landmark_required'),
            "buyer.governate.required"              => __('messages.buyer.governate_required'),
            "buyer.regionCity.required"             => __('messages.buyer.regionCity_required'),
            "buyer.street.required"                 => __('messages.buyer.street_required'),
            "buyer.buildingNumber.required"         => __('messages.buyer.buildingNumber_required'),
            "buyer.postalCode.required"             => __('messages.buyer.postalCode_required'),
            "buyer.floor.required"                  => __('messages.buyer.floor_required'),
            "buyer.room.required"                   => __('messages.buyer.room_required'),
            "buyer.landmark.required"               => __('messages.buyer.landmark_required'),
            "settings.company_name.required"        => __('messages.settings.company_name_required'),
            "settings.activity_id.required"         => __('messages.settings.activity_id_required'),
            "settings.branch_id.required"           => __('messages.settings.branch_id_required'),
            "settings.country_id.required"          => __('messages.settings.country_id_required'),
            "settings.currency_id.required"         => __('messages.settings.currency_id_required'),
            "settings.registration_id.required"     => __('messages.settings.registration_id_required'),
            "settings.client_id.required"           => __('messages.settings.client_id_required'),
            "settings.client_secret.required"       => __('messages.settings.client_secret_required'),
        ];

        // Check if vats count is greater than 0 then add the rules.
        if ($vats->count() > 0) {
            $rules["vats"]                         = ['required', 'array'];
            $rules["vats.*.add"]                   = ['required', 'array'];
            $rules["vats.*.vat_id"]                = ['required', 'numeric', new NotNull];
            $rules["vats.*.vat_type"]              = ['required', 'numeric', new NotNull];
            $rules["vats.*.vat_value"]             = ['required', 'numeric', new NotNull];
            $rules["vats.*.add.einvoice_code"]     = ['required', 'string', new NotNull];
            $rules["vats.*.add.sub_einvoice_code"] = ['required', 'string', new NotNull];

            $messages["vats.required"]                         = __('messages.vats.required');
            $messages["vats.*.add.required"]                   = __('messages.vats_add_required');
            $messages["vats.*.vat_id.required"]                = __('messages.vats_vat_id_required');
            $messages["vats.*.vat_type.required"]              = __('messages.vats_vat_type_required');
            $messages["vats.*.vat_value.required"]             = __('messages.vats_vat_value_required');
            $messages["vats.*.add.einvoice_code.required"]     = __('messages.vats_add_einvoice_code_required');
            $messages["vats.*.add.sub_einvoice_code.required"] = __('messages.vats_add_sub_einvoice_code_required');
        }

        $validator = Validator::make($request, $rules, $messages);
        return $validator->fails() ? $validator->errors()->all() : [];
    }

    /**
     * This Method to get the invoice details.
     *
     * @param invoiceDetails
     * @param invoice
     * @param totalSalesAmount
     * @param totalItemsDiscountAmount
     * @param settings
     * @param totalDiscount
     * @param invoiceLines
     * @param taxes
     * @return array
     */
    public function invoiceDetails($invoiceDetails, $invoice, $totalSalesAmount, $totalItemsDiscountAmount, $settings, $totalDiscount, $invoiceLines, $taxes)
    {
        foreach ($invoiceDetails as $item_) {

            // Get the item
            $item = Item::where('id', $item_->item_id)->first();

            /**
             * Get the document type
             * (invoice_type = 0) => فاتورة شراء , (invoice_type = 1) => فاتورة بيع
             * (invoice_type = 2) => فاتورة مرتجع شراء , (invoice_type = 3) => فاتورة مرتجع بيع
             */
            $item_price = in_array($invoice->invoice_type, [0, 2]) ? $item_->basic_pay_price : $item_->basic_sale_price;

            /**
             * Item Discount
             * (discount_type = 0) => نسبة , (discount_type = 1) => مبلغ
             * (discount_operation = 0) => تخفيض , (discount_operation = 1) => زيادة
             */
            $discount         = 0;
            $discount_percent = 0;

            if ($item_->discount > 0) {
                // $discount         = (($item_->discount_type == 1) ? $item_->discount : $item_->discount * $item_price) * $item_->quantity;
                // $discount_percent = ($item_->discount_type == 0) ? $item_->discount * 100 : ($item_->discount / $item_price) * 100;
                $discount         = (($item_->discount_type == 1) ? $item_->discount : ($item_->discount * $item_price) / 100) * $item_->quantity;
                $discount_percent = ($item_->discount_type == 0) ? $item_->discount : ($item_->discount / $item_price) * 100;
            }

            // dd([
            //     'discount_type'     => $item_->discount_type,
            //     'discount'          => $item_->discount,
            //     'item_price'        => $item_price,
            //     'discount_'         => $discount,
            //     'discount_percent_' => $discount_percent,
            //     'quantity'          => $item_->quantity,
            // ]);

            // Item Unit Price
            if ($item_->unit_id != $item->unit_id) {
                $unit       = ItemUnit::where('unit_id', $item_->unit_id)->where('item_id', $item_->item_id)->first();
                $item_price = $unit->unit_value * $item_price;
            } else {
                $unit = Unit::where('id', $item_->unit_id)->first();
            }

            // net price is the unit price - discount
            $net_price  = ($item_price * $item_->quantity) - $discount;

            $taxableItems     = collect();
            $totalTaxableFees = 0;
            $invoiceVats      = $invoice->vats->where('item_id', $item_->item_id);

            // Get item vats.
            $totalItemTax = count($invoiceVats) > 0
                ? $this->invoiceVats($item_, $invoiceVats, $taxableItems, $totalTaxableFees, $taxes, $net_price, $item_->quantity)
                : 0;

            // Calculate the total amount of the item.
            $itemTotal = $this->invoiceTotal($item_price, $item_->quantity, $discount, $totalItemTax);

            // Total amount of all the items sales.
            $totalSalesAmount += $itemTotal['salesTotal'];

            // Non-taxable items discount.
            $itemsDiscount = $item_->items_discount ?? 0;

            // Total amount of all the items discount.
            $totalItemsDiscountAmount += $itemsDiscount;

            // Add the item to the invoice items.
            $invoiceItems = [
                "description"      => $item->about,
                "itemType"         => "EGS",
                "itemCode"         => $item->item_einvoice_code,
                "unitType"         => $unit->unit_einvoice_code,
                "quantity"         => (float) $item_->quantity,
                "salesTotal"       => einvoice_decimal($itemTotal['salesTotal']),
                "total"            => einvoice_decimal($itemTotal['total']),
                "valueDifference"  => einvoice_decimal($itemTotal['valueDifference']),
                "totalTaxableFees" => einvoice_decimal($totalTaxableFees),
                "netTotal"         => einvoice_decimal($itemTotal['netTotal']),
                "itemsDiscount"    => einvoice_decimal($itemsDiscount),
                "unitValue" => [
                    "currencySold" => $settings->currency_id,
                    "amountEGP"    => einvoice_decimal($item_price),
                ],
            ];

            // If the item has discount.
            if (isset($discount)) {
                $invoiceItems['discount'] = [
                    "rate"   => einvoice_decimal($discount_percent),
                    "amount" => einvoice_decimal($discount),
                ];
            }

            // if currency is not equal to EGP.
            if ($invoiceItems['unitValue']['currencySold'] != 'EGP') {
                $invoiceItems['unitValue']['amountSold'] = (float) $item_price;
            }

            // Calculate the total discount amount.
            $totalDiscount += $invoiceItems['discount']['amount'];

            // Add the taxable items to the invoice items if item has vats.
            if ($taxableItems->count() > 0) {
                $invoiceItems['taxableItems'] = $taxableItems;
            }

            $invoiceLines->push($invoiceItems);
        }

        return [
            'totalTaxableFees'         => einvoice_decimal($totalTaxableFees),
            'totalSalesAmount'         => einvoice_decimal($totalSalesAmount),
            'totalDiscount'            => einvoice_decimal($totalDiscount),
            'totalItemsDiscountAmount' => einvoice_decimal($totalItemsDiscountAmount),
        ];
    }

    /**
     * This Method to get the item vats.
     *
     * @param item_ item in the invoice
     * @param vats
     * @param taxableItems
     * @param totalTaxableFees
     * @param taxes
     * @param net_price
     * @return totalItemTax
     */
    public function invoiceVats($item_, $vats, $taxableItems, $totalTaxableFees, $taxes, $net_price, $quantity)
    {
        foreach ($vats as $vat) {

            // Get the vat from the bills_add
            $vat_ = Vat::where('id', $vat->vat_id)->first();

            /**
             * Add the vat to the taxable items
             * (vat_type = 0) => نسبة , (vat_type = 1) => مبلغ
             * (vat_operation = 0) => إضافة , (vat_operation = 1) => خصم
             */
            $rate   = $vat->vat_type == 0 ? $vat->vat_value : ($vat->vat_value / $net_price) * 100;
            $amount = ($vat->vat_type == 1 ? $vat->vat_value : ($vat->vat_value * $net_price) / 100);
            // dd([
            //     'vat_type' => $vat->vat_type,
            //     'vat_value' => $vat->vat_value,
            //     'net_price' => $net_price,
            //     'rate'   => $rate,
            //     'amount' => $amount,
            // ]);

            // Check if the vat is already added to the taxable items.
            $allTaxableItem = $taxableItems->pluck('taxType')->toArray();
            $allTaxes       = $taxes->pluck('taxType')->toArray();

            if (!in_array($vat_->einvoice_code, $allTaxableItem)) {
                $taxableItems->push([
                    "taxType" => $vat_->einvoice_code,
                    "rate"    => einvoice_decimal($rate),
                    "amount"  => einvoice_decimal($amount),
                    "subType" => $vat_->sub_einvoice_code
                ]);
            }

            // echo $taxableItems;
            if (!in_array($vat_->einvoice_code, $allTaxes)) {
                $taxes->push([
                    "taxType" => $vat_->einvoice_code,
                    "amount"  => einvoice_decimal($amount),
                ]);
            } else {
                $tax            = $taxes->where('taxType', $vat_->einvoice_code)->first();
                $key            = $taxes->search($tax);

                $taxes->forget($key);
                $taxes->push([
                    "taxType" => $vat_->einvoice_code,
                    "amount"  => einvoice_decimal($tax['amount'] + $amount),
                ]);
            }

            // Total amount of additional taxable fees to be used in final tax calculation.
            $totalTaxableFees += $item_->additional_taxable_fees ?? 0;
        }

        // Total item tax.
        $totalItemTax = $taxableItems->sum('amount') + $totalTaxableFees;

        return einvoice_decimal($totalItemTax);
    }

    /**
     * This Method is used to calculate the tax totals.
     *
     * @param taxes of the invoice
     * @return taxTotals
     */
    public function invoiceTaxTotal($taxes)
    {
        $taxTotals = collect();

        // Calculate the tax totals amount.
        $taxes->groupBy('taxType')->each(function ($taxes, $key) use ($taxTotals) {
            $taxTotals->push([
                "taxType" => $key,
                "amount"  => einvoice_decimal($taxes->sum('amount')),
            ]);
        });

        // Calculate the sum of all the tax totals.
        $totalTaxAmount = $taxTotals->sum('amount');

        return [
            "taxTotals"      => $taxTotals,
            "totalTaxAmount" => einvoice_decimal($totalTaxAmount),
        ];
    }

    /**
     * This Method is used to calculate totals.
     *
     * @param item_price
     * @param quantity
     * @param totalTaxableFees
     * @param discount
     * @return itemTotal
     */
    public function invoiceTotal($item_price, $quantity, $discount, $totalItemTax)
    {
        // Total amount considering quantity and unit price.
        $salesTotal = $item_price * $quantity;

        // Total amount after adding all pricing items, taxes, removing discounts.
        $total = $salesTotal + $totalItemTax - $discount;

        // Value difference when selling goods already taxed.
        $valueDifference = $total - $salesTotal;

        // Total amount for the invoice line after applying discount.
        $netTotal = $total - $totalItemTax;

        return [
            "salesTotal"      => einvoice_decimal($salesTotal),
            "total"           => einvoice_decimal($total),
            "valueDifference" => einvoice_decimal(0 ?? $valueDifference),
            "netTotal"        => einvoice_decimal($netTotal),
        ];
    }
}
