<?php

namespace App\Traits\Einvoice;

use Ixudra\Curl\Facades\Curl;
use App\Traits\Einvoice\AuthTrait;
use Illuminate\Support\Facades\Log;

trait ItemCodeTrait
{
    use AuthTrait;

    /**
     * This Method allow taxpayers to create their own codes in the system (EGS Codes).
     *
     * @param array $data
     */
    public function createEgsCode($data)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            $body = [
                "items" => $data,
            ];

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/requests/codes')
                ->withData($body)
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->post();

            // Retry if the response is not successful
            while (!in_array($response->status, [200, 201])) {
                sleep(1);
                Log::channel('einvoice')->error('Retrying to create EGS Code , Error: ' . $response->content . ' Status: ' . $response->status);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/requests/codes')
                    ->withData($body)
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->post();
            }

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in createEgsCode Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to update their own codes in the system (EGS Codes).
     *
     * @param string $itemCode
     * @param array $body
     * @param string $codeType
     * @return response
     */
    public function updateCode($itemCode, $body, $codeType = 'EGS')
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/' . $codeType . '/codes/' . $itemCode)
                ->withData($body)
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->put();

            // Retry if the response is not successful
            while (!in_array($response->status, [200, 201])) {
                sleep(1);
                Log::channel('einvoice')->error('Retrying to update code: ' . $itemCode . ' Error: ' . $response->content . ' Status: ' . $response->status);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/' . $codeType . '/codes/' . $itemCode)
                    ->withData($body)
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->put();
            }

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in updateCode Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayer to Update his previous code usage requests as long they are under ‘Submitted’ state.
     *
     * @param string $codeUsageRequestID
     * @param array $body
     * @param string $codeType
     * @return response
     */
    public function updateEgsCode($codeUsageRequestID, $body)
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/requests/codes/' . $codeUsageRequestID)
                ->withData($body)
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->put();

            // Retry if the response is not successful
            while (!in_array($response->status, [200, 201])) {
                sleep(1);
                Log::channel('einvoice')->error('Retrying to update EGS code , Error: ' . $response->content . ' Status: ' . $response->status);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/requests/codes/' . $codeUsageRequestID)
                    ->withData($body)
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->put();
            }

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in updateEgsCode Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }


    /**
     * This Method allow taxpayers to get code details from the system (EGS Codes).
     *
     * @param string $codeType
     * @param string $itemCode
     * @return response
     */
    public function getEgsCodeDetails($itemCode, $codeType = 'EGS')
    {
        try {
            // Login to the Taxpayer API
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Send a request to the Taxpayer API
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/' . $codeType . '/codes/' . $itemCode)
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->get();

                // Retry if the response is not successful
                while ($response->status != 200) {
                    sleep(1);
                    Log::channel('einvoice')->error('Retrying to get EGS Code Details , Error: ' . $response->content . ' Status: ' . $response->status);
                    $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/' . $codeType . '/codes/' . $itemCode)
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->get();
                }

            $data = $response->content;

            // Data from the response
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in getEgsCodeDetails Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to search for codes in the system (EGS Codes).
     *
     * @param string $query
     * @return response
     */
    public function searchEgsCode($query)
    {
        try {
            // Login to the Taxpayer API.
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Check if there is a query or not.
            $url = in_array($query, [' ', '', null])
                ? self::$apiBaseUrl . '/api/v1.0/codetypes/requests/my'
                : self::$apiBaseUrl . '/api/v1.0/codetypes/requests/my?' . $query;

            // Send a request to the Taxpayer API.
            $response = Curl::to($url)
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            // Retry if the response is not successful.
            while ($response->status != 200) {
                sleep(1);
                Log::channel('einvoice')->error('Retrying to get EGS Codes , Error: ' . $response->content . ' Status: ' . $response->status);
                $response = Curl::to($url)
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->get();
            }

            $data = $response->content;

            // Data from the response.
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in searchEgsCode Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to request code reusue from current existing code.
     *
     * @param array $data
     * @return response
     */
    public function requestCodeReuse($data)
    {
        try {
            // Login to the Taxpayer API.
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            $body = [
                "items" => $data,
            ];

            // Send a request to the Taxpayer API.
            $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/requests/codeusages')
                ->withData($body)
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->put();

            // Retry if the response is not successful.
            while (!in_array($response->status, [200, 201])) {
                sleep(1);
                Log::channel('einvoice')->error('Retrying to request code reuse , Error: ' . $response->content . ' Status: ' . $response->status);
                $response = Curl::to(self::$apiBaseUrl . '/api/v1.0/codetypes/requests/codeusages')
                    ->withData($body)
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->put();
            }

            $data = $response->content;

            // Data from the response.
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in requestCodeReuse Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }

    /**
     * This Method allow taxpayers to search EGS and GS1 published codes in the solution.
     *
     * @param string $query
     * @param string $codeType
     * @return response
     */
    public function searchPublishedCodes($query, $codeType)
    {
        try {
            // Login to the Taxpayer API.
            $this->token = in_array($this->token, [null, '']) ? $this->loginTaxpayer() : $this->token;

            // Check if there is a query or not.
            $url = in_array($query, [' ', '', null])
                ? self::$apiBaseUrl . '/api/v1.0/codetypes/' . $codeType . '/codes'
                : self::$apiBaseUrl . '/api/v1.0/codetypes/' . $codeType . '/codes?' . $query;

            // Send a request to the Taxpayer API.
            $response = Curl::to($url)
                ->withBearer($this->token)
                ->returnResponseObject()
                ->asJson(true)
                ->get();

            // Retry if the response is not successful.
            while ($response->status != 200) {
                sleep(1);
                Log::channel('einvoice')->error('Retrying to get published codes , Error: ' . $response->content . ' Status: ' . $response->status);
                $response = Curl::to($url)
                    ->withBearer($this->token)
                    ->returnResponseObject()
                    ->asJson(true)
                    ->get();
            }

            $data = $response->content;

            // Data from the response.
            return response($data, $response->status);
        } catch (\Exception $e) {
            Log::channel('einvoice')->error('Error in searchPublishedCodes Method: ' . $e->getMessage() . ' Status: ' . $e->getCode());
            return response($e->getMessage(), 500);
        }
    }
}
