<?php

namespace App\Traits\Einvoice;

use Illuminate\Support\Str;
use Ixudra\Curl\Facades\Curl;
use App\Traits\Einvoice\AuthTrait;
use Modules\EInvoice\Entities\Vat;
use Modules\EInvoice\Entities\Item;
use Modules\EInvoice\Entities\Unit;
use Modules\EInvoice\Entities\Clients;
use Modules\EInvoice\Entities\Invoice;
use Modules\EInvoice\Entities\ItemUnit;
use Modules\EInvoice\Entities\Settings;

trait ReceiptDocumentTrait
{
    use AuthTrait;

    /**
     * This Method allows taxpayers to retrieve receipt/return receipt source JSON with additional tax authority added metadata.
     *
     * @param $receiptUUID
     * @return $data
     */
    public function getReceipt($receiptUUID)
    {
        // Login to POS API
        $this->posToken = in_array($this->posToken, [null, '']) ? $this->authenticatePOS() : $this->posToken;

        // Send a request to the POS API
        $response = Curl::to(self::$apiBaseUrl . '/api/v1/receipts/' . $receiptUUID . '/raw')
            ->withBearer($this->posToken)
            ->returnResponseObject()
            ->asJson(true)
            ->get();

        $data = $response->content;

        // Data from the response
        return $data;
    }

    /**
     * This Method allows taxpayers to retrieve a single receipt/return receipt full details including validation results.
     *
     * @param $receiptUUID
     * @return $data
     */
    public function getReceiptDetails($receiptUUID)
    {
        // Login to POS API
        $this->posToken = in_array($this->posToken, [null, '']) ? $this->authenticatePOS() : $this->posToken;

        // Send a request to the POS API
        $response = Curl::to(self::$apiBaseUrl . '/api/v1/receipts/' . $receiptUUID . '/details')
            ->withBearer($this->posToken)
            ->returnResponseObject()
            ->asJson(true)
            ->get();

        $data = $response->content;

        // Data from the response
        return $data;
    }

    /**
     * This Method allows taxpayers to retrieve document source JSON with additional tax authority added metadata.
     *
     * @param $receiptUUID
     * @param $dateTimeIssued
     * @return $data
     */
    public function getReceiptDetailsAnonymously($receiptUUID, $dateTimeIssued)
    {
        // Send a request to the POS API
        $response = Curl::to(self::$apiBaseUrl . '/api/v1/receipts/' . $receiptUUID . '/share?dateTimeIssued=' . $dateTimeIssued)
            ->returnResponseObject()
            ->asJson(true)
            ->get();

        $data = $response->content;

        // Data from the response
        return $data;
    }

    /**
     * This Method returns information on receipts submitted during a single submission by taxpayer.
     *
     * @param $submitionUUID
     * @param $pageSize
     * @param $pageNumber
     * @return $data
     */
    public function getSubmittedReceipts($submitionUUID, $pageSize = 10, $pageNumber = 1)
    {
        // Login to POS API
        $this->posToken = in_array($this->posToken, [null, '']) ? $this->authenticatePOS() : $this->posToken;

        // Send a request to the POS API
        $response = Curl::to(self::$apiBaseUrl . '/api/v1/receiptsubmissions/' . $submitionUUID . '/details?pageSize=' . $pageSize . '&pageNo=' . $pageNumber)
            ->withBearer($this->posToken)
            ->returnResponseObject()
            ->asJson(true)
            ->get();

        $data = $response->content;

        // Data from the response
        return $data;
    }

    /**
     * This Method uses to set the receipt to be submitted.
     *
     * @param $bills
     * @return $receipts
     */
    public function setReceipt($bills)
    {
        // All the bills to be submitted.
        $documents = collect();

        // Check if the bills are array or not.
        $bills = gettype($bills) == "array" ? $bills : collect($bills);

        foreach ($bills as $bill) {
            // All the data used to create the document
            $invoice  = Invoice::where('id', $bill)->first();
            $settings = Settings::shop()->first();
            $seller   = Clients::where('id', $invoice->add_user)->first();
            $customer = Clients::where('id', $invoice->client_id)->first();

            // Seller Data.
            $seller = [
                "rin"              => $settings->registration_id,
                "companyTradeName" => $settings->company_name,
                "branchCode"       => $settings->branch_id,
                "branchAddress"    => [
                    "country"               => $seller->country->iso_alpha2,
                    "governate"             => $seller->governate,
                    "regionCity"            => $seller->regionCity,
                    "street"                => $seller->street,
                    "buildingNumber"        => (string) $seller->buildingNumber,
                    "postalCode"            => (string) $seller->postalCode,
                    "floor"                 => (string) $seller->floor,
                    "room"                  => (string) $seller->room,
                    "landmark"              => $seller->landmark,
                ],
                "deviceSerialNumber" => $settings->device_serial_number,
                "activityCode"       => $settings->activity_id,
            ];

            if (isset($settings->syndicate_license_number)) {
                $seller['syndicateLicenseNumber'] = $settings->syndicate_license_number;
            }

            // Buyer Data.
            $buyer = [
                "type" => in_array($customer->client_role, ['1', '2']) ? 'B' : 'P',
            ];

            // if the customer has name add it to the buyer.
            if (isset($customer->client_name)) {
                $buyer['name'] = $customer->client_name;
            }

            // if the customer is a person or foreigner add his/her identity to the buyer.
            if (!in_array($buyer['type'], ['P', 'F']) || $invoice['net_price'] >= 50000) {
                $buyer['id'] = in_array($customer['type'], ['P', 'F']) ? $customer->national_id : $customer->client_tax_number;
            }

            // if the customer has mobile add it to the buyer.
            if (isset($customer->mobileNumber)) {
                $buyer['mobileNumber'] = $customer->mobileNumber;
            }

            // if the customer has phone add it to the buyer.
            if (isset($customer->paymentNumber)) {
                $buyer['paymentNumber'] = $customer->paymentNumber;
            }

            // Receipt Header.
            $header = [
                "dateTimeIssued"   => now()->format('Y-m-d\TH:i:s\Z'),
                "receiptNumber"    => Str::random(10),
                "uuid"             => "b4c08c8b273c45d6b824b359d87d957345b0ec6859eb4ad7b68221244ca55d86",
                "previousUUID"     => "0",
                "referenceOldUUID" => "b4c08c8b273c45d6b824b359d87d957345b0ec6859eb4ad7b68221244ca55d86",
                "currency"         => $settings->currency_id,
            ];

            // Receipt Document Type.
            $documentType = [
                "receiptType" => "S",
                "typeVersion" => "1.0",
            ];

            // Receipt Items.
            $itemData                = collect();
            $taxes                   = collect();
            $totalSales              = 0;
            $totalCommercialDiscount = 0;
            $totalItemsDiscount      = 0;
            $extraReceiptDiscount    = 0;
            $netAmount               = 0;
            $feesAmount              = 0;
            $totalAmount             = 0;

            // Get the items from the invoice.
            $details = $this->receiptDetails($invoice->details, $invoice, $taxes, $itemData, $totalSales, $totalCommercialDiscount, $totalItemsDiscount, $extraReceiptDiscount, $netAmount, $feesAmount, $totalAmount);

            // Total Tax Amount.
            $totalTaxes = $taxes->count() > 0 ? $this->receiptTaxTotal($taxes) : 0;

            // Sum of all tax amounts.
            $totalTaxAmount = $totalTaxes['totalTaxAmount'] ?? 0;

            // Total amount of total sales after applying discount.
            $netAmount = $details['netAmount'] ?? 0;

            // Total amount of the invoice calculated as NetAmount + Totals of tax amounts.
            $totalAmount = $netAmount + $totalTaxAmount;

            // Receipt Document.
            $document = [
                "header"                  => $header,
                "documentType"            => $documentType,
                "seller"                  => $seller,
                "buyer"                   => $buyer,
                "itemData"                => $itemData,
                "totalSales"              => (float) $details['totalSales'],
                "netAmount"               => (float) $details['netAmount'],
                "feesAmount"              => (float) $details['feesAmount'],
                "totalAmount"             => (float) $details['totalAmount'],
                "paymentMethod"           => "Cash",
            ];

            // if items have discount then add it to the document.
            if (isset($details['totalItemsDiscount']) && $details['totalItemsDiscount'] > 0) {
                $document['totalItemsDiscount'] = $details['totalItemsDiscount'];
            }

            // if there is any commercial discount then add it to the document.
            if (isset($details['totalCommercialDiscount']) && $details['totalCommercialDiscount'] > 0) {
                $document['totalCommercialDiscount'] = $details['totalCommercialDiscount'];
            }

            // if extra discount is added then add it to the document.
            if (isset($details['extraReceiptDiscount']) && $details['extraReceiptDiscount'] > 0) {
                $document['extraReceiptDiscount'] = $details['extraReceiptDiscount'];
            }

            // If the item has a vat.
            if ($totalTaxes > 0) {
                $document['taxTotals'] = $totalTaxes['taxTotals'];
            }

            // Add the data to the document
            $documents->put($bill, $document);
        }

        // Return the documents.
        return $documents;
    }

    /**
     * This Method uses to get the invoice details.
     *
     * @param details
     * @param invoice
     * @param taxes
     * @param receiptItem
     * @return array
     */
    public function receiptDetails($receiptDetails, $invoice, $taxes, $itemData, $totalSales, $totalCommercialDiscount, $totalItemsDiscount, $extraReceiptDiscount, $netAmount, $feesAmount, $totalAmount)
    {
        // Details of the invoice that will be used to generate the receipt.
        $details = [];

        // Get the items from the invoice.
        foreach ($receiptDetails as $item_) {
            // Get the item
            $item = Item::where('id', $item_->item_id)->first();

            /**
             * Get the document type
             * (invoice_type = 0) => فاتورة شراء , (invoice_type = 1) => فاتورة بيع
             * (invoice_type = 2) => فاتورة مرتجع شراء , (invoice_type = 3) => فاتورة مرتجع بيع
             */
            $item_price = in_array($invoice->invoice_type, [0, 2]) ? $item_->basic_pay_price : $item_->basic_sale_price;

            /**مم
             * Item Discount
             * (discount_type = 0) => نسبة , (discount_type = 1) => مبلغ
             * (discount_operation = 0) => تخفيض , (discount_operation = 1) => زيادة
             */
            $discount = $item_->discount > 0
                ? (($item_->discount_type == 1) ? $item_->discount : ($item_->discount * $item_price) / 100)
                : 0;

            // Item Unit Price
            if ($item_->unit_id != $item->unit_id) {
                $unit       = ItemUnit::where('unit_id', $item_->unit_id)->where('item_id', $item_->item_id)->first();
                $item_price = $unit->unit_value * $item_price;
            } else {
                $unit = Unit::where('id', $item_->unit_id)->first();
            }

            // net price is the unit price - discount
            $net_price  = $item_price - $discount;

            // Item Tax.
            $taxableItems = collect();

            // Get item vats.
            $totalItemTax = count($invoice->vats) > 0
                ? $this->receiptVats($invoice->vats, $taxableItems, $taxes, $net_price)
                : 0;

            // Calculate the total amount of the item.
            $itemTotal = $this->receiptTotal($item_price, $item_->quantity, $discount, $totalItemTax);

            // Non-taxable items discount.
            $itemDiscount = $item_->items_discount ?? 0;

            // Total amount of all the items sales.
            $totalSales += $itemTotal['totalSale'];

            // Total amount of all the items discount.
            $totalItemsDiscount += $itemDiscount;

            // Total amount of all the items extra discount.
            $extraReceiptDiscount += $item_->extra_discount ?? 0;

            // Total amount of all the items net price.
            $netAmount += $itemTotal['netTotal'];

            // Total amount of all the items fees.
            $feesAmount += $item_->fees ?? 0;

            // Total amount of all the items.
            $totalAmount += $itemTotal['total'];

            // Add the item to the item data.
            $receiptItem = [
                "internalCode"    => Str::random(12),
                "description"     => $item->about,
                "itemType"        => "EGS",
                "itemCode"        => $item->item_einvoice_code,
                "unitType"        => $unit->unit_einvoice_code,
                "quantity"        => (float) $item_->quantity,
                "unitPrice"       => (float) $item_price,
                "netSale"         => (float) $itemTotal['netSale'],
                "totalSale"       => (float) $itemTotal['totalSale'],
                "total"           => (float) $itemTotal['total'],
                "valueDifference" => (float) $itemTotal['valueDifference'],
            ];

            // If the item has a discount.
            if (isset($discount) && $discount > 0) {
                $receiptItem['commercialDiscount'] = [(float) $discount];
            }

            // If non-taxable items have a discount.
            if (isset($itemDiscount) && $itemDiscount > 0) {
                $receiptItem['itemsDiscount'] = [(float) $itemDiscount];
            }

            // If the item has a vat.
            if (count($taxableItems) > 0) {
                $receiptItem['taxableItems'] = $taxableItems;
            }

            // Total amount of all commercial items discount.
            $totalCommercialDiscount += $discount ?? 0;

            // Add the item to the item data.
            $itemData->push($receiptItem);
        }

        // Add the totals to the details.
        $details = [
            "totalSales"  => $totalSales,
            "netAmount"   => $netAmount,
            "feesAmount"  => $feesAmount,
            "totalAmount" => $totalAmount,
        ];

        // If totalCommercialDiscount is greater than 0 add it to the details.
        if ($totalCommercialDiscount > 0) {
            $details['totalCommercialDiscount'] = $totalCommercialDiscount;
        }

        // If totalItemsDiscount is greater than 0 add it to the details.
        if ($totalItemsDiscount > 0) {
            $details['totalItemsDiscount'] = $totalItemsDiscount;
        }

        // If extraReceiptDiscount is greater than 0 add it to the details.
        if ($extraReceiptDiscount > 0) {
            $details['extraReceiptDiscount'] = $extraReceiptDiscount;
        }

        // Return the details.
        return $details;
    }

    /**
     * This Method uses to get the item vats.
     * @param item_
     * @param vats
     * @param taxableItems
     * @param totalTaxableFees
     * @param taxes
     * @param net_price
     * @return array
     */
    public function receiptVats($vats, $taxableItems, $taxes, $net_price)
    {
        foreach ($vats as $vat) {

            // Get the vat from the bills_add
            $vat_ = Vat::where('id', $vat->vat_id)->first();

            /**
             * Add the vat to the taxable items
             * (vat_type = 0) => نسبة , (vat_type = 1) => مبلغ
             * (vat_operation = 0) => إضافة , (vat_operation = 1) => خصم
             */
            $rate   = $vat->vat_type == 0 ? $vat->vat_value : ($vat->vat_value / $net_price) * 100;
            $amount = $vat->vat_type == 1 ? $vat->vat_value : ($vat->vat_value * $net_price) / 100;

            $taxableItems->push([
                "taxType" => $vat_->einvoice_code,
                "rate"    => (float) $rate,
                "amount"  => (float) $amount,
                "subType" => $vat_->sub_einvoice_code
            ]);

            $taxes->push([
                "taxType" => $vat_->einvoice_code,
                "amount"  => (float) $amount,
            ]);
        }

        // Total item tax.
        $totalItemTax = $taxableItems->sum('amount');

        // Return the total item tax.
        return $totalItemTax;
    }

    /**
     * This Method is used to calculate the tax totals.
     *
     * @param taxes of the invoice
     * @return taxTotals
     */
    public function receiptTaxTotal($taxes)
    {
        $taxTotals = collect();

        // Calculate the tax totals amount.
        $taxes->groupBy('taxType')->each(function ($taxes, $key) use ($taxTotals) {
            $taxTotals->push([
                "taxType" => $key,
                "amount"  => (float) $taxes->sum('amount'),
            ]);
        });

        // Calculate the sum of all the tax totals.
        $totalTaxAmount = $taxTotals->sum('amount');

        // Return the tax totals.
        return [
            "taxTotals"      => $taxTotals,
            "totalTaxAmount" => $totalTaxAmount,
        ];
    }

    /**
     * This Method is used to calculate totals.
     *
     * @param item_price
     * @param quantity
     * @param totalTaxableFees
     * @param discount
     * @return itemTotal
     */
    public function receiptTotal($item_price, $quantity, $discount, $totalItemTax)
    {
        // Total amount considering quantity and unit price.
        $totalSale = $item_price * $quantity;

        // Net sale is the total sale - discount.
        $netSale = $totalSale - $discount;

        // Total amount after adding all pricing items, taxes, removing discounts.
        $total = $totalSale + $totalItemTax - $discount;

        // Value difference when selling goods already taxed.
        $valueDifference = $total - $totalSale;

        // Total amount for the invoice line after applying discount.
        $netTotal = $total - $totalItemTax;

        // Return the totals.
        return [
            "netSale"         => $netSale,
            "totalSale"       => $totalSale,
            "total"           => $total,
            'netTotal'        => $netTotal,
            "valueDifference" => 0 ?? $valueDifference,
        ];
    }
}
